// Mft0.cpp : Implementation of CMft0

#include "stdafx.h"
#include "Mft0.h"
#include "SampleHelpers.h"
#include <WinString.h>
#include "csc/csc.h"
#include "rotation/rotation.h"
#include "utils_mft.h"

#include <time.h>

HRESULT GetFormatWidthHeightFps(IMFMediaType *pMediaType, GUID *pGuidSubType, UINT *puiWidth, UINT *puiHeight, UINT *puiFps)
{
    HRESULT hr;
    UINT uiNumerator, uiDenominator;

    GOTO_END_NULL(pMediaType, "Failed to GetFormatWidthHeightFps() due to null pMediaType \n");
    GOTO_END_NULL(pGuidSubType, "Failed to GetFormatWidthHeightFps() due to null pGuidSubType \n");
    GOTO_END_NULL(puiWidth, "Failed to GetFormatWidthHeightFps() due to null puiWidth \n");
    GOTO_END_NULL(puiHeight, "Failed to GetFormatWidthHeightFps() due to null puiHeight \n");
    GOTO_END_NULL(puiFps, "Failed to GetFormatWidthHeightFps() due to null puiFps \n");

    hr = pMediaType->GetGUID(MF_MT_SUBTYPE, pGuidSubType);
    GOTO_END_FAILED(hr, "Failed to GetGUID() \n");

    hr = MFGetAttributeSize(pMediaType, MF_MT_FRAME_SIZE, puiWidth, puiHeight);
    GOTO_END_FAILED(hr, "Failed to MFGetAttributeSize() \n");

    hr = MFGetAttributeRatio(pMediaType, MF_MT_FRAME_RATE, &uiNumerator, &uiDenominator);
    GOTO_END_FAILED(hr, "Failed to MFGetAttributeRatio() \n");

    *puiFps = (uiDenominator == 0) ? 0 : (UINT)(uiNumerator/uiDenominator);

END:
    return hr;
}


STDMETHODIMP GetDefaultStride(IMFMediaType *m_pInputType, LONG *plStride)
{
    LONG lStride = 0;

    // Try to get the default stride from the media type.
    HRESULT hr = m_pInputType->GetUINT32(MF_MT_DEFAULT_STRIDE, (UINT32*)&lStride);
    if (FAILED(hr))
    {
        // Attribute not set. Try to calculate the default stride.
        GUID subtype = GUID_NULL;
        UINT32 width = 0;
        UINT32 height = 0;

        // Get the subtype and the image size.
        hr = m_pInputType->GetGUID(MF_MT_SUBTYPE, &subtype);
        if (SUCCEEDED(hr))
        {
            hr = MFGetAttributeSize(m_pInputType, MF_MT_FRAME_SIZE, &width, &height);
        }
        if (SUCCEEDED(hr))
        {
            hr = MFGetStrideForBitmapInfoHeader(subtype.Data1, width, &lStride);
        }

        // Set the attribute for later reference.
        if (SUCCEEDED(hr))
        {
            (void)m_pInputType->SetUINT32(MF_MT_DEFAULT_STRIDE, UINT32(lStride));
        }
    }

    if (SUCCEEDED(hr))
    {
        *plStride = lStride;
    }

    return hr;
}



// CMft0
STDMETHODIMP CMft0::UpdateDsp(UINT32 uiPercentOfScreen)
{
    m_percentOfScreen = uiPercentOfScreen;
    return S_OK;
}

STDMETHODIMP CMft0::Enable(void)
{
    m_bEnableEffects = TRUE;
    return S_OK;
}

STDMETHODIMP CMft0::Disable(void)
{
    m_bEnableEffects = FALSE;
    return S_OK;
}

STDMETHODIMP CMft0::GetDspSetting(UINT* puiPercentOfScreen, BOOL* pIsEnabled)
{
    HRESULT hr = S_OK;
    do {
        CHK_NULL_PTR_BRK(puiPercentOfScreen);
        CHK_NULL_PTR_BRK(pIsEnabled);
        *puiPercentOfScreen = m_percentOfScreen;
        *pIsEnabled = m_bEnableEffects;
    } while (FALSE);

    return S_OK;
}

STDMETHODIMP CMft0::GetIids( 
    /* [out] */ _Out_ ULONG *iidCount,
    /* [size_is][size_is][out] */ _Outptr_result_buffer_maybenull_(*iidCount) IID **iids)
{
    HRESULT hr = S_OK;
    do {
        CHK_NULL_PTR_BRK(iidCount);
        CHK_NULL_PTR_BRK(iids);
        *iids = NULL;
        *iidCount = 0;
    } while (FALSE);

    return hr;
}

STDMETHODIMP CMft0::GetRuntimeClassName( 
    /* [out] */ _Outptr_result_maybenull_ HSTRING *className)
{
    HRESULT hr = S_OK;
    if(className != nullptr)
    {
        hr = WindowsCreateString(NULL, 0, className);
        if(FAILED(hr)){
            hr = E_OUTOFMEMORY;
        }
    }else {
        hr = E_INVALIDARG;
    }
    return  hr;
}

STDMETHODIMP CMft0::GetTrustLevel( 
    /* [out] */ _Out_ TrustLevel *trustLevel)
{
    HRESULT hr = S_OK;
    do {
        CHK_NULL_PTR_BRK(trustLevel);
        *trustLevel = TrustLevel::BaseTrust;
    } while (FALSE);

    return hr;
}


STDMETHODIMP CMft0::GetStreamLimits( 
    /* [out] */ _Out_ DWORD *pdwInputMinimum,
    /* [out] */ _Out_ DWORD *pdwInputMaximum,
    /* [out] */ _Out_ DWORD *pdwOutputMinimum,
    /* [out] */ _Out_ DWORD *pdwOutputMaximum)
{
    HRESULT hr = S_OK;

    do {
        if ((pdwInputMinimum == NULL) ||
            (pdwInputMaximum == NULL) ||
            (pdwOutputMinimum == NULL) ||
            (pdwOutputMaximum == NULL))
        {
            hr = E_POINTER;
            break;
        }

        if (m_bMft0IsEnabled)
        {
            // This MFT has a fixed number of streams.
            *pdwInputMinimum = 1;
            *pdwInputMaximum = 1;
            *pdwOutputMinimum = 1;
            *pdwOutputMaximum = 1;
        }
        else
        {
            *pdwInputMinimum = 0;
            *pdwInputMaximum = 0;
            *pdwOutputMinimum = 0;
            *pdwOutputMaximum = 0;
        }
    } while (FALSE);

    return hr;    
}


STDMETHODIMP CMft0::GetStreamCount( 
    /* [out] */ _Out_ DWORD *pcInputStreams,
    /* [out] */ _Out_ DWORD *pcOutputStreams)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::GetStreamCount() \n");

    if ((pcInputStreams == NULL) || (pcOutputStreams == NULL))
    {
        hr = E_POINTER;
        PLOG("Failed to CMft0::GetStreamCount() due to null pointer pcInputStreams or pcOutputStreams \n");
        goto END;
    }

    if (m_bMft0IsEnabled)
    {
        // This MFT has a fixed number of streams.
        *pcInputStreams = 1;
        *pcOutputStreams = 1;
    }
    else
    {
        *pcInputStreams = 0;
        *pcOutputStreams = 0;
        PLOG("====== CMft0::GetStreamCount() Disabled the MFT0 due to the current system configures \n");
    }
END:
    PLOG("<<<=== CMft0::GetStreamCount() \n");
    return hr;
}

STDMETHODIMP CMft0::GetStreamIDs( 
    DWORD dwInputIDArraySize,
    /* [size_is][out] */ _Out_writes_(dwInputIDArraySize) DWORD *pdwInputIDs,
    DWORD dwOutputIDArraySize,
    /* [size_is][out] */ _Out_writes_(dwOutputIDArraySize) DWORD *pdwOutputIDs)
{
    dwOutputIDArraySize = 0;
    dwInputIDArraySize = 0;
    pdwInputIDs = NULL;
    pdwOutputIDs = NULL;
    return E_NOTIMPL; 
}

STDMETHODIMP CMft0::GetInputStreamInfo( 
    DWORD dwInputStreamID,
    /* [out] */ _Out_ MFT_INPUT_STREAM_INFO *pStreamInfo)
{
    HRESULT hr = S_OK;
    EnterCriticalSection(&m_critSec);
    
    do {
        CHK_NULL_BRK(pStreamInfo);
        if (!IsValidInputStream(dwInputStreamID))
        {
            CHK_LOG_BRK(MF_E_INVALIDSTREAMNUMBER);
        }
        if(m_pInputType) {
            pStreamInfo->cbAlignment = 0;
            pStreamInfo->cbSize = 0;
            pStreamInfo->dwFlags = MFT_INPUT_STREAM_WHOLE_SAMPLES | MFT_INPUT_STREAM_SINGLE_SAMPLE_PER_BUFFER ;
            pStreamInfo->hnsMaxLatency = 0;
        } else {
            hr = MF_E_TRANSFORM_TYPE_NOT_SET;
        }
    } while (FALSE);

    LeaveCriticalSection(&m_critSec);
    return hr;
}


STDMETHODIMP CMft0::GetOutputStreamInfo( 
    DWORD dwOutputStreamID,
    /* [out] */ _Out_ MFT_OUTPUT_STREAM_INFO *pStreamInfo)
{
    HRESULT hr = S_OK;
    EnterCriticalSection(&m_critSec);

    do {
        CHK_NULL_BRK(pStreamInfo);
        if (!IsValidInputStream(dwOutputStreamID))
        {
            CHK_LOG_BRK(MF_E_INVALIDSTREAMNUMBER);
        }
        if(m_pOutputType) {
            pStreamInfo->cbAlignment = 0;
            pStreamInfo->cbSize = 0;
            pStreamInfo->dwFlags = MFT_OUTPUT_STREAM_WHOLE_SAMPLES  | MFT_OUTPUT_STREAM_SINGLE_SAMPLE_PER_BUFFER | MFT_OUTPUT_STREAM_PROVIDES_SAMPLES | MFT_OUTPUT_STREAM_FIXED_SAMPLE_SIZE; 
        } else {
            hr = MF_E_TRANSFORM_TYPE_NOT_SET;
        }

    } while (FALSE);
    
    LeaveCriticalSection(&m_critSec);
    return hr;
}

STDMETHODIMP CMft0::GetAttributes( 
    /* [out] */ _Outptr_result_maybenull_ IMFAttributes **ppAttributes)
{
    HRESULT hr = S_OK;

    do {
        CHK_NULL_PTR_BRK(ppAttributes);
        if(!m_pGlobalAttributes) {
            CHK_LOG_BRK(MFCreateAttributes(&m_pGlobalAttributes, 3));
            CHK_LOG_BRK(m_pGlobalAttributes->SetUINT32(MF_TRANSFORM_ASYNC, FALSE));
            CHK_LOG_BRK(m_pGlobalAttributes->SetString(MFT_ENUM_HARDWARE_URL_Attribute, L"Sample_CameraExtensionMft"));
            CHK_LOG_BRK(m_pGlobalAttributes->SetUINT32(MFT_SUPPORT_DYNAMIC_FORMAT_CHANGE, FALSE));
        }
        *ppAttributes = m_pGlobalAttributes;
        (*ppAttributes)->AddRef();
    } while (FALSE);
    
    return hr;
}

STDMETHODIMP CMft0::GetInputStreamAttributes( 
    DWORD dwInputStreamID,
    /* [out] */ _Outptr_result_maybenull_ IMFAttributes **ppAttributes)
{
    HRESULT hr = S_OK;
    IUnknown *pUnk = NULL;
    IMFAttributes *pSourceAttributes = NULL;

    PLOG(">>>=== CMft0::GetInputStreamAttributes dwInputStreamID=[%d] \n", dwInputStreamID);

    do {
        if (dwInputStreamID > 0) { hr = MF_E_INVALIDSTREAMNUMBER; break; }

        CHK_NULL_PTR_BRK(ppAttributes);
        if (!m_pInputAttributes)
        {
            CHK_LOG_BRK(MFCreateAttributes(&m_pInputAttributes, 2));
            CHK_LOG_BRK(m_pInputAttributes->SetUINT32(MFT_SUPPORT_DYNAMIC_FORMAT_CHANGE, TRUE));
            CHK_LOG_BRK(m_pInputAttributes->SetString(MFT_ENUM_HARDWARE_URL_Attribute, L"Sample_CameraExtensionMft"));
        }
        *ppAttributes = m_pInputAttributes;
        (*ppAttributes)->AddRef();
    } while (FALSE);
    
    SAFERELEASE(pUnk);
    SAFERELEASE(pSourceAttributes);
    PLOG("<<<=== CMft0::GetInputStreamAttributes \n");
    return hr;
}

STDMETHODIMP CMft0::GetOutputStreamAttributes( 
    DWORD dwOutputStreamID,
    /* [out] */ _Outptr_result_maybenull_ IMFAttributes **ppAttributes)
{
    HRESULT hr = S_OK;

    do {
        if(dwOutputStreamID > 0) {
            hr = MF_E_INVALIDSTREAMNUMBER;
            break;
        }
        CHK_NULL_PTR_BRK(ppAttributes);
        CHK_NULL_PTR_BRK(m_pInputAttributes);

        *ppAttributes = m_pInputAttributes;
        (*ppAttributes)->AddRef();
    } while (FALSE);
    
    return hr;
}

STDMETHODIMP CMft0::DeleteInputStream( 
    DWORD dwStreamID)
{
    HRESULT hr = S_OK;
    if(dwStreamID > 0) {
        hr = MF_E_INVALIDSTREAMNUMBER;
    } else {
        hr = E_NOTIMPL;
    }
    return hr; 
}

// https://msdn.microsoft.com/en-us/library/windows/desktop/ms696211(v=vs.85).aspx
STDMETHODIMP CMft0::AddInputStreams( 
    DWORD cStreams,
    /* [in] */ _In_ DWORD *adwStreamIDs)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::AddInputStreams() \n");
    
    if( !adwStreamIDs || cStreams > 0) {
        hr = E_INVALIDARG;
    } else {
        hr = E_NOTIMPL;
    }
    PLOG("<<<=== CMft0::AddInputStreams() \n");
    return hr; 
}

STDMETHODIMP CMft0::GetInputAvailableType( 
    DWORD dwInputStreamID,
    DWORD dwTypeIndex,
    /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType)
{
    HRESULT hr = S_OK;
    IUnknown *pUnk = NULL;
    IMFAttributes *pSourceAttributes = NULL;
    wchar_t *pszName;
    PLOG(">>>=== CMft0::GetInputAvailableType() dwInputStreamID=[%d] dwTypeIndex=[%d] \n", dwInputStreamID, dwTypeIndex);

    EnterCriticalSection(&m_critSec);

    if (!m_bMft0IsEnabled)
    {
        hr = MF_E_NO_MORE_TYPES;
        goto END;
    }

    do {
        CHK_BOOL_BRK(IsValidInputStream(dwInputStreamID));

        if(!m_pSourceTransform && m_pInputAttributes) {
            CHK_LOG_BRK(m_pInputAttributes->GetUnknown(MFT_CONNECTED_STREAM_ATTRIBUTE, IID_IMFAttributes, IID_PPV_ARGS_Helper(&pSourceAttributes)));
            CHK_LOG_BRK(pSourceAttributes->GetUnknown(MF_DEVICESTREAM_EXTENSION_PLUGIN_CONNECTION_POINT, IID_IUnknown, IID_PPV_ARGS_Helper(&pUnk)));
            CHK_LOG_BRK(pUnk->QueryInterface(__uuidof(IMFTransform), (void**)&m_pSourceTransform));
            CHK_LOG_BRK(pSourceAttributes->GetGUID( MF_DEVICESTREAM_STREAM_CATEGORY, &m_stStreamType));

            if (IsEqualGUID(m_stStreamType, PINNAME_VIDEO_CAPTURE)) {
                wprintf(L"Stream type: PINNAME_VIDEO_CAPTURE\n");
            } else if (IsEqualGUID(m_stStreamType, PINNAME_VIDEO_PREVIEW)) {
                wprintf(L"Stream type: PINNAME_VIDEO_PREVIEW\n");
            } else if (IsEqualGUID(m_stStreamType, PINNAME_VIDEO_STILL)) {
                wprintf(L"Stream type: PINNAME_VIDEO_STILL\n");
            } else if (IsEqualGUID(m_stStreamType, PINNAME_IMAGE)) {
                wprintf(L"Stream type: PINNAME_IMAGE\n");
            } else {
                StringFromCLSID(m_stStreamType, &pszName);
                if(pszName){
                    wprintf_s(L"Stream type: %s\n", pszName);
                    CoTaskMemFree(pszName);
                }
            }
            CHK_LOG_BRK((IsEqualGUID(m_stStreamType, PINNAME_VIDEO_PREVIEW) || IsEqualGUID(m_stStreamType, PINNAME_VIDEO_CAPTURE)) ? S_OK : E_UNEXPECTED);
            CHK_LOG_BRK(GenerateMFMediaTypeListFromDevice(dwInputStreamID));
        }

        CHK_LOG_BRK(GetMediaType(dwInputStreamID, dwTypeIndex, ppType));
    } while (FALSE);

END:
    LeaveCriticalSection(&m_critSec);
    SAFERELEASE(pUnk);
    SAFERELEASE(pSourceAttributes);
    PLOG("<<<=== CMft0::GetInputAvailableType() \n");
    return hr;
}

STDMETHODIMP CMft0::GetOutputAvailableType( 
    DWORD dwOutputStreamID,
    DWORD dwTypeIndex,
    /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType)
{
    HRESULT hr = S_OK;
    IUnknown *pUnk = NULL;
    IMFAttributes *pSourceAttributes = NULL;
    BOOL bStreamIdIsValid;

    if (!m_bMft0IsEnabled)
    {
        hr = MF_E_NO_MORE_TYPES;
        goto END;
    }

    PLOG(">>>=== CMft0::GetOutputAvailableType() dwOutputStreamID=[%d] dwTypeIndex=[%d] \n", dwOutputStreamID, dwTypeIndex);

    EnterCriticalSection(&m_critSec);
    bStreamIdIsValid = IsValidOutputStream(dwOutputStreamID);
    GOTO_END_FALSE(bStreamIdIsValid, "Failed to GetOutputAvailableType() due to invalid dwOutputStreamID \n");

    if (!m_pSourceTransform && m_pInputAttributes)
    {
        wchar_t *pszName;

        hr = m_pInputAttributes->GetUnknown(MFT_CONNECTED_STREAM_ATTRIBUTE, IID_IMFAttributes, IID_PPV_ARGS_Helper(&pSourceAttributes));
        GOTO_END_FAILED(hr , "Failed to m_pInputAttributes->GetUnknown \n");

        hr = pSourceAttributes->GetUnknown(MF_DEVICESTREAM_EXTENSION_PLUGIN_CONNECTION_POINT, IID_IUnknown, IID_PPV_ARGS_Helper(&pUnk));
        GOTO_END_FAILED(hr , "Failed to pSourceAttributes->GetUnknown \n");

        hr = pUnk->QueryInterface(__uuidof(IMFTransform), (void**)&m_pSourceTransform);
        GOTO_END_FAILED(hr , "Failed to pUnk->QueryInterface IMFTransform \n");

        hr = pSourceAttributes->GetGUID(MF_DEVICESTREAM_STREAM_CATEGORY, &m_stStreamType);

        if (IsEqualGUID(m_stStreamType, PINNAME_VIDEO_CAPTURE)) {
            PLOG("====== CMft0::GetOutputAvailableType() Stream type: PINNAME_VIDEO_CAPTURE \n");
            hr = S_OK;
        } else if (IsEqualGUID(m_stStreamType, PINNAME_VIDEO_PREVIEW)) {
            PLOG("====== CMft0::GetOutputAvailableType() Stream type: PINNAME_VIDEO_PREVIEW \n");
            hr = S_OK;
        } else if (IsEqualGUID(m_stStreamType, PINNAME_VIDEO_STILL)) {
            PLOG("====== CMft0::GetOutputAvailableType() Stream type: PINNAME_VIDEO_STILL \n");
            hr = E_UNEXPECTED;
        } else if (IsEqualGUID(m_stStreamType, PINNAME_IMAGE)) {
            PLOG("====== CMft0::GetOutputAvailableType() Stream type: PINNAME_IMAGE \n");
            hr = E_UNEXPECTED;
        } else {
            StringFromCLSID(m_stStreamType, &pszName);
            PLOG("====== CMft0::GetOutputAvailableType() Stream type: %s \n", pszName);
            hr = E_UNEXPECTED;
        }

        hr = GenerateMFMediaTypeListFromDevice(dwOutputStreamID);
        GOTO_END_FAILED(hr, "Failed to GenerateMFMediaTypeListFromDevice() \n");
    }

    hr = GetMediaTypeOut(dwOutputStreamID, dwTypeIndex, ppType); 
    GOTO_END_FAILED(hr, "Failed to GetMediaTypeOut() \n");

END:
    LeaveCriticalSection(&m_critSec);
    SAFERELEASE(pUnk);
    SAFERELEASE(pSourceAttributes);
    if (ppType) {
        PLOG_MEDIATYPE(*ppType, "<<<=== CMft0::GetOutputAvailableType() ");
    } else {
        PLOG("<<<=== CMft0::GetOutputAvailableType() ppType=[NULL] \n");
    }
    return hr;
}

// https://msdn.microsoft.com/en-us/library/windows/desktop/ms700113%28v=vs.85%29.aspx
STDMETHODIMP CMft0::SetInputType(DWORD dwInputStreamID, IMFMediaType *pType, DWORD dwFlags)
{
    HRESULT hr = S_OK;

    PLOG_MEDIATYPE(pType, ">>>=== CMft0::SetInputType() MFT_SET_TYPE_TEST_ONLY=[%d] ", dwFlags);
    EnterCriticalSection(&m_critSec);

    BOOL bReallySet = ((dwFlags & MFT_SET_TYPE_TEST_ONLY) == 0);
    CComPtr<IMFMediaType> pFullType;

    if (dwInputStreamID != 0)
    {
        hr = MF_E_INVALIDSTREAMNUMBER;
        goto END;
    }

    hr = IsMediaTypeSupported(dwInputStreamID, pType, &pFullType);
    GOTO_END_FAILED(hr, "Failed to IsMediaTypeSupported() \n");

    // Validate flags.
    if (bReallySet)
    {
        SAFERELEASE(m_pInputType);
        m_pInputType = pFullType;
        m_pInputType->AddRef();

        hr = MFGetAttributeSize(pType, MF_MT_FRAME_SIZE, &m_InputTypeWidth, &m_InputTypeHeight);
        GOTO_END_FAILED(hr, "Failed to MFGetAttributeSize() \n");
    }

END:
    LeaveCriticalSection(&m_critSec);
    PLOG("<<<=== CMft0::SetInputType() hr=[%d] \n", hr);
    return hr;
}


HRESULT CMft0::FindMediaType(IMFMediaType **ppMediaTypeFound, GUID guidSubTypeWanted, UINT uiWidthWanted, UINT uiHeightWanted, UINT uiFpsWanted)
{
    HRESULT hr = S_OK;
    BOOL bFound = FALSE;
    CHAR szGuidSubtype[64] = {0};

    sGuidPrint(szGuidSubtype, 64, guidSubTypeWanted);
    PLOG(">>>=== CMft0::FindMediaType() guidSubTypeWanted=[%s] uiWidthWanted=[%d] uiHeightWanted=[%d] uiFpsWanted=[%d] \n",
        szGuidSubtype, uiWidthWanted, uiHeightWanted, uiFpsWanted);

    GOTO_END_NULL(m_pSourceTransform, "Failed to FindMediaType() due to null pointer m_pSourceTransform \n");
    GOTO_END_NULL(ppMediaTypeFound, "Failed to FindMediaType() due to null pointer ppMediaTypeFound \n");

    for (UINT iIdx = 0; hr == S_OK; iIdx++) 
    {
        CComPtr<IMFMediaType> pMediaType;
        UINT uiWidth, uiHeight, uiFps;
        GUID GuidSubType;

        hr = m_pSourceTransform->GetOutputAvailableType(0, iIdx, &pMediaType);
        if (hr == MF_E_NO_MORE_TYPES)
        {
            hr = S_OK; // This is an exception case, so we treat it as S_OK.
            break;
        }

        GOTO_END_FAILED(hr, "Failed to GetOutputAvailableType() \n");

        hr = GetFormatWidthHeightFps(pMediaType, &GuidSubType, &uiWidth, &uiHeight, &uiFps);
        GOTO_END_FAILED(hr, "Failed to GetFormatWidthHeightFps() \n");

        if (GuidSubType == guidSubTypeWanted &&
            uiWidth == uiWidthWanted &&
            uiHeightWanted == uiHeight &&
            uiFpsWanted == uiFps)
        {
            bFound = TRUE;
            *ppMediaTypeFound = pMediaType;
            goto END;
        }
    }
END:
    hr = (bFound)? S_OK: E_FAIL;
    PLOG("<<<=== CMft0::FindMediaType() hr=[%d] bFound=[%d] \n", hr, bFound);
    return hr;
}


STDMETHODIMP CMft0::SetOutputType(DWORD dwOutputStreamID, IMFMediaType *pMediaTypeOut, DWORD dwFlags)
{
    HRESULT hr = S_OK;
    BOOL bReallySet = ((dwFlags & MFT_SET_TYPE_TEST_ONLY) == 0);
    GUID guidSubtypeOut = {0};
    UINT uiWidthOut, uiHeightOut, uiFpsOut;
    IMFMediaType *pMediaTypeInput = NULL;

    UINT uiWidthIn, uiHeightIn, uiFpsIn;
    GUID guidSubtypeIn = {0};

    PLOG_MEDIATYPE(pMediaTypeOut, ">>>=== CMft0::SetOutputType() dwOutputStreamID=[%d] bReallySet=[%d] ", dwOutputStreamID, bReallySet);
    EnterCriticalSection(&m_critSec);

    // check input param
    if (dwOutputStreamID != 0)
    {
        PLOG("Failed to CMft0::SetOutputType() due to invalid dwOutputStreamID \n");
        hr = MF_E_INVALIDINDEX; 
        goto END;
    }

    if (!pMediaTypeOut)
    {
        PLOG("Failed to CMft0::SetOutputType() due to null pointer pMediaTypeOut \n");
        hr = MF_E_INVALIDMEDIATYPE;
        goto END;
    }

    hr = GetFormatWidthHeightFps(pMediaTypeOut, &guidSubtypeOut, &uiWidthOut, &uiHeightOut, &uiFpsOut);
    GOTO_END_FAILED(hr, "Failed to GetFormatWidthHeightFps() \n");

    for (INT i = 0; i < m_listOfMediaTypesIn.GetCount(); i++)
    {
        hr = GetFormatWidthHeightFps(m_listOfMediaTypesIn[i], &guidSubtypeIn, &uiWidthIn, &uiHeightIn, &uiFpsIn);
        GOTO_END_FAILED(hr, "Failed to GetFormatWidthHeightFps() \n");
        
        if (uiWidthOut == uiWidthIn && uiHeightOut == uiHeightIn && uiFpsOut == uiFpsIn)
        {
            pMediaTypeInput = m_listOfMediaTypesIn[i];
            break;
        }
    }

    if (!pMediaTypeInput)
    {
        PLOG("Failed to CMft0::SetOutputType() \n");
        hr = E_FAIL;
        goto END;
    }

    if (bReallySet)
    {
        SAFERELEASE(m_pOutputType);

        hr = MFCreateMediaType(&m_pOutputType);
        GOTO_END_FAILED(hr, "Failed to MFCreateMediaType() \n");

        hr = pMediaTypeOut->CopyAllItems(m_pOutputType);
        GOTO_END_FAILED(hr, "Failed to CopyAllItems() \n");

        m_pOutputType->AddRef();

        SAFERELEASE(m_pInputType);
        m_pInputType = pMediaTypeInput;
        m_pInputType->AddRef();

        m_InputTypeWidth    = uiWidthIn;
        m_InputTypeHeight   = uiHeightIn;

        m_OutputTypeWidth   = uiWidthOut;
        m_OutputTypeHeight  = uiHeightOut;
    }

END:
    LeaveCriticalSection(&m_critSec);
    PLOG_MEDIATYPE(m_pInputType, "====== CMft0::SetOutputType() m_pInputType: ");
    PLOG_MEDIATYPE(m_pOutputType, "====== CMft0::SetOutputType() m_pOutputType: ");
    PLOG("<<<=== CMft0::SetOutputType() hr=[%d] \n", hr);
    return hr;
}


STDMETHODIMP CMft0::GetInputCurrentType( 
    DWORD dwInputStreamID,
    /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::GetInputCurrentType() dwInputStreamID=[%d] \n", dwInputStreamID);

    EnterCriticalSection(&m_critSec);

    do {
        CHK_NULL_BRK(ppType);

        if (!IsValidInputStream(dwInputStreamID))
        {
            CHK_LOG_BRK(MF_E_INVALIDSTREAMNUMBER);
        }
        else if(m_pInputType)
        {
            *ppType = m_pInputType;
            (*ppType)->AddRef();
        }
        else 
        {
            CHK_LOG_BRK(MF_E_TRANSFORM_TYPE_NOT_SET);
        }
    } while (FALSE);

    LeaveCriticalSection(&m_critSec);
    PLOG_MEDIATYPE(m_pInputType, "<<<=== CMft0::GetInputCurrentType() m_pInputType= ");
    return hr;
}

STDMETHODIMP CMft0::GetOutputCurrentType( 
    DWORD dwOutputStreamID,
    /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::GetOutputCurrentType() dwOutputStreamID=[%d] \n", dwOutputStreamID);

    EnterCriticalSection(&m_critSec);

    do {
        CHK_NULL_BRK(ppType);

        if (!IsValidOutputStream(dwOutputStreamID))
        {
            CHK_LOG_BRK(MF_E_INVALIDSTREAMNUMBER);
        }
        else if(m_pOutputType) 
        {
            *ppType = m_pOutputType;
            (*ppType)->AddRef();
        }
        else
        {
            CHK_LOG_BRK(MF_E_TRANSFORM_TYPE_NOT_SET);
        }
    } while (FALSE);

    LeaveCriticalSection(&m_critSec);
    PLOG_MEDIATYPE(m_pOutputType, "<<<=== CMft0::GetOutputCurrentType() m_pOutputType= ");
    return hr;
}

STDMETHODIMP CMft0::GetInputStatus( 
    DWORD dwInputStreamID,
    /* [out] */ _Out_ DWORD *pdwFlags)
{
    HRESULT hr = S_OK;

    EnterCriticalSection(&m_critSec);

    do {
        CHK_BOOL_BRK(pdwFlags);

        if (!IsValidInputStream(dwInputStreamID))
        {
            CHK_LOG_BRK(MF_E_INVALIDSTREAMNUMBER);
        }

        // If we already have an input sample, we don't accept
        // another one until the client calls ProcessOutput or Flush.
        if (m_pSample == NULL)
        {
            *pdwFlags = MFT_INPUT_STATUS_ACCEPT_DATA;
        }
        else
        {
            *pdwFlags = 0;
        }
    } while (FALSE);
    
    LeaveCriticalSection(&m_critSec);
    return hr;
}

STDMETHODIMP CMft0::GetOutputStatus( 
    /* [out] */ _Out_ DWORD *pdwFlags)
{
    HRESULT hr = S_OK;

    EnterCriticalSection(&m_critSec);

    do {
        CHK_NULL_BRK(pdwFlags);
        // We can produce an output sample if (and only if)
        // we have an input sample.
        if (m_pSample != NULL)
        {
            *pdwFlags = MFT_OUTPUT_STATUS_SAMPLE_READY;
        }
        else
        {
            *pdwFlags = 0;
        }
    } while (FALSE);

    LeaveCriticalSection(&m_critSec);
    return hr;
}

STDMETHODIMP CMft0::SetOutputBounds( 
    LONGLONG hnsLowerBound,
    LONGLONG hnsUpperBound)
{
    UNREFERENCED_PARAMETER(hnsLowerBound);
    UNREFERENCED_PARAMETER(hnsUpperBound);
    return S_OK;
}

STDMETHODIMP CMft0::ProcessEvent( 
    DWORD dwInputStreamID,
    /* [in] */ _In_opt_ IMFMediaEvent *pEvent)
{
    UNREFERENCED_PARAMETER(dwInputStreamID);
    UNREFERENCED_PARAMETER(pEvent);
    return E_NOTIMPL;
}

STDMETHODIMP CMft0::ProcessMessage( 
    MFT_MESSAGE_TYPE eMessage,
    ULONG_PTR ulParam)
{
    UNREFERENCED_PARAMETER(ulParam);
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::ProcessMessage() \n");

    EnterCriticalSection(&m_critSec);

    switch (eMessage)
    {
    case MFT_MESSAGE_COMMAND_FLUSH:
        // Flush the MFT.
        hr = OnFlush();
        break;

    case MFT_MESSAGE_COMMAND_DRAIN:
        // Drain: Tells the MFT not to accept any more input until
        // all of the pending output has been processed. That is our
        // default behevior already, so there is nothing to do.
        break;

    case MFT_MESSAGE_SET_D3D_MANAGER:
        // The pipeline should never send this message unless the MFT
        // has the MF_SA_D3D_AWARE attribute set to TRUE. However, if we
        // do get this message, it's invalid and we don't implement it.
        hr = E_NOTIMPL;
        break;

    case MFT_MESSAGE_NOTIFY_BEGIN_STREAMING:
        PLOG("====== CMft0::ProcessMessage() MFT_MESSAGE_NOTIFY_BEGIN_STREAMING \n");

        break;
    case MFT_MESSAGE_NOTIFY_END_STREAMING:
        PLOG("====== CMft0::ProcessMessage() MFT_MESSAGE_NOTIFY_END_STREAMING \n");

        hr = m_FdfbFilter.Destroy();
        GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Destroy() \n");

        hr = m_MjpgDecoder.Destroy();
        GOTO_END_FAILED(hr, "Failed m_MjpgDecoder.Destroy() \n");

        hr = m_RotationCropFilter.Destroy();
        GOTO_END_FAILED(hr, "Failed m_RotationCropFilter.Destroy() \n");

        hr = m_RotationFillFilter.Destroy();
        GOTO_END_FAILED(hr, "Failed m_RotationFillFilter.Destroy() \n");

        break;
    case MFT_MESSAGE_NOTIFY_END_OF_STREAM:
        PLOG("====== CMft0::ProcessMessage() MFT_MESSAGE_NOTIFY_END_OF_STREAM \n");

        if (CreateDirectory(MFT0_NOTIFY_TAG_FOLDER, NULL) || ERROR_ALREADY_EXISTS == GetLastError())
        {
            CHAR szPathMft0Close[128] = {0};
            sprintf_s(szPathMft0Close, 128 , "%s\\close", MFT0_NOTIFY_TAG_FOLDER);

            if (CreateDirectory(szPathMft0Close, NULL) || ERROR_ALREADY_EXISTS == GetLastError())
            {
                sprintf_s(szPathMft0Close, 128 , "%s\\close\\fb_close.log", MFT0_NOTIFY_TAG_FOLDER);
                std::fstream inputFileStream(szPathMft0Close, std::ios::out);
                inputFileStream.close();
                PLOG("====== CMft0::ProcessMessage(), Create close notifying file path=[%s]", szPathMft0Close);
            }
        }
        break;
    case MFT_MESSAGE_NOTIFY_START_OF_STREAM:
        PLOG("====== CMft0::ProcessMessage() MFT_MESSAGE_NOTIFY_START_OF_STREAM \n");

        hr = m_MjpgDecoder.Destroy();
        GOTO_END_FAILED(hr, "Failed m_MjpgDecoder.Destroy() \n");

        hr = m_MjpgDecoder.Init(FOURCC_NV12, m_InputTypeWidth, m_InputTypeHeight);
        GOTO_END_FAILED(hr, "Failed to m_MjpgDecoder.Init() \n");

        hr = m_FdfbFilter.Destroy();
        GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Destroy() \n");

#if HOT_FIX_FB_HEIGHT_LIMIT_HW
        hr = m_RotationCropFilter.Destroy();
        GOTO_END_FAILED(hr, "Failed m_RotationCropFilter.Destroy() \n");

        hr = m_RotationFillFilter.Destroy();
        GOTO_END_FAILED(hr, "Failed m_RotationFillFilter.Destroy() \n");
#endif

        hr = m_CscFilter_Yuy2ToNv12.Destroy();
        GOTO_END_FAILED(hr, "Failed m_CscFilter_Yuy2ToNv12.Destroy() \n");

        if (!GetDwordRegKey("VCAM_FB_DETECTION_ANGLE", m_dwRegKeyFbDetectionAngleCurrent))
        {
            m_dwRegKeyFbDetectionAngleCurrent = 0;
            m_dwRegKeyFbDetectionAnglePrevious = 0;
        }
        else
        {
            m_dwRegKeyFbDetectionAnglePrevious = m_dwRegKeyFbDetectionAngleCurrent;
        }

        if (m_dwRegKeyFbDetectionAngleCurrent == 90 || m_dwRegKeyFbDetectionAngleCurrent == 270)
        {
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
            INT iTargetW = m_InputTypeHeight;
            INT iTargetH = (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)? HOT_FIX_FB_HEIGHT_MAX: m_InputTypeWidth;

            hr = m_RotationCropFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight, iTargetW, iTargetH);
            GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.Init() \n");

            hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, iTargetW, iTargetH);
            GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");

            hr = m_RotationFillFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, iTargetW, iTargetH, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.Init() \n");

#else // HOT_FIX_FB_HEIGHT_LIMIT_SW
            INT iTargetW = m_InputTypeHeight;
            INT iTargetH = (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)? HOT_FIX_FB_HEIGHT_MAX: m_InputTypeWidth;

            hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, iTargetW, iTargetH);
#endif
        }
        else if (m_dwRegKeyFbDetectionAngleCurrent == 180)
        {
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
            hr = m_RotationCropFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.Init() \n");

            hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");

            hr = m_RotationFillFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.Init() \n");
#else // HOT_FIX_FB_HEIGHT_LIMIT_SW
            hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");
#endif
        }
        else
        {
            hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");
        }

        if (SUCCEEDED(hr))
        {
            m_FdfbFilterState = stInitialized;
        }
        else
        {
            m_FdfbFilterState = stInitFailed;
            PLOG("Failed to m_FdfbFilter.Init() \n");
            hr = S_OK; // continue even init fdfb failed (case if hardware not support fb)
        }

        GUID guidInputSubType = {0};
        GUID guidOutputSubType = {0};

        hr = m_pInputType->GetGUID(MF_MT_SUBTYPE, &guidInputSubType);
        GOTO_END_FAILED(hr, "Failed to GetGUID() \n");

        hr = m_pInputType->GetGUID(MF_MT_SUBTYPE, &guidOutputSubType);
        GOTO_END_FAILED(hr, "Failed to GetGUID() \n");

        if (guidInputSubType == MFVideoFormat_YUY2 && guidOutputSubType == MFVideoFormat_YUY2)
        {
            hr = m_CscFilter_Yuy2ToNv12.Init(FOURCC_NV12, FOURCC_YUY2, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_CscFilter_Yuy2ToNv12() \n");
        }

        if (!GetDwordRegKey("VCAM_FB_GLOBAL_ENABLED", m_dwRegKeyFbGlobalEnabled))
        {
            m_dwRegKeyFbGlobalEnabled = FALSE;
        }

        m_pRegkeyListener = new RegKeyListener();
        GOTO_END_NULL(m_pRegkeyListener, "Failed to new RegKeyListener() \n");

        hr = m_pRegkeyListener->Init();
        GOTO_END_FAILED(hr, "Failed to m_pRegkeyListener->Init() \n");

        hr = m_pRegkeyListener->Listen();
        if (FAILED(hr))
        {
            PLOG("Failed to m_pRegkeyListener->Listen(), registry keys under [LMHK\\%s] didn't exist \n", VCAM_CONTROL_REGKEY_PATH);
            hr = S_OK; // ignore the error so MFT0 won't break the camera metro app or SkypePreview.
        }

        if (CreateDirectory(MFT0_NOTIFY_TAG_FOLDER, NULL) || ERROR_ALREADY_EXISTS == GetLastError())
        {
            CHAR szPathMft0Start[128] = {0};
            sprintf_s(szPathMft0Start, 128 , "%s\\start", MFT0_NOTIFY_TAG_FOLDER);

            if (CreateDirectory(szPathMft0Start, NULL) || ERROR_ALREADY_EXISTS == GetLastError())
            {
                sprintf_s(szPathMft0Start, 128 , "%s\\start\\fb_start.log", MFT0_NOTIFY_TAG_FOLDER);
                std::fstream inputFileStream(szPathMft0Start, std::ios::out);
                inputFileStream.close();
                PLOG("====== CMft0::ProcessMessage(), Create start notifying file path=[%s]", szPathMft0Start);
            }
        }
        break;
    }

END:
    LeaveCriticalSection(&m_critSec);
    PLOG("<<<=== CMft0::ProcessMessage() \n");
    return hr;
}


// check if Intel graphics driver supports FB and MJPG decode, and registry key exist.
HRESULT CMft0::FinalConstruct()
{
    HRESULT hr = S_OK;
    BOOL bRegKeyIsExisted;
    DWORD dwDummyValue;

    PLOG(">>>=== CMft0::FinalConstruct() \n");
    m_bMft0IsEnabled = TRUE;

    bRegKeyIsExisted = GetDwordRegKey("VCAM_FB_GLOBAL_ENABLED", dwDummyValue);
    if (!bRegKeyIsExisted)
    {
        PLOG("Failed to read MFT0 registry key=[HKEY_LOCAL_MACHINE\\%s\\VCAM_FB_GLOBAL_ENABLED], disable MFT0 \n", VCAM_CONTROL_REGKEY_PATH);
        m_bMft0IsEnabled = FALSE;
        goto END;
    }

    hr = m_MjpgDecoder.Init(FOURCC_NV12, 640, 480);
    if (SUCCEEDED(hr))
    {
        m_MjpgDecoder.Destroy();
    }
    else
    {
        PLOG("Failed to m_MjpgDecoder.Init(), disable MFT0 \n");
        m_bMft0IsEnabled = FALSE;
        goto END;
    }

    hr = CreateD3d11DeviceAndContext();
    GOTO_END_FAILED(hr, "Failed to CreateD3d11DeviceAndContext() \n");

    hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, 640, 480);
    if (SUCCEEDED(hr))
    {
        m_FdfbFilter.Destroy();
    }
    else
    {
        PLOG("Failed to m_FdfbFilter.Init(), disable MFT0 \n");
        m_bMft0IsEnabled = FALSE;
        goto END;
    }

END:
    hr = S_OK;
    PLOG("<<<=== CMft0::FinalConstruct() hr=[%d] \n", hr);
    return hr;
}


VOID CMft0::FinalRelease()
{
    HRESULT hr = S_OK;
    PLOG(">>>=== FinalRelease() \n");

    PLOG("<<<=== FinalRelease() hr=[%d] \n", hr);
}


STDMETHODIMP CMft0::ProcessInput( 
    DWORD dwInputStreamID,
    IMFSample *pSample,
    DWORD dwFlags)
{
    HRESULT hr = S_OK;
    DWORD   dwBufferCount = 0;

    EnterCriticalSection(&m_critSec);

    do {
        CHK_NULL_BRK(pSample);

        CHK_BOOL_BRK(dwFlags == 0);

        if (!IsValidInputStream(dwInputStreamID))
        {
            CHK_LOG_BRK(hr = MF_E_INVALIDSTREAMNUMBER);
        }

        if (!m_pInputType || !m_pOutputType)
        {
            CHK_LOG_BRK(MF_E_NOTACCEPTING);   // Client must set input and output types.
        }

        if (m_pSample != NULL)
        {
            CHK_LOG_BRK(MF_E_NOTACCEPTING);   // We already have an input sample.
        }

        // Validate the number of buffers. There should only be a single buffer to hold the video frame.
        CHK_LOG_BRK(pSample->GetBufferCount(&dwBufferCount));

        if (dwBufferCount == 0)
        {
            CHK_LOG_BRK(E_FAIL);
        }
        if (dwBufferCount > 1)
        {
            CHK_LOG_BRK(MF_E_SAMPLE_HAS_TOO_MANY_BUFFERS);
        }

        // Cache the sample. We do the actual work in ProcessOutput.
        m_pSample = pSample;
        pSample->AddRef();  // Hold a reference count on the sample.
    } while (FALSE);

    LeaveCriticalSection(&m_critSec);
    return hr;
}


HRESULT DebugDumpMfMediaBuf(IMFMediaBuffer *pMediaBufIn, char *szDumpPathPrefix, IMFMediaType *pMediaType)
{
    HRESULT     hr             = S_OK;
    LONG        lDefaultStride = 0;
    LONG        lSrcStride     = 0;
    UINT        uiWidth        = 0;
    UINT        uiHeight       = 0;
    BYTE        *pSrc          = NULL;
    GUID        stSubType      = {0};

    PLOG_MEDIATYPE(pMediaType, ">>>=== DebugDumpMfMediaBuf() ");

    if (!pMediaType || !pMediaBufIn) 
    {
        hr = E_FAIL;
        PLOG("Failed to DebugDumpMfMediaBuf() due to null pointer pMediaType or pMediaBufIn \n");
        goto END;
    }

    MFGetAttributeSize(pMediaType, MF_MT_FRAME_SIZE, &uiWidth, &uiHeight);
    pMediaType->GetGUID(MF_MT_SUBTYPE, &stSubType);

    if (stSubType == MFVideoFormat_MJPG)
    {
        PLOG("====== DebugDumpMfMediaBuf() stSubType=[MFVideoFormat_MJPG] \n");
        DWORD dwMaxLen;
        DWORD dwCurrentLen;

        hr = pMediaBufIn->Lock(&pSrc, &dwMaxLen, &dwCurrentLen);
        GOTO_END_FAILED(hr, "Failed pMediaBufInput() \n");

        DUMP_BUFFER(szDumpPathPrefix, stSubType, pSrc, dwCurrentLen, uiWidth, uiHeight);

        hr = pMediaBufIn->Unlock();
        GOTO_END_FAILED(hr, "Failed pMediaBufInput->Unlock() \n");
    }
    else
    {
        VideoBufferLock bufferLock(pMediaBufIn);

        hr = GetDefaultStride(pMediaType, &lDefaultStride);
        GOTO_END_FAILED(hr, "Failed toGetDefaultStride() \n");

        hr = bufferLock.LockBuffer(lDefaultStride, uiHeight, &pSrc, &lSrcStride);
        GOTO_END_FAILED(hr, "Failed toGetDefaultStride() \n");

        PLOG("====== DebugDumpMfMediaBuf() stSubType=[MFVideoFormat_NV12] lSrcStride=[%d] uiHeight=[%d] \n", lSrcStride, uiHeight);

        if (stSubType == MFVideoFormat_NV12 || stSubType == MFVideoFormat_I420) {
            DUMP_BUFFER(szDumpPathPrefix, stSubType, pSrc, lSrcStride*uiHeight*3/2, uiWidth, uiHeight);
        } else if (stSubType == MFVideoFormat_YUY2) {
            DUMP_BUFFER(szDumpPathPrefix, stSubType, pSrc, lSrcStride*uiHeight*2, uiWidth, uiHeight);
        } else {
            PLOG("Warning, not support color format stSubType=[%d]", stSubType);
            DUMP_BUFFER(szDumpPathPrefix, stSubType, pSrc, lSrcStride*uiHeight, uiWidth, uiHeight);
        }
    }

END:
    PLOG("<<<=== DebugDumpMfMediaBuf() \n");
    return hr;
}


STDMETHODIMP CMft0::ProcessOutput(
    DWORD dwFlags,
    DWORD cOutputBufferCount,
    /* [size_is][out][in] */ MFT_OUTPUT_DATA_BUFFER *pOutputSamples,
    /* [out] */ DWORD *pdwStatus)
{
    HRESULT        hr = S_OK;
    IMFMediaBuffer *pMediaBufInput   = NULL;
    IMFMediaBuffer *pMediaBufOutput  = NULL;
    IMFSample      *pOutputIMFSample = NULL;
    LONGLONG       hnsDuration       = 0;
    LONGLONG       hnsTime           = 0;
    BOOL           bRegKeyHasBeenChanged;

    LONG lClockStart = clock();
    LONG lClockEnd = 0;

    GUID InputSubtype;
    GUID OutputSubtype;

    if (m_bFirstRun)
    {
        PLOG(">>>=== CMft0::ProcessOutput() \n");
        PLOG_MEDIATYPE(m_pInputType, "====== CMft0::ProcessOutput() m_pInputType= ");
        PLOG_MEDIATYPE(m_pOutputType, "====== CMft0::ProcessOutput() m_pOutputType= ");
    }

    EnterCriticalSection(&m_critSec);

    GOTO_END_FALSE(dwFlags == 0, "Failed to ProcessOutput() due to invalid Flags \n");
    GOTO_END_NULL(pOutputSamples, "Failed to ProcessOutput() due to null pointer pOutputSamples \n");
    GOTO_END_NULL(pdwStatus, "Faied to ProcessOutput() due to null pointer pdwStatus \n");
    GOTO_END_FALSE(cOutputBufferCount == 1, "Failed to ProcessOutput() due to only support one output buffer \n");

    if (m_pSample == NULL)
    {
        // PLOG("Failed to CMft0::ProcessOutput() MF_E_TRANSFORM_NEED_MORE_INPUT \n");
        hr = MF_E_TRANSFORM_NEED_MORE_INPUT;
        goto END;
    }

    hr = m_pSample->ConvertToContiguousBuffer(&pMediaBufInput);
    GOTO_END_FAILED(hr, "Failed to ConvertToContiguousBuffer() \n");
    if (m_bFirstRun)
    {
        DebugDumpMfMediaBuf(pMediaBufInput, "C:\\Temp\\mft0_all_in", m_pInputType);
    }

    hr = MFCreateSample(&pOutputIMFSample);
    GOTO_END_FAILED(hr, "Failed to MFCreateSample() \n");

    if (pOutputSamples[0].pSample)
    {
        hr = pOutputSamples[0].pSample->ConvertToContiguousBuffer(&pMediaBufOutput);
        GOTO_END_FAILED(hr, "Failed to ConvertToContiguousBuffer() \n");
    }
    else
    {
        hr = m_pSample->ConvertToContiguousBuffer(&pMediaBufOutput);
        GOTO_END_FAILED(hr, "Failed to ConvertToContiguousBuffer() \n");
    }

    hr = pOutputIMFSample->AddBuffer(pMediaBufOutput);
    GOTO_END_FAILED(hr, "Failed to pOutputIMFSample->AddBuffer() \n");

    if (m_pRegkeyListener->CheckIfChanged())
    {
        PLOG("Registry key has been changed \n");
        bRegKeyHasBeenChanged = TRUE;

        if (!GetDwordRegKey("VCAM_FB_GLOBAL_ENABLED", m_dwRegKeyFbGlobalEnabled))
        {
            m_dwRegKeyFbGlobalEnabled = FALSE;
        }

        if (!GetDwordRegKey("VCAM_FB_DETECTION_ANGLE", m_dwRegKeyFbDetectionAngleCurrent))
        {
            m_dwRegKeyFbDetectionAngleCurrent = 0;
            m_dwRegKeyFbDetectionAnglePrevious = 0;
        }

        m_pRegkeyListener->Reset();
        hr = m_pRegkeyListener->Listen();
        GOTO_END_FAILED(hr, "Failed to m_pRegkeyListener->Listen() \n");
    }
    else
    {
        bRegKeyHasBeenChanged = FALSE;
    }

    hr = m_pInputType->GetGUID(MF_MT_SUBTYPE, &InputSubtype);
    GOTO_END_FAILED(hr, "Failed to GetGUID() \n");

    hr = m_pOutputType->GetGUID(MF_MT_SUBTYPE, &OutputSubtype);
    GOTO_END_FAILED(hr, "Failed to GetGUID() \n");

    if (m_FdfbFilterState == stInitialized && m_dwRegKeyFbGlobalEnabled)
    {
        static BYTE pBufferSrcNv12[CSC_BUF_SIZE_MAX] = {0};
        static BYTE pBufferRotationOut[CSC_BUF_SIZE_MAX] = {0};
        static BYTE pBufferFaceBeautyOut[CSC_BUF_SIZE_MAX] = {0};

        if (m_dwRegKeyFbDetectionAngleCurrent != m_dwRegKeyFbDetectionAnglePrevious)
        {
            m_dwRegKeyFbDetectionAnglePrevious = m_dwRegKeyFbDetectionAngleCurrent;
            m_FdfbFilter.Destroy();
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
            m_RotationCropFilter.Destroy();
            m_RotationFillFilter.Destroy();
#endif

            if (m_dwRegKeyFbDetectionAngleCurrent == 90 || m_dwRegKeyFbDetectionAngleCurrent == 270)
            {
                INT iTargetW = m_InputTypeHeight;
                INT iTargetH = (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)? HOT_FIX_FB_HEIGHT_MAX: m_InputTypeWidth;
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
                hr = m_RotationCropFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight, iTargetW, iTargetH);
                GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.Init() \n");

                hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, iTargetW, iTargetH);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");

                hr = m_RotationFillFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, iTargetW, iTargetH, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.Init() \n");

#else // HOT_FIX_FB_HEIGHT_LIMIT_SW
                hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, iTargetW, iTargetH);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");
#endif
            }
            else
            {
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
                hr = m_RotationCropFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.Init() \n");

                hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");

                hr = m_RotationFillFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.Init() \n");
#else
                hr = m_FdfbFilter.Init(m_pD3d11Device, m_pD3d11DeviceContext, FDFB_OUTPUT_FORMAT, FDFB_INPUT_FORMAT, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");
#endif
            }
            GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Init() \n");
        }

        if (InputSubtype == MFVideoFormat_MJPG && OutputSubtype == MFVideoFormat_NV12)
        {
            BYTE *pSrc = NULL;
            BYTE *pDst = NULL;
            DWORD dwCurrentLen, dwMaxLen;
            LONG lDefaultStride;
            LONG lDestStride;
            static BYTE pBufferCropOut[CSC_BUF_SIZE_MAX] = {0};
            static BYTE pBufferFillOut[CSC_BUF_SIZE_MAX] = {0};

            hr = pMediaBufInput->Lock(&pSrc, &dwMaxLen, &dwCurrentLen);
            GOTO_END_FAILED(hr, "Failed pMediaBufInput() \n");

            hr = m_MjpgDecoder.Decode(pSrc, dwCurrentLen, pBufferSrcNv12);
#if WA_SKIP_MJPEG_DECODE_ERROR
            if (FAILED(hr))
            {
                PLOG("Warning: Skip failed m_MjpgDecoder.Decode() hr=[%d] \n", hr);
                hr = S_OK;
            }
#endif
            GOTO_END_FAILED(hr, "Failed m_MjpgDecoder.Decode() \n");

            hr = pMediaBufInput->Unlock();
            GOTO_END_FAILED(hr, "Failed pMediaBufInput->Unlock() \n");

            VideoBufferLock outputLock(pMediaBufOutput);
            hr = GetDefaultStride(m_pOutputType, &lDefaultStride);
            GOTO_END_FAILED(hr, "Failed toGetDefaultStride() \n");

            hr = outputLock.LockBuffer(lDefaultStride, m_InputTypeHeight, &pDst, &lDestStride);
            GOTO_END_FAILED(hr, "Failed outputLock.LockBuffer() \n");

            if (m_dwRegKeyFbDetectionAngleCurrent == 90)
            {
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
                INT iTargetW = m_InputTypeHeight;
                INT iTargetH = (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)? HOT_FIX_FB_HEIGHT_MAX: m_InputTypeWidth;

                hr = m_RotationCropFilter.BltBuf(pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight, D3D11_VIDEO_PROCESSOR_ROTATION_270);
                GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.BltBuf() \n");

                hr = m_FdfbFilter.BltText(m_RotationCropFilter.GetBltTextOut(), bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.BltText() \n");

                m_RotationFillFilter.BltText(m_FdfbFilter.GetBltTextOut(), D3D11_VIDEO_PROCESSOR_ROTATION_90);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.BltText() \n");

                m_RotationFillFilter.GetBltBufOut(pDst, CSC_BUF_SIZE_MAX);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.BltText() \n");

#else // HOT_FIX_FB_HEIGHT_LIMIT_SW
                hr = RotateNv12SurfaceDegree270(pBufferRotationOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree270() \n");

                if (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)
                {
                    CropSurfaceTopBottomNv12(pBufferCropOut, pBufferRotationOut, m_InputTypeHeight, m_InputTypeWidth, HOT_FIX_FB_HEIGHT_MAX);

                    hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferCropOut, m_InputTypeHeight, HOT_FIX_FB_HEIGHT_MAX, bRegKeyHasBeenChanged);
                    GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                    CopyAndFillTopBottomSurfNv12(pBufferFillOut, m_InputTypeWidth, pBufferFaceBeautyOut, m_InputTypeHeight, HOT_FIX_FB_HEIGHT_MAX);

                    hr = RotateNv12SurfaceDegree90(pDst, pBufferFillOut, m_InputTypeHeight, m_InputTypeWidth);
                    GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree90() \n");
                }
                else
                {
                    hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferRotationOut, m_InputTypeHeight, m_InputTypeWidth, bRegKeyHasBeenChanged);
                    GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                    hr = RotateNv12SurfaceDegree90(pDst, pBufferFaceBeautyOut, m_InputTypeHeight, m_InputTypeWidth);
                    GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree90() \n");
                }
#endif
            }
            else if (m_dwRegKeyFbDetectionAngleCurrent == 180)
            {
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
                hr = m_RotationCropFilter.BltBuf(pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight, D3D11_VIDEO_PROCESSOR_ROTATION_180);
                GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.BltBuf() \n");

                hr = m_FdfbFilter.BltText(m_RotationCropFilter.GetBltTextOut(), bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.BltText() \n");

                hr = m_RotationFillFilter.BltText(m_FdfbFilter.GetBltTextOut(), D3D11_VIDEO_PROCESSOR_ROTATION_180);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.BltText() \n");

                hr = m_RotationFillFilter.GetBltBufOut(pDst, CSC_BUF_SIZE_MAX);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.GetBltBufOut() \n");

#else // HOT_FIX_FB_HEIGHT_LIMIT_SW
                hr = RotateNv12SurfaceDegree180(pBufferRotationOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree180() \n");

                hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferRotationOut, m_InputTypeWidth, m_InputTypeHeight, bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                hr = RotateNv12SurfaceDegree180(pDst, pBufferFaceBeautyOut, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree180() \n");
#endif
            }
            else if (m_dwRegKeyFbDetectionAngleCurrent == 270)
            {
#if HOT_FIX_FB_HEIGHT_LIMIT_HW
                INT iTargetW = m_InputTypeHeight;
                INT iTargetH = (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)? HOT_FIX_FB_HEIGHT_MAX: m_InputTypeWidth;

                m_RotationCropFilter.BltBuf(pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight, D3D11_VIDEO_PROCESSOR_ROTATION_90);
                GOTO_END_FAILED(hr, "Failed to m_RotationCropFilter.BltBuf() \n");

                hr = m_FdfbFilter.BltText(m_RotationCropFilter.GetBltTextOut(), bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.BltText() \n");

                hr = m_RotationFillFilter.BltText(m_FdfbFilter.GetBltTextOut(), D3D11_VIDEO_PROCESSOR_ROTATION_270);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.BltText() \n");

                hr = m_RotationFillFilter.GetBltBufOut(pDst, CSC_BUF_SIZE_MAX);
                GOTO_END_FAILED(hr, "Failed to m_RotationFillFilter.GetBltBufOut() \n");

#else // HOT_FIX_FB_HEIGHT_LIMIT_SW
                hr = RotateNv12SurfaceDegree90(pBufferRotationOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree90() \n");

                if (m_InputTypeWidth > HOT_FIX_FB_HEIGHT_MAX)
                {
                    CropSurfaceTopBottomNv12(pBufferCropOut, pBufferRotationOut, m_InputTypeHeight, m_InputTypeWidth, HOT_FIX_FB_HEIGHT_MAX);

                    hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferCropOut, m_InputTypeHeight, HOT_FIX_FB_HEIGHT_MAX, bRegKeyHasBeenChanged);
                    GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                    CopyAndFillTopBottomSurfNv12(pBufferFillOut, m_InputTypeWidth, pBufferFaceBeautyOut, m_InputTypeHeight, HOT_FIX_FB_HEIGHT_MAX);
                    hr = RotateNv12SurfaceDegree270(pDst, pBufferFillOut, m_InputTypeHeight, m_InputTypeWidth);
                    GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree270() \n");
                }
                else
                {
                    hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferRotationOut, m_InputTypeHeight, m_InputTypeWidth, bRegKeyHasBeenChanged);
                    GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                    hr = RotateNv12SurfaceDegree270(pDst, pBufferFaceBeautyOut, m_InputTypeHeight, m_InputTypeWidth);
                    GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree270() \n");
                }
#endif
            }
            else // for m_dwRegKeyFbDetectionAngle == 0 or others
            {
                hr = m_FdfbFilter.Blt(pDst, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight, bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");
            }
        }
        // Samsung's cameras go through MJPG->NV12, so it's used for camera only supports YUY2->YUY2
        else if (InputSubtype == MFVideoFormat_YUY2 && OutputSubtype == MFVideoFormat_YUY2)
        {
            BYTE *pSrc = NULL;
            BYTE *pDst = NULL;
            LONG lDefaultStride;
            LONG lDestStride;

            VideoBufferLock inputLock(pMediaBufInput);
            VideoBufferLock outputLock(pMediaBufOutput);

            hr = GetDefaultStride(m_pInputType, &lDefaultStride);
            GOTO_END_FAILED(hr, "Failed toGetDefaultStride() \n");

            hr = inputLock.LockBuffer(lDefaultStride, m_InputTypeHeight, &pSrc, &lDestStride);
            GOTO_END_FAILED(hr, "Failed inputLock.LockBuffer() \n");

            hr = outputLock.LockBuffer(lDefaultStride, m_OutputTypeHeight, &pDst, &lDestStride);
            GOTO_END_FAILED(hr, "Failed outputLock.LockBuffer() \n");

            hr = m_CscFilter_Yuy2ToNv12.Blt(pBufferSrcNv12, pSrc, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to m_CscFilter_Yuy2ToNv12.Blt() \n");

            if (m_dwRegKeyFbDetectionAngleCurrent == 90)
            {
                hr = RotateNv12SurfaceDegree270(pBufferRotationOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree270() \n");

                hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferRotationOut, m_InputTypeHeight, m_InputTypeWidth, bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                hr = RotateNv12SurfaceDegree90(pBufferFaceBeautyOut, pBufferFaceBeautyOut, m_InputTypeHeight, m_InputTypeWidth);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree90() \n");
            }
            else if (m_dwRegKeyFbDetectionAngleCurrent == 180)
            {
                hr = RotateNv12SurfaceDegree180(pBufferRotationOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree180() \n");

                hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferRotationOut, m_InputTypeWidth, m_InputTypeHeight, bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                hr = RotateNv12SurfaceDegree180(pBufferFaceBeautyOut, pBufferFaceBeautyOut, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree180() \n");
            }
            else if (m_dwRegKeyFbDetectionAngleCurrent == 270)
            {
                hr = RotateNv12SurfaceDegree90(pBufferRotationOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree90() \n");

                hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferRotationOut, m_InputTypeHeight, m_InputTypeWidth, bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");

                hr = RotateNv12SurfaceDegree270(pBufferFaceBeautyOut, pBufferFaceBeautyOut, m_InputTypeHeight, m_InputTypeWidth);
                GOTO_END_FAILED(hr, "Failed to RotateNv12SurfaceDegree270() \n");
            }
            else // for m_dwRegKeyFbDetectionAngle == 0 or others
            {
                hr = m_FdfbFilter.Blt(pBufferFaceBeautyOut, pBufferSrcNv12, m_InputTypeWidth, m_InputTypeHeight, bRegKeyHasBeenChanged);
                GOTO_END_FAILED(hr, "Failed to m_FdfbFilter.Blt() \n");
            }

            hr = Nv12_to_Yuy2(pDst, lDefaultStride*m_InputTypeHeight*2, pBufferFaceBeautyOut, m_InputTypeWidth, m_InputTypeHeight);
            GOTO_END_FAILED(hr, "Failed to Nv12_to_Yuy2() \n");
        }
        else
        {
            if (m_uiPerformanceCounter % 300)
            {
                PLOG("====== CMft0::ProcessOutput() Not apply FB effect due to unsupported color format \n");
            }
        }
    }
    else
    {
        GUID guidInputSubType = {0};
        UINT uiWidth, uiHeight, uiFps;

        if (InputSubtype == MFVideoFormat_MJPG && OutputSubtype == MFVideoFormat_NV12)
        {
            BYTE *pSrc = NULL;
            BYTE *pDst = NULL;
            DWORD dwCurrentLen, dwMaxLen;
            LONG lDefaultStride;
            LONG lDestStride;

            VideoBufferLock outputLock(pMediaBufOutput);
            hr = GetDefaultStride(m_pOutputType, &lDefaultStride);
            GOTO_END_FAILED(hr, "Failed toGetDefaultStride() \n");

            hr = GetFormatWidthHeightFps(m_pInputType, &guidInputSubType, &uiWidth, &uiHeight, &uiFps);
            GOTO_END_FAILED(hr, "Failed GetFormatWidthHeightFps() \n");

            hr = outputLock.LockBuffer(lDefaultStride, uiHeight, &pDst, &lDestStride);
            GOTO_END_FAILED(hr, "Failed outputLock.LockBuffer() \n");

            hr = pMediaBufInput->Lock(&pSrc, &dwMaxLen, &dwCurrentLen);
            GOTO_END_FAILED(hr, "Failed pMediaBufInput() \n");

            hr = m_MjpgDecoder.Decode(pSrc, dwCurrentLen, pDst);
#if WA_SKIP_MJPEG_DECODE_ERROR
            if (FAILED(hr))
            {
                PLOG("Warning: Skip failed m_MjpgDecoder.Decode() hr=[%d] \n", hr);
                hr = S_OK;
            }
#endif
            GOTO_END_FAILED(hr, "Failed m_MjpgDecoder.Decode() \n");

            hr = pMediaBufInput->Unlock();
            GOTO_END_FAILED(hr, "Failed pMediaBufInput->Unlock() \n");
        }
        else
        {
            // do nothing, output is input buffer
        }
    }

    pOutputSamples[0].pSample  = pOutputIMFSample;
    pOutputSamples[0].dwStatus = 0;
    *pdwStatus                 = 0;
    pOutputIMFSample->AddRef();

    hr = m_pSample->GetSampleDuration(&hnsDuration);
    if (SUCCEEDED(hr))
    {
        hr = pOutputSamples[0].pSample->SetSampleDuration(hnsDuration);
        GOTO_END_FAILED(hr, "Failed to pSample->SetSampleDuration() \n");
    }

    hr = m_pSample->GetSampleTime(&hnsTime);
    if (SUCCEEDED(hr))
    {
        hr = pOutputSamples[0].pSample->SetSampleTime(hnsTime);
        GOTO_END_FAILED(hr, "Failed to pSample->SetSampleTime() \n");
    }

END:
    if (m_bFirstRun)
    {
        PLOG("<<<=== CMft0::ProcessOutput() hr=[%d] \n", hr);
    }

    if (m_bFirstRun && SUCCEEDED(hr))
    {
        m_bFirstRun = FALSE;
        DebugDumpMfMediaBuf(pMediaBufOutput, "C:\\Temp\\mft0_all_out", m_pOutputType);
    }

    lClockEnd = clock();
    m_uiPerformanceCounter++;
    m_lPerformanceClockSum += (lClockEnd - lClockStart);
    if (m_uiPerformanceCounter % 300 == 0)
    {
        PLOG("performance evaluation used ticks in [%07d]-times ProcessOutput() = [%08d] CLOCKS_PER_SEC=[%d] \n", m_uiPerformanceCounter, m_lPerformanceClockSum, CLOCKS_PER_SEC);
    }

    SAFERELEASE(pOutputIMFSample);
    SAFERELEASE(pMediaBufInput);
    SAFERELEASE(pMediaBufOutput);
    SAFERELEASE(m_pSample);   // Release our input sample.
    LeaveCriticalSection(&m_critSec);
    return hr;
}

BOOL CMft0::IsValidInputStream(DWORD dwInputStreamID)
{
    return dwInputStreamID == 0;
}

// IsValidOutputStream: Returns TRUE if dwOutputStreamID is a valid output stream identifier.
BOOL CMft0::IsValidOutputStream(DWORD dwOutputStreamID)
{
    //update
    return dwOutputStreamID == 0;
}


STDMETHODIMP CMft0::OnFlush()
{
    HRESULT hr = S_OK;

    // For this MFT, flushing just means releasing the input sample.
    SAFERELEASE(m_pSample);

    return hr;
}

STDMETHODIMP CMft0::GetMediaType( DWORD dwStreamId, DWORD dwTypeIndex, IMFMediaType **ppType)
{
    HRESULT hr = S_OK;

    do {
        CHK_NULL_PTR_BRK(ppType);
        if (dwStreamId != 0)
            CHK_LOG_BRK(MF_E_INVALIDSTREAMNUMBER);
        if (dwTypeIndex >= m_listOfMediaTypesIn.GetCount())
            CHK_LOG_BRK(MF_E_NO_MORE_TYPES);

        *ppType = m_listOfMediaTypesIn[dwTypeIndex];
        (*ppType)->AddRef();
    } while (FALSE);

    return hr;
}

STDMETHODIMP CMft0::GetMediaTypeOut(DWORD dwStreamId, DWORD dwTypeIndex, IMFMediaType **ppType)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::GetMediaTypeOut() dwStreamId=[%d] dwTypeIndex=[%d] \n", dwStreamId, dwTypeIndex);

    GOTO_END_NULL(ppType, "Failed to CMft0::GetMediaTypeOut() due to null pointer ppType \n");

    if (dwStreamId != 0) { hr = MF_E_INVALIDSTREAMNUMBER; goto END; }

    if (dwTypeIndex >= m_listOfMediaTypesOut.GetCount()) { hr = MF_E_NO_MORE_TYPES; goto END; }

    *ppType = m_listOfMediaTypesOut[dwTypeIndex];
    (*ppType)->AddRef();

    PLOG_MEDIATYPE(*ppType, "====== CMft0::GetMediaTypeOut() ");
END:
    if (hr == MF_E_NO_MORE_TYPES) {
        PLOG("<<<=== CMft0::GetMediaTypeOut() hr=[MF_E_NO_MORE_TYPES] \n");
    } else if (hr == MF_E_INVALIDSTREAMNUMBER) {
        PLOG("<<<=== CMft0::GetMediaTypeOut() hr=[MF_E_INVALIDSTREAMNUMBER] \n");
    } else {
        PLOG("<<<=== CMft0::GetMediaTypeOut() hr=[%d] \n", hr);
    }
    return hr;
}

STDMETHODIMP CMft0::IsMediaTypeSupported(UINT uiStreamId, IMFMediaType *pIMFMediaType, IMFMediaType **ppIMFMediaTypeFull)
{
    HRESULT hr = S_OK;

    do {
        CHK_NULL_PTR_BRK(pIMFMediaType);

        if(uiStreamId != 0)
        {
            CHK_LOG_BRK(MF_E_INVALIDINDEX);
        }
        BOOL bFound =FALSE;
        for(UINT i = 0; i< m_listOfMediaTypesIn.GetCount(); i++)
        {
            DWORD   dwResult = 0;
            hr = m_listOfMediaTypesIn[i]->IsEqual(pIMFMediaType, &dwResult);
            if(hr == S_FALSE)
            {

                if((dwResult & MF_MEDIATYPE_EQUAL_MAJOR_TYPES) && 
                    (dwResult& MF_MEDIATYPE_EQUAL_FORMAT_TYPES) && 
                    (dwResult& MF_MEDIATYPE_EQUAL_FORMAT_DATA))
                {
                    hr = S_OK;
                }
            }
            if(hr == S_OK)
            {
                bFound = TRUE;
                if(ppIMFMediaTypeFull) {
                    *ppIMFMediaTypeFull = m_listOfMediaTypesIn[i];
                    (*ppIMFMediaTypeFull)->AddRef();
                }
                break;
            }
            else if(FAILED(hr))
            {
                CHK_LOG_BRK(hr);
            }

        }
        if(bFound == FALSE)
        {
            CHK_LOG_BRK(MF_E_INVALIDMEDIATYPE);
        }
    } while (FALSE);

    return hr;
}


HRESULT CMft0::AddInputMediaTypeList(GUID guidSubtypeAdd)
{
    HRESULT hr = S_OK;
    CHAR szSubGuid[64] = {0};

    sGuidPrint(szSubGuid, 64, guidSubtypeAdd);
    PLOG(">>>=== AddInputMediaTypeList() szSubGuid=[%s] \n", szSubGuid);

    if (!m_pSourceTransform)
    {
        PLOG("Failed to CMft0::AddInputMediaTypeList() due to null pointer m_pSourceTransform \n");
        hr = E_FAIL;
        goto END;
    }

    for (UINT iMediaType = 0; hr == S_OK; iMediaType++) 
    {
        CComPtr<IMFMediaType> pMediaType;
        GUID GuidSubType;
        UINT uiWidth, uiHeight, uiFps;

        hr = m_pSourceTransform->GetOutputAvailableType(0, iMediaType, &pMediaType);
        if (hr == MF_E_NO_MORE_TYPES)
        {
            hr = S_OK; // This is an exception case, so we treat it as S_OK.
            break;
        }
        GOTO_END_FAILED(hr, "Failed to GetOutputAvailableType() \n");

        hr = GetFormatWidthHeightFps(pMediaType, &GuidSubType, &uiWidth, &uiHeight, &uiFps);
        GOTO_END_FAILED(hr, "Failed GetFormatWidthHeightFps() \n");

        if (GuidSubType == guidSubtypeAdd)
        {
            if (uiWidth <= MFT0_MAX_WIDTH && uiHeight <= MFT0_MAX_HEIGHT)
            {
                m_listOfMediaTypesIn[(ULONG)(m_listOfMediaTypesIn.GetCount())] = pMediaType;
            }
        }
    }

    PLOG("<<<=== AddInputMediaTypeList() m_listOfMediaTypesIn.GetCount()=[%d] \n", m_listOfMediaTypesIn.GetCount());
END:
    return hr;
}


HRESULT CMft0::AddOutputMediaTypeListBaseOnInput()
{
    HRESULT hr = S_OK;
    GUID GuidSubType;
    UINT uiWidth, uiHeight, uiFps;
    PLOG(">>>=== CMft0::AddOutputMediaTypeListBaseOnInput()  m_listOfMediaTypesIn.GetCount()=[%d] \n",  m_listOfMediaTypesIn.GetCount());

    for (UINT iListInIdx = 0; iListInIdx < m_listOfMediaTypesIn.GetCount() ; iListInIdx++) 
    {
        hr = GetFormatWidthHeightFps(m_listOfMediaTypesIn[iListInIdx], &GuidSubType, &uiWidth, &uiHeight, &uiFps);
        GOTO_END_FAILED(hr, "Failed GetFormatWidthHeightFps() \n");

        for (INT iIdx = 0; iIdx < MFT0_AVAILABLE_FORMAT_TABLE_LEN; iIdx++)
        {
            if (m_pMft0AvailableFormatTable[iIdx].InputGuid == GuidSubType)
            {
                CComPtr<IMFMediaType> pMediaTypeOut;

                hr = MFCreateMediaType(&pMediaTypeOut);
                GOTO_END_FAILED(hr, "Failed to MFCreateMediaType() \n");

                hr = m_listOfMediaTypesIn[iListInIdx]->CopyAllItems(pMediaTypeOut);
                GOTO_END_FAILED(hr, "Failed to CopyAllItems() \n");

                hr = pMediaTypeOut->SetGUID(MF_MT_SUBTYPE, m_pMft0AvailableFormatTable[iIdx].OutputGuid);
                GOTO_END_FAILED(hr, "Failed to SetGUID() \n");

                m_listOfMediaTypesOut[(ULONG)(m_listOfMediaTypesOut.GetCount())] = pMediaTypeOut;
            }
        }
    }

END:
    PLOG("<<< CMft0::AddOutputMediaTypeListBaseOnInput() m_listOfMediaTypesOut.GetCount()=[%d] \n", m_listOfMediaTypesOut.GetCount());
    return hr;
}


HRESULT CMft0::CheckSupportFormat()
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CMft0::CheckSupportFormat() \n");

    m_bWebcamSupportI420 = m_bWebcamSupportYuy2 = m_bWebcamSupportMjpg = FALSE;
    m_bMft0InputMjpg = m_bMft0InputYuy2 = m_bMft0InputI420 = FALSE;

    for (UINT iIdx = 0; hr == S_OK; iIdx++) 
    {
        CComPtr<IMFMediaType> pMediaType;
        UINT uiWidth, uiHeight, uiFps;
        GUID GuidSubType;

        hr = m_pSourceTransform->GetOutputAvailableType(0, iIdx, &pMediaType);
        if (hr == MF_E_NO_MORE_TYPES)
        {
            hr = S_OK; // This is an exception case, so we treat it as S_OK.
            break;
        }

        GOTO_END_FAILED(hr, "Failed to GetOutputAvailableType() \n");

        hr = GetFormatWidthHeightFps(pMediaType, &GuidSubType, &uiWidth, &uiHeight, &uiFps);
        GOTO_END_FAILED(hr, "Failed to GetFormatWidthHeightFps() \n");

        if (uiWidth == 640 && uiHeight == 480 && uiFps == 30)
        {
            if (GuidSubType == MFVideoFormat_MJPG ) { m_bWebcamSupportMjpg = TRUE; }
            if (GuidSubType == MFVideoFormat_YUY2 ) { m_bWebcamSupportYuy2 = TRUE; }
            if (GuidSubType == MFVideoFormat_I420 ) { m_bWebcamSupportI420 = TRUE; }
        }
    }

    for (INT i = 0; i < MFT0_AVAILABLE_FORMAT_TABLE_LEN; i++)
    {
        if (m_pMft0AvailableFormatTable[i].InputGuid == MFVideoFormat_MJPG) { m_bMft0InputMjpg = TRUE; }
        if (m_pMft0AvailableFormatTable[i].InputGuid == MFVideoFormat_YUY2) { m_bMft0InputYuy2 = TRUE; }
        if (m_pMft0AvailableFormatTable[i].InputGuid == MFVideoFormat_I420) { m_bMft0InputI420 = TRUE; }
    }

    hr = S_OK;

    if (!m_bWebcamSupportYuy2 && !m_bWebcamSupportI420 && !m_bWebcamSupportMjpg)
    {
        PLOG("Failed to CheckSupportFormat() due to unsupported format \n");
        hr = E_FAIL;
        goto END;
    }

    if (!m_bMft0InputMjpg && !m_bMft0InputYuy2 && !m_bMft0InputI420)
    {
        PLOG("Failed to CheckSupportFormat() due to unsupported format \n");
        hr = E_FAIL;
        goto END;
    }

END:
    PLOG("<<<=== CMft0::CheckSupportFormat() \n");
    return hr;
}


STDMETHODIMP CMft0::GenerateMFMediaTypeListFromDevice(UINT uiStreamId)
{
    HRESULT hr = S_OK;
    CComPtr<IMFMediaType> pMediaType;

    PLOG(">>>=== CMft0::GenerateMFMediaTypeListFromDevice() \n");

    if (!m_pSourceTransform)
    {
        PLOG("Failed to GenerateMFMediaTypeListFromDevice() due to null ponter m_pSourceTransform \n");
        goto END;
    }

    m_listOfMediaTypesIn.RemoveAll();
    m_listOfMediaTypesOut.RemoveAll();

    hr = CheckSupportFormat();
    GOTO_END_FAILED(hr, "Failed to CheckSupportFormat() \n");

    if (m_bWebcamSupportMjpg && m_bMft0InputMjpg)
    {
        hr = AddInputMediaTypeList(MFVideoFormat_MJPG);
    }
    else if (m_bWebcamSupportYuy2 && m_bMft0InputYuy2)
    {
        hr = AddInputMediaTypeList(MFVideoFormat_YUY2);
    }
    else if (m_bWebcamSupportI420 && m_bMft0InputI420)
    {
        hr = AddInputMediaTypeList(MFVideoFormat_I420);
    }
    else
    {
        PLOG("Failed to GenerateMFMediaTypeListFromDevice() due to supported format \n");
        hr = E_FAIL;
        goto END;
    }

    hr = AddOutputMediaTypeListBaseOnInput();
    GOTO_END_FAILED(hr, "Failed to AddOutputMediaTypeListBaseOnInput() \n");

    for (INT iIdx = 0; iIdx < m_listOfMediaTypesIn.GetCount(); iIdx++)
    {
        PLOG_MEDIATYPE(m_listOfMediaTypesIn[iIdx], "====== CMft0::GenerateMFMediaTypeListFromDevice() m_listOfMediaTypesIn[%d]: ", iIdx);
    }

    for (INT iIdx = 0; iIdx < m_listOfMediaTypesOut.GetCount(); iIdx++)
    {
        PLOG_MEDIATYPE(m_listOfMediaTypesOut[iIdx], "====== CMft0::GenerateMFMediaTypeListFromDevice() m_listOfMediaTypesOut[%d]: ", iIdx);
    }
END:
    PLOG("<<<=== CMft0::GenerateMFMediaTypeListFromDevice() \n");
    return hr;
}


HRESULT CMft0::CreateD3d11DeviceAndContext()
{
    HRESULT hr = S_OK;

    D3D_FEATURE_LEVEL __FeatureLevel;
    D3D_FEATURE_LEVEL pFeatureLevels[] = { D3D_FEATURE_LEVEL_11_0 };
    INT flCount = sizeof(pFeatureLevels) / sizeof(D3D_FEATURE_LEVEL);
    IDXGIFactory* dxgiFactory   = NULL;
    UINT          adapterIdx    = 0;
    IDXGIAdapter* pAdapter      = NULL;
    IDXGIAdapter* pIntelAdapter = NULL;
    D3D_DRIVER_TYPE eDriverType = D3D_DRIVER_TYPE_HARDWARE;

    PLOG(">>>=== CMft0::CreateDevice() \n");

    CreateDXGIFactory(__uuidof(IDXGIFactory), reinterpret_cast<void**>(&dxgiFactory));
    GOTO_END_FAILED(hr, "Failed to CreateDXGIFactory() \n");
    GOTO_END_NULL(dxgiFactory, "Failed to CreateDXGIFactory() \n");

    for (adapterIdx = 0; dxgiFactory->EnumAdapters(adapterIdx, &pAdapter) != DXGI_ERROR_NOT_FOUND; adapterIdx++)
    {
        DXGI_ADAPTER_DESC adapterDesc;
        hr = pAdapter->GetDesc(&adapterDesc);
        if (FAILED(hr))
        {
            PLOG("Failed to pAdapter->GetDesc() \n");
            pAdapter->Release();
            pAdapter = NULL;
            continue;
        }

        // eDriverType must be D3D_DRIVER_TYPE_UNKNOWN if pIntelAdapter isn't NULL
        // https://msdn.microsoft.com/en-us/library/windows/desktop/ff476082(v=vs.85).aspx
        if (wcsstr(adapterDesc.Description, L"Intel"))
        {
            pIntelAdapter = pAdapter;
            eDriverType   = D3D_DRIVER_TYPE_UNKNOWN;
            break;
        }
        else
        {
            pAdapter->Release();
            pAdapter = NULL;
        }
    }

    if (pIntelAdapter)
    {
        PLOG("====== CMft0::CreateDevice() Found Intel graphic adapter \n");
    } else {
        PLOG("====== Warning: cannot find Intel graphic adapter, D3D11CreateDevice() with NULL pIntelAdapter \n");
    }

    hr = D3D11CreateDevice(pIntelAdapter, eDriverType, NULL, D3D11_CREATE_DEVICE_BGRA_SUPPORT,
        pFeatureLevels, flCount, D3D11_SDK_VERSION, &m_pD3d11Device, &__FeatureLevel, &m_pD3d11DeviceContext);
    GOTO_END_FAILED(hr, "Failed toD3D11CreateDevice() \n");

END:
    if (pIntelAdapter)
    {
        pIntelAdapter ->Release();
        pIntelAdapter = NULL;
    }

    if (dxgiFactory)
    {
        dxgiFactory->Release();
        dxgiFactory = NULL;
    }

    PLOG("<<<=== CMft0::CreateDevice() hr=[%08x] \n", hr);
    return hr;
}

