// Mft0.h : Declaration of the CMft0
#ifndef _MFT0_H_
#define _MFT0_H_
#pragma once
#include "resource.h"       // main symbols
#include "SampleMft0.h"
#include "SampleHelpers.h"
#if defined(_WIN32_WCE) && !defined(_CE_DCOM) && !defined(_CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA)
#error "Single-threaded COM objects are not properly supported on Windows CE platform, such as the Windows Mobile platforms that do not include full DCOM support. Define _CE_ALLOW_SINGLE_THREADED_OBJECTS_IN_MTA to force ATL to support creating single-thread COM object's and allow use of it's single-threaded COM object implementations. The threading model in your rgs file was set to 'Free' as that is the only threading model supported in non DCOM Windows CE platforms."
#endif
#include "fdfb/fdfb.h"
#include "utils_mft.h"
//using namespace ATL;
#include "decode_mjpg\decode_mjpg.h"
#include "rotation\rotation.h"
#include "csc\csc.h"

typedef struct __MFT0_AVAILABLE_FORMAT
{
    GUID InputGuid;
    GUID OutputGuid;
    INT iFps;
} MFT0_AVAILABLE_FORMAT;


enum FDFBFilterState
{
    stUnInit,
    stInitialized,
    stInitFailed,
};

// CMft0
class ATL_NO_VTABLE CMft0 :
    public CComObjectRootEx<CComMultiThreadModel>,
    public CComCoClass<CMft0, &CLSID_Mft0>,
    public IMft0, 
    public IMFTransform,
    public IInspectable
{
public:
    CMft0() :   m_pSample(NULL),
        m_pInputType(NULL),
        m_pOutputType(NULL),
        m_pInputAttributes(0),
        m_pGlobalAttributes(0),
        m_pSourceTransform(0),
        m_bEnableEffects(TRUE),
        m_uiStreamId(0),
        m_percentOfScreen(5),
        m_nRefCount(1),
        m_FdfbFilterState(stUnInit),
        m_bFirstRun(TRUE),
        m_uiPerformanceCounter(0),
        m_lPerformanceClockSum(0),
        m_bMft0IsEnabled(TRUE),
        m_pD3d11Device(NULL),
        m_pD3d11DeviceContext(NULL)
    {
        PLOG(">>>=== CMft0()::CMft0() Intel MFT0 version=[0.%d.0.0] \n", MFT0_MINOR_VERSION);
        InitializeCriticalSection(&m_critSec);
        PLOG("<<<=== CMft0()::CMft0() \n");
    }

    DECLARE_REGISTRY_RESOURCEID(IDR_MFT0)

    DECLARE_NOT_AGGREGATABLE(CMft0)

    BEGIN_COM_MAP(CMft0)
        COM_INTERFACE_ENTRY(IMft0)
        COM_INTERFACE_ENTRY(IMFTransform)
        COM_INTERFACE_ENTRY(IInspectable)
    END_COM_MAP()

    DECLARE_PROTECT_FINAL_CONSTRUCT()

    HRESULT FinalConstruct();
    void FinalRelease();

    HRESULT FindMediaType(IMFMediaType **ppMediaTypeFound, GUID guidSubTypeWanted, UINT uiWidth, UINT uiHeight, UINT uiFps);
    HRESULT CreateD3d11DeviceAndContext();

public:
    STDMETHOD(UpdateDsp)(UINT32 uiPercentOfScreen);
    STDMETHOD(Enable)(void);
    STDMETHOD(Disable)(void);
    STDMETHOD(GetDspSetting)(UINT* puiPercentOfScreen, BOOL* pIsEnabled);

    STDMETHODIMP GetIids( 
        /* [out] */ _Out_ ULONG *iidCount,
        /* [size_is][size_is][out] */ _Outptr_result_buffer_maybenull_(*iidCount) IID **iids);

    STDMETHODIMP GetRuntimeClassName( 
        /* [out] */ _Outptr_result_maybenull_ HSTRING *className);

    STDMETHODIMP GetTrustLevel( 
        /* [out] */ _Out_ TrustLevel *trustLevel);

    STDMETHODIMP GetStreamLimits( 
        /* [out] */ _Out_ DWORD *pdwInputMinimum,
        /* [out] */ _Out_ DWORD *pdwInputMaximum,
        /* [out] */ _Out_ DWORD *pdwOutputMinimum,
        /* [out] */ _Out_ DWORD *pdwOutputMaximum);

    STDMETHODIMP GetStreamCount( 
        /* [out] */ _Out_ DWORD *pcInputStreams,
        /* [out] */ _Out_ DWORD *pcOutputStreams);

    STDMETHODIMP GetStreamIDs( 
        DWORD dwInputIDArraySize,
        /* [size_is][out] */ _Out_writes_(dwInputIDArraySize) DWORD *pdwInputIDs,
        DWORD dwOutputIDArraySize,
        /* [size_is][out] */ _Out_writes_(dwOutputIDArraySize) DWORD *pdwOutputIDs);

    STDMETHODIMP GetInputStreamInfo( 
        DWORD dwInputStreamID,
        /* [out] */ _Out_ MFT_INPUT_STREAM_INFO *pStreamInfo);

    STDMETHODIMP GetOutputStreamInfo( 
        DWORD dwOutputStreamID,
        /* [out] */ _Out_ MFT_OUTPUT_STREAM_INFO *pStreamInfo);

    STDMETHODIMP GetAttributes( 
        /* [out] */ _Outptr_result_maybenull_ IMFAttributes **pAttributes);

    STDMETHODIMP GetInputStreamAttributes( 
        DWORD dwInputStreamID,
        /* [out] */ _Outptr_result_maybenull_ IMFAttributes **pAttributes);

    STDMETHODIMP GetOutputStreamAttributes( 
        DWORD dwOutputStreamID,
        /* [out] */ _Outptr_result_maybenull_ IMFAttributes **pAttributes);

    STDMETHODIMP DeleteInputStream( 
        DWORD dwStreamID);

    STDMETHODIMP AddInputStreams( 
        DWORD cStreams,
        /* [in] */ _In_ DWORD *adwStreamIDs);

    STDMETHODIMP GetInputAvailableType( 
        DWORD dwInputStreamID,
        DWORD dwTypeIndex,
        /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType);

    STDMETHODIMP GetOutputAvailableType( 
        DWORD dwOutputStreamID,
        DWORD dwTypeIndex,
        /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType);

    STDMETHODIMP SetInputType( 
        DWORD dwInputStreamID,
        /* [in] */ _In_opt_ IMFMediaType *pType,
        DWORD dwFlags);

    STDMETHODIMP SetOutputType( 
        DWORD dwOutputStreamID,
        /* [in] */ _In_opt_ IMFMediaType *pType,
        DWORD dwFlags);

    STDMETHODIMP GetInputCurrentType( 
        DWORD dwInputStreamID,
        /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType);

    STDMETHODIMP GetOutputCurrentType( 
        DWORD dwOutputStreamID,
        /* [out] */ _Outptr_result_maybenull_ IMFMediaType **ppType);

    STDMETHODIMP GetInputStatus( 
        DWORD dwInputStreamID,
        /* [out] */ _Out_ DWORD *pdwFlags);

    STDMETHODIMP GetOutputStatus( 
        /* [out] */ _Out_ DWORD *pdwFlags);

    STDMETHODIMP SetOutputBounds( 
        LONGLONG hnsLowerBound,
        LONGLONG hnsUpperBound);

    STDMETHODIMP ProcessEvent( 
        DWORD dwInputStreamID,
        /* [in] */ _In_opt_ IMFMediaEvent *pEvent);

    STDMETHODIMP ProcessMessage( 
        MFT_MESSAGE_TYPE eMessage,
        ULONG_PTR ulParam);

    STDMETHODIMP ProcessInput( 
        DWORD dwInputStreamID,
        IMFSample *pSample,
        DWORD dwFlags);

    STDMETHODIMP ProcessOutput( 
        DWORD dwFlags,
        DWORD cOutputBufferCount,
        /* [size_is][out][in] */ MFT_OUTPUT_DATA_BUFFER *pOutputSamples,
        /* [out] */ DWORD *pdwStatus);

protected:
    virtual ~CMft0()
    {
PLOG(">>>=== CMft0()::~CMft0() \n");
        SAFERELEASE(m_pInputAttributes);
        SAFERELEASE(m_pGlobalAttributes);
        SAFERELEASE(m_pSourceTransform);
        SAFERELEASE(m_pInputType);
        SAFERELEASE(m_pOutputType);
        SAFERELEASE(m_pSample);
        SAFERELEASE(m_pRegkeyListener);
        DeleteCriticalSection(&m_critSec);
        m_FdfbFilterState = stUnInit;
PLOG("<<<=== CMft0()::~CMft0() \n");
    }

    STDMETHODIMP GetMediaType(DWORD  dwStreamID, DWORD dwTypeIndex, IMFMediaType **ppType);
    STDMETHODIMP GetMediaTypeOut( DWORD dwStreamId, DWORD dwTypeIndex, IMFMediaType **ppType);
    STDMETHODIMP IsMediaTypeSupported(UINT uiStreamId, IMFMediaType *pIMFMediaType, IMFMediaType **ppIMFMediaTypeFull = NULL);
    STDMETHODIMP GenerateMFMediaTypeListFromDevice(UINT uiStreamId);

    // HasPendingOutput: Returns TRUE if the MFT is holding an input sample.
    BOOL HasPendingOutput() const { return m_pSample != NULL; }

    // IsValidInputStream: Returns TRUE if dwInputStreamID is a valid input stream identifier.
    BOOL IsValidInputStream(DWORD dwInputStreamID);

    // IsValidOutputStream: Returns TRUE if dwOutputStreamID is a valid output stream identifier.
    BOOL IsValidOutputStream(DWORD dwOutputStreamID);

    // STDMETHODIMP GetDefaultStride(LONG *plStride);

    STDMETHODIMP OnFlush();

    CRITICAL_SECTION            m_critSec;

    IMFSample                   *m_pSample;                 // Input sample.
    IMFMediaType                *m_pInputType;              // Input media type.
    IMFMediaType                *m_pOutputType;             // Output media type.

    // Image transform function. (Changes based on the media type.)
    IMFAttributes               *m_pInputAttributes;
    IMFAttributes               *m_pGlobalAttributes;
    GUID                        m_stStreamType;
    IMFTransform                *m_pSourceTransform;
    volatile BOOL               m_bEnableEffects;
    UINT                        m_uiStreamId;
    UINT                        m_percentOfScreen; // TODO: remvoe this variable since it's unused
    UINT                        m_nRefCount;
    CAtlMap<DWORD, CComPtr<IMFMediaType>> m_listOfMediaTypesIn;
    CAtlMap<DWORD, CComPtr<IMFMediaType>> m_listOfMediaTypesOut;

    FdfbFilter                  m_FdfbFilter;               // FDFB filter
    FDFBFilterState             m_FdfbFilterState;
    UINT                        m_InputTypeWidth;
    UINT                        m_InputTypeHeight;
    UINT                        m_OutputTypeWidth;
    UINT                        m_OutputTypeHeight;

    RotationFilter              m_RotationCropFilter;
    RotationFilter              m_RotationFillFilter;

    CscFilter                   m_CscFilter_Yuy2ToNv12;

    MjpgDecoder                 m_MjpgDecoder;

    RegKeyListener              *m_pRegkeyListener;
    DWORD                       m_dwRegKeyFbGlobalEnabled;
    DWORD                       m_dwRegKeyFbDetectionAngleCurrent; // only can be 0, 90, 180, 270
    DWORD                       m_dwRegKeyFbDetectionAnglePrevious; // only can be 0, 90, 180, 270

    IMFMediaBuffer*             m_pMediaBuffer;
	BYTE                        m_pFDFBInBuf;
    BYTE                        m_pFDFBOutBuf[FDFB_OUTPUT_BUF_SIZE];
	PBYTE                       m_pCscPhase1InBuf;
	BYTE                        m_pCscPhase1OutBuf[FDFB_OUTPUT_BUF_SIZE];
	PBYTE                       m_pCscPhase2InBuf;
	BYTE                        m_pCscPhase2OutBuf[FDFB_OUTPUT_BUF_SIZE];

    BOOL                        m_bFirstRun;
    UINT                        m_uiPerformanceCounter;
    LONG                        m_lPerformanceClockSum;

    CComPtr<ID3D11Device> m_pD3d11Device;
    CComPtr<ID3D11DeviceContext> m_pD3d11DeviceContext;

private:

    BOOL m_bWebcamSupportMjpg;
    BOOL m_bWebcamSupportYuy2;
    BOOL m_bWebcamSupportI420;
    BOOL m_bMft0InputMjpg;
    BOOL m_bMft0InputYuy2;
    BOOL m_bMft0InputI420;

    BOOL m_bMft0IsEnabled;

    HRESULT CheckSupportFormat();
    HRESULT AddInputMediaTypeList(GUID guidSubtypeAdd);

    HRESULT AddOutputMediaTypeListBaseOnInput();

    CONST MFT0_AVAILABLE_FORMAT m_pMft0AvailableFormatTable[MFT0_AVAILABLE_FORMAT_TABLE_LEN] = MFT0_AVAILABLE_FORMAT_TABLE;
};

OBJECT_ENTRY_AUTO(__uuidof(Mft0), CMft0)
#endif //_MFT0_H_