// FDFBSample.cpp : Defines the entry point for the console application.
// This app should work on SKL, windows 10, with FDFB enabled drivers
// Change this line to use FB only or FDFB mode:
// VPE_FDFB_MODE_ENUM FDFBMode = VPE_FDFB_MODE_FB_ONLY;
// VPE_FDFB_MODE_ENUM FDFBMode = VPE_FDFB_MODE_FDFB;

#include <atlcomcli.h>
#include <d3d11.h>
#include "../utils_mft.h"
#include "csc.h"
#include "../config.h"


HRESULT CscFilter::CreateStagingTexture(CComPtr<ID3D11Texture2D> &TextureStaging, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    D3D11_TEXTURE2D_DESC stagingDesc = {0};
    HRESULT hr;

    stagingDesc.Width            = iWidth;
    stagingDesc.Height           = iHeight;
    stagingDesc.MipLevels        = 1;
    stagingDesc.ArraySize        = 1;
    stagingDesc.Format           = format_fourcc_to_dxgi(eFourcc);
    stagingDesc.SampleDesc.Count = 1;
    stagingDesc.Usage            = D3D11_USAGE_STAGING;
    stagingDesc.CPUAccessFlags   = D3D11_CPU_ACCESS_WRITE | D3D11_CPU_ACCESS_READ;

    hr = m_pD3d11Device->CreateTexture2D(&stagingDesc, NULL, &TextureStaging);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateStagingTexture() eFourcc=[%d] \n", eFourcc);
    }

    return hr;
}

HRESULT CscFilter::CreateInputFrame(CComPtr<ID3D11Texture2D> &pTextureIn, CComPtr<ID3D11VideoProcessorInputView> &pViewIn, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    D3D11_TEXTURE2D_DESC inputDesc = {0};
    D3D11_VIDEO_PROCESSOR_INPUT_VIEW_DESC inputViewDesc = {0};
    HRESULT hr;

    inputDesc.Width                    = iWidth;
    inputDesc.Height                   = iHeight;
    inputDesc.MipLevels                = 1;
    inputDesc.ArraySize                = 1;
    inputDesc.Format                   = format_fourcc_to_dxgi(eFourcc);
    inputDesc.SampleDesc.Count         = 1;
    inputDesc.Usage                    = D3D11_USAGE_DEFAULT;

    inputViewDesc.FourCC               = eFourcc;
    inputViewDesc.ViewDimension        = D3D11_VPIV_DIMENSION_TEXTURE2D;
    inputViewDesc.Texture2D.MipSlice   = 0;
    inputViewDesc.Texture2D.ArraySlice = 0;

    hr = m_pD3d11Device->CreateTexture2D(&inputDesc, NULL, &pTextureIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateTexture2D() input \n");
        goto END;
    }

    hr = m_pD3d11VideoDevice->CreateVideoProcessorInputView(pTextureIn, m_pD3d11VideoProcessorEnum, &inputViewDesc, &pViewIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateVideoProcessorInputView() \n");
        goto END;
    }
END:
    return hr;
}

HRESULT CscFilter::CreateOutputFrame(CComPtr<ID3D11Texture2D> &pTextureOut, CComPtr<ID3D11VideoProcessorOutputView> &pViewOut, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    D3D11_TEXTURE2D_DESC outputDesc = {0};
    HRESULT hr;
    D3D11_VIDEO_PROCESSOR_OUTPUT_VIEW_DESC outputViewDesc;

    outputDesc.Width            = iWidth;
    outputDesc.Height           = iHeight;
    outputDesc.MipLevels        = 1;
    outputDesc.ArraySize        = 1;
    outputDesc.Format           = format_fourcc_to_dxgi(eFourcc);
    outputDesc.SampleDesc.Count = 1;
    outputDesc.Usage            = D3D11_USAGE_DEFAULT;
    outputDesc.BindFlags        = D3D11_BIND_RENDER_TARGET;

    hr = m_pD3d11Device->CreateTexture2D(&outputDesc, NULL, &pTextureOut);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateTexture2D output \n");
        goto END;
    }

    ZeroMemory(&outputViewDesc, sizeof(outputViewDesc));
    outputViewDesc.ViewDimension = D3D11_VPOV_DIMENSION_TEXTURE2D;
    hr = m_pD3d11VideoDevice->CreateVideoProcessorOutputView(pTextureOut, m_pD3d11VideoProcessorEnum, &outputViewDesc, &pViewOut);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateVideoProcessorOutputView \n");
        goto END;
    }

END:
    return hr;
}

HRESULT CscFilter::CopyBufToTextureStaging(ID3D11Texture2D *pStagingTexture, BYTE *pBufIn, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    HRESULT hr = S_OK;
    BYTE *dst;
    UINT pitch;
    D3D11_MAPPED_SUBRESOURCE lr;

    ZeroMemory(&lr, sizeof(lr));
    hr = m_pD3d11DeviceContext->Map(pStagingTexture, 0, D3D11_MAP_WRITE, 0, &lr);
    if (FAILED(hr))
    {
        PLOG("Failed to m_pD3d11DeviceContext->Map() \n");
        goto END;
    }

    dst   = (BYTE*)lr.pData;
    pitch = lr.RowPitch;

    if (eFourcc == FOURCC_YUY2)
    {
        // YUY2 plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth*2);
	        dst += pitch;
	        pBufIn += iWidth*2;
        }
    }
    else if (eFourcc == FOURCC_I420)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth);
	        dst += pitch;
	        pBufIn += iWidth;
        }
        // U plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth/2);
	        dst += pitch;
	        pBufIn += iWidth;
        }
        // V plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth/2);
	        dst += pitch;
	        pBufIn += iWidth;
        }
    }
    else if (eFourcc == FOURCC_NV12)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(dst, pBufIn, iWidth);
            dst += pitch;
            pBufIn += iWidth;
        }
 
        // U,V plane
        for (INT cnt = 0; cnt < iHeight / 2; cnt++)
        {
            memcpy(dst, pBufIn, iWidth);
            dst += pitch;
            pBufIn += iWidth;
        }
    }
    else
    {
        PLOG("Failed to CopyBufToTextureStaging due to invalid eFourcc=[%d] \n", eFourcc);
        hr = E_INVALIDARG;
        goto END;
    }

    m_pD3d11DeviceContext->Unmap(pStagingTexture, 0);
END:
    return hr;
}

HRESULT CscFilter::CopyTextureStagingToBuf(BYTE *pBufNv12Out, ID3D11Texture2D *pStagingTexture, INT iWidth, INT iHeight, FOURCC eFourccOut)
{
    HRESULT hr = S_OK;
    UINT pitch;
    D3D11_MAPPED_SUBRESOURCE lr;
    BYTE *src;

    ZeroMemory(&lr, sizeof(lr));
    hr = m_pD3d11DeviceContext->Map(pStagingTexture, 0, D3D11_MAP_READ, 0, &lr);
    if (FAILED(hr))
    {
        PLOG("Failed to m_pD3d11DeviceContext->Map() \n");
        goto END;
    }

    src   = (BYTE*)lr.pData;
    pitch = lr.RowPitch;

    if (eFourccOut == FOURCC_YUY2)
    {
        // YUY2 plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth*2);
            src += pitch;
            pBufNv12Out += iWidth*2;
        }
    }
    else if (eFourccOut == FOURCC_I420)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth);
            src += pitch;
            pBufNv12Out += iWidth;
        }
        // U plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth/2);
            src += pitch;
            pBufNv12Out += iWidth;
        }
        // V plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth/2);
            src += pitch;
            pBufNv12Out += iWidth;
        }
    }
    else if (eFourccOut == FOURCC_NV12)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth);
            src += pitch;
            pBufNv12Out += iWidth;
        }
 
        // U,V plane
        for (INT cnt = 0; cnt < iHeight / 2; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth);
            src += pitch;
            pBufNv12Out += iWidth;
        }
    }
    else
    {
        PLOG("Failed to CopyTextureStagingToBuf() due to invalid eFourcc=[%d] \n", eFourccOut);
        hr = E_INVALIDARG;
        goto END;
    }

    m_pD3d11DeviceContext->Unmap(pStagingTexture, 0);

END:
    return hr;
}

HRESULT CscFilter::VideoProcessorBlt(CComPtr<ID3D11VideoProcessorOutputView> pViewOut,  CComPtr<ID3D11VideoProcessorInputView> pViewIn)
{
    HRESULT hr = S_OK;

    // blt
    D3D11_VIDEO_PROCESSOR_STREAM StreamData;

    ZeroMemory(&StreamData, sizeof(StreamData));
    StreamData.Enable                = TRUE;
    StreamData.OutputIndex           = 0;
    StreamData.InputFrameOrField     = 0;
    StreamData.PastFrames            = 0;
    StreamData.FutureFrames          = 0;
    StreamData.ppPastSurfaces        = NULL;
    StreamData.ppFutureSurfaces      = NULL;
    StreamData.pInputSurface         = pViewIn;
    StreamData.ppPastSurfacesRight   = NULL;
    StreamData.ppFutureSurfacesRight = NULL;

    hr = m_pD3d11VideoContext->VideoProcessorBlt(m_pD3d11VideoProcessor, pViewOut, 0, 1, &StreamData);
    if (FAILED(hr))
    {
        PLOG("ERROR: VideoProcessorBlt \n");
        goto END;
    }

END:
    return hr;
}

HRESULT CscFilter::CreateDevice()
{
    HRESULT hr = S_OK;

    D3D_FEATURE_LEVEL __FeatureLevel;
    D3D_FEATURE_LEVEL pFeatureLevels[] = { D3D_FEATURE_LEVEL_11_1, D3D_FEATURE_LEVEL_11_0 };
    INT flCount = sizeof(pFeatureLevels) / sizeof(D3D_FEATURE_LEVEL);

    hr = D3D11CreateDevice(NULL, D3D_DRIVER_TYPE_HARDWARE, NULL, D3D11_CREATE_DEVICE_BGRA_SUPPORT,
        pFeatureLevels, flCount, D3D11_SDK_VERSION, &m_pD3d11Device, &__FeatureLevel, (ID3D11DeviceContext**)&m_pD3d11DeviceContext);
    if (FAILED(hr))
    {
        PLOG("ERROR: D3D11CreateDevice failed \n");
        goto END;
    }

    hr = m_pD3d11Device->QueryInterface(__uuidof(ID3D11VideoDevice), (void**)&m_pD3d11VideoDevice);
    if (FAILED(hr))
    {
        PLOG("ERROR: QueryInterface ID3D11VideoDevice failed \n");
        goto END;
    }

END:
    return hr;
}

HRESULT CscFilter::CreateVideoProcessor(INT iWidth, INT iHeight)
{
    HRESULT hr;
    D3D11_VIDEO_PROCESSOR_CAPS caps;
    UINT flags;
    D3D11_VIDEO_PROCESSOR_CONTENT_DESC desc;

#if TARGET_PLATFORM_WIN10
    hr = m_pD3d11DeviceContext->QueryInterface(__uuidof(ID3D11VideoContext1), (void**)&m_pD3d11VideoContext);
#else
    hr = m_pD3d11DeviceContext->QueryInterface(__uuidof(ID3D11VideoContext), (void**)&m_pD3d11VideoContext);
#endif

    if (FAILED(hr))
    {
        PLOG("Failed to QueryInterface() ID3D11VideoContext1 and ID3D11VideoContext \n");
        goto END;
    }

    ZeroMemory(&desc, sizeof(D3D11_VIDEO_PROCESSOR_CONTENT_DESC));
    desc.InputFrameFormat           = D3D11_VIDEO_FRAME_FORMAT_PROGRESSIVE;
    desc.InputFrameRate.Numerator   = desc.OutputFrameRate.Numerator   = 30;
    desc.InputFrameRate.Denominator = desc.OutputFrameRate.Denominator = 1;
    desc.InputWidth                 = desc.OutputWidth  = iWidth;
    desc.InputHeight                = desc.OutputHeight = iHeight;
    desc.Usage                      = D3D11_VIDEO_USAGE_PLAYBACK_NORMAL;

    hr = m_pD3d11VideoDevice->CreateVideoProcessorEnumerator(&desc, &m_pD3d11VideoProcessorEnum);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateVideoProcessorEnumerator failed \n");
        goto END;
    }

    hr = m_pD3d11VideoProcessorEnum->GetVideoProcessorCaps(&caps);
    if (FAILED(hr))
    {
        PLOG("ERROR: GetVideoProcessorCaps failed \n");
        goto END;
    }

    hr = m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat(DXGI_FORMAT_YUY2, &flags);
    if (FAILED(hr))
    {
        PLOG("ERROR: CheckVideoProcessorFormat input format failed \n");
        goto END;
    }

    if ((flags & D3D11_VIDEO_PROCESSOR_FORMAT_SUPPORT_INPUT) == 0)
    {
        PLOG("ERROR: don't supported input format \n");
        goto END;
    }

    hr = m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat(DXGI_FORMAT_NV12, &flags);
    if (FAILED(hr))
    {
        PLOG("ERROR: CheckVideoProcessorFormat output format failed \n");
        goto END;
    }

    if ((flags & D3D11_VIDEO_PROCESSOR_FORMAT_SUPPORT_OUTPUT) == 0)
    {
        PLOG("ERROR: don't supported output format \n");
        goto END;
    }

    hr = m_pD3d11VideoDevice->CreateVideoProcessor(m_pD3d11VideoProcessorEnum, 0, &m_pD3d11VideoProcessor);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateVideoProcessor() failed \n");
        goto END;
    }

    m_pD3d11VideoContext->VideoProcessorSetStreamAutoProcessingMode(m_pD3d11VideoProcessor, 0, FALSE);

END:
    return hr;
}

HRESULT CscFilter::Init(FOURCC eFourccOut, FOURCC eFourccIn, INT iWidth, INT iHeight)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== CscFilter::Init() \n");
    CAutoLock cObjectLock(&m_MutexLock);

    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextureIn               = NULL;
    m_pTextureOut              = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;
                               
    m_pD3d11DeviceContext      = NULL;
    m_pD3d11Device             = NULL;

    m_iWidth                   = iWidth;
    m_iHeight                  = iHeight;

    if (eFourccIn == FOURCC_I420)
    {
        PLOG("Cannot use HW color space conversion since D3D11 doesn't support DXGI_FORMAT_I420 \n", eFourccIn);
        hr = E_INVALIDARG;
        goto END;
    }

    if (eFourccIn != FOURCC_YUY2 && eFourccIn != FOURCC_NV12)
    {
        PLOG("Failed to Init(), color space conversion doesn't support this input format eFourccIn=[%d] \n", eFourccIn);
        hr = E_INVALIDARG;
        goto END;
    }

    if (eFourccOut != FOURCC_YUY2 && eFourccOut != FOURCC_NV12)
    {
        PLOG("Failed to Init(), color space conversion doesn't support this output format eFourccOut=[%d] \n", eFourccOut);
        hr = E_INVALIDARG;
        goto END;
    }

    m_eFourccIn  = eFourccIn;
    m_eFourccOut = eFourccOut;

    hr = CreateDevice();
    if (FAILED(hr))
    {
        PLOG("Failed to CreateDevice() \n");
        goto END;
    }

    hr = CreateVideoProcessor(iWidth, iHeight);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateVideoProcessor() \n");
        goto END;
    }

    hr = CreateStagingTexture(m_pTextureStagingIn, iWidth, iHeight, eFourccIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateStagingTexture() m_pTextureStagingIn \n");
        goto END;
    }

    hr = CreateStagingTexture(m_pTextureStagingOut, iWidth, iHeight, eFourccOut);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateStagingTexture() m_pTextureStagingOut \n");
        goto END;
    }

    hr = CreateInputFrame(m_pTextureIn, m_pViewIn, iWidth, iHeight, eFourccIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateInputFrame() \n");
        goto END;
    }

    hr = CreateOutputFrame(m_pTextureOut, m_pViewOut, iWidth, iHeight, eFourccOut);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateOutputFrame() \n");
        goto END;
    }

    m_bInited = TRUE;

END:
    PLOG("<<<=== CscFilter::Init() \n");
    return hr;
}

HRESULT CscFilter::Destroy()
{
    PLOG(">>>=== CscFilter::Destroy() \n");
    CAutoLock cObjectLock(&m_MutexLock);

    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextureIn               = NULL;
    m_pTextureOut              = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;
                               
    m_pD3d11DeviceContext      = NULL;
    m_pD3d11Device             = NULL;

    PLOG("<<<=== CscFilter::Destroy() \n");
    return S_OK;
}

HRESULT CscFilter::Blt(BYTE *pBufOut, BYTE *pBufIn, INT iWidth, INT iHeight)
{
    HRESULT hr = S_OK;
    CAutoLock cObjectLock(&m_MutexLock);
    if (!m_bInited)
    {
        goto END;
    }

    hr = CopyBufToTextureStaging(m_pTextureStagingIn, pBufIn, iWidth, iHeight, m_eFourccIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CopyBufToTextureStaging() \n");
        goto END;
    }

    // copy staging to vpblt input, needing a temp texture for copying CPU to GPU memory.
    m_pD3d11DeviceContext->CopyResource(m_pTextureIn, m_pTextureStagingIn);

    VideoProcessorBlt(m_pViewOut, m_pViewIn);

    // copy vpblt output to staging, needing a temp texture for copying GPU to CPU memory.
    m_pD3d11DeviceContext->CopyResource(m_pTextureStagingOut, m_pTextureOut);

    hr = CopyTextureStagingToBuf(pBufOut, m_pTextureStagingOut, iWidth, iHeight, m_eFourccOut);
    if (FAILED(hr))
    {
        PLOG("Failed to CopyTextureStagingToBuf() \n");
    }

END:
    return hr;
}

#if TARGET_PLATFORM_WIN10
HRESULT CscFilter::SetHorizontalMirrorGpu(BOOL bEnable)
{
    m_pD3d11VideoContext->VideoProcessorSetStreamMirror(m_pD3d11VideoProcessor, 0, bEnable, bEnable, 0);
    return S_OK;
}
#endif