#include "windows.h"
#include "string.h"
#include "stdlib.h"
#include "decode_mjpg.h"
#include "..\utils_mft.h"

// porting from MSDK sample code: 
// (1) https://software.intel.com/en-us/articles/media-sdk-tutorials-for-client-and-server, and 
// (2) https://software.intel.com/en-us/articles/supported-output-color-formats-in-mjpeg-decoding

HRESULT copy_mfx_surf_to_buf(BYTE *byteBufOut, mfxFrameSurface1* pSurface)
{
    mfxFrameInfo* pInfo = &pSurface->Info;
    mfxFrameData* pData = &pSurface->Data;
    HRESULT hr = S_OK;
    BYTE *pTempOut = byteBufOut;
    BYTE *pTempInYplane = pData->Y;

    if (!byteBufOut)
    {
        PLOG("Failed to copy_mfx_surf_to_buf() due to null pointer byteBufOut \n");
        goto END;
    }

    // Y plane
    for (INT i = 0; i < pInfo->CropH; i++)
    {
        memcpy_s(pTempOut, pInfo->CropW, pTempInYplane, pInfo->CropW);
        pTempOut      += pInfo->CropW;
        pTempInYplane += pData->Pitch;
    }

    // UV-plane
    pTempInYplane = pData->UV;
    for (INT i = 0; i < pInfo->CropH/2; i++)
    {
        memcpy_s(pTempOut, pInfo->CropW, pTempInYplane, pInfo->CropW);
        pTempOut      += pInfo->CropW;
        pTempInYplane += pData->Pitch;
    }
END:
    return hr;
}

#if 0 // 1st version we port from Media SDK sample code, which is unused anymore but still keep it if case.
HRESULT GetMfxBitStream(mfxBitstream *pMfxBsOut, BYTE *pMjpgBufIn, INT iMjpgBufLen)
{
    memset(pMfxBsOut, 0, sizeof(*pMfxBsOut));
    pMfxBsOut->DataOffset = 0;
    pMfxBsOut->Data       = pMjpgBufIn;
    pMfxBsOut->MaxLength  = CSC_BUF_SIZE_MAX;
    pMfxBsOut->DataLength = iMjpgBufLen;

    return S_OK;
}

HRESULT GetMfxVideoParamsMjpeg(mfxVideoParam *pMfxVp)
{
    memset(pMfxVp, 0, sizeof(*pMfxVp));
    pMfxVp->mfx.CodecId                = MFX_CODEC_JPEG;
    pMfxVp->IOPattern                  = MFX_IOPATTERN_OUT_SYSTEM_MEMORY;
    pMfxVp->mfx.FrameInfo.FourCC       = MFX_FOURCC_NV12;
    pMfxVp->mfx.FrameInfo.ChromaFormat = MFX_CHROMAFORMAT_YUV420;

    return S_OK;
}

mfxFrameSurface1* AllocateMfxSurfaceMjpeg(MFXVideoDECODE *pMfxDecIn, mfxVideoParam *pMfxVpIn)
{
    mfxStatus sts;

    // Query number of required surfaces for decoder
    mfxFrameAllocRequest Request = {0};
    sts = pMfxDecIn->QueryIOSurf(pMfxVpIn, &Request);
    MSDK_IGNORE_MFX_STS(sts, MFX_WRN_PARTIAL_ACCELERATION);
    if (sts < MFX_ERR_NONE) { return NULL; }

    mfxU16 width                  = (mfxU16) MSDK_ALIGN32(Request.Info.Width);
    mfxU16 height                 = (mfxU16) MSDK_ALIGN32(Request.Info.Height);
    mfxU8  bitsPerPixel           = 12; // NV12 format is a 12 bits per pixel format
    mfxU32 surfaceSize            = width * height * bitsPerPixel / 8;
    mfxU8* pSurfaceBuf            = (mfxU8*) new mfxU8[surfaceSize];
    mfxFrameSurface1 *pMfxSurface = (mfxFrameSurface1 *) new mfxFrameSurface1;

    memset(pMfxSurface, 0, sizeof(*pMfxSurface));
    memcpy(&(pMfxSurface->Info), &(pMfxVpIn->mfx.FrameInfo), sizeof(mfxFrameInfo));
    pMfxSurface->Data.Y     = pSurfaceBuf;
    pMfxSurface->Data.UV    = pMfxSurface->Data.Y + width * height;
    pMfxSurface->Data.Pitch = width;

END:
    return pMfxSurface;
}

void DeleteMfxSurfaceMjpeg(mfxFrameSurface1* pMfxSurf)
{
    if (!pMfxSurf)
    {
        goto END;
    }
    if (pMfxSurf->Data.Y)
    {
        delete pMfxSurf->Data.Y;
    }
    delete pMfxSurf;
END:
    return;
}

HRESULT mjpg_decode(BYTE *pMjpgBitStreamIn, INT iMjpgBitStreamLenIn, BYTE *byteBufOut)
{
    mfxStatus         sts            = MFX_ERR_NONE;
    mfxFrameSurface1* pMfxDecodedSurf = NULL;
    MFXVideoSession   session;
    mfxBitstream      mfxBS;
    mfxVideoParam     mfxVideoParams;
    mfxFrameSurface1* pMfxSurface = NULL;
    HRESULT           hr = S_OK;
    static BOOL       bFirstRun = TRUE;
    static BYTE       pMjpgDecoded[CSC_BUF_SIZE_MAX] = {0};
    MFXVideoDECODE    *pMfxDEC = NULL;

    if (!byteBufOut)
    {
        PLOG("Failed to mjpg_decode() due to null pointer byteBufOut() \n");
        goto END;
    }

    sts = Initialize(MFX_IMPL_AUTO_ANY, /* mfxVersion */{{0, 1}}, &session, NULL);
    GOTO_END_FALSE(MFX_OK(sts), "Failed to Initialize() \n");

    hr = GetMfxBitStream(&mfxBS, pMjpgBitStreamIn, iMjpgBitStreamLenIn);
    GOTO_END_FAILED(hr, "Failed to GetMfxBitStream() \n");

    hr = GetMfxVideoParamsMjpeg(&mfxVideoParams);
    GOTO_END_FAILED(hr, "Failed to GetMfxVideoParamsMjpeg() \n");

    pMfxDEC = new MFXVideoDECODE(session);
    sts = pMfxDEC->DecodeHeader(&mfxBS, &mfxVideoParams);
    MSDK_IGNORE_MFX_STS(sts, MFX_WRN_PARTIAL_ACCELERATION);
    GOTO_END_FALSE(MFX_OK(sts), "Failed to pMfxDEC->DecodeHeader() \n");

    // sts = mfxDEC.Query(&mfxVideoParams, &mfxVideoParams); // Validate video decode parameters (optional)

    pMfxSurface = AllocateMfxSurfaceMjpeg(pMfxDEC, &mfxVideoParams);

    // Initialize the Media SDK decoder
    sts = pMfxDEC->Init(&mfxVideoParams);
    MSDK_IGNORE_MFX_STS(sts, MFX_WRN_PARTIAL_ACCELERATION);
    GOTO_END_FALSE(MFX_OK(sts), "Failed to pMfxDEC->Init() \n");

    for (int iTryTimes = 0; iTryTimes < 10; iTryTimes++)
    {
        mfxSyncPoint syncp;
        sts = pMfxDEC->DecodeFrameAsync(&mfxBS, pMfxSurface, &pMfxDecodedSurf, &syncp);

        // Wait if device is busy, then repeat the same call to DecodeFrameAsync
        if (MFX_WRN_DEVICE_BUSY == sts) { MSDK_SLEEP(1); goto CONTINUE; }

        // Ignore warnings if output is available, if no output and no action required just repeat the DecodeFrameAsync call
        if (MFX_ERR_NONE < sts && syncp) { sts = MFX_ERR_NONE; }

        if (MFX_ERR_NONE == sts)
        {
            sts = session.SyncOperation(syncp, 60000); // Synchronize. Wait until decoded frame is ready
            GOTO_END_FALSE(MFX_OK(sts), "Failed to session.SyncOperation() \n");
            break;
        }
CONTINUE:
        PLOG("retry pMfxDEC->DecodeFrameAsync() iTryTimes=[%d] sts=[%d] \n", iTryTimes, sts);
    }

    GOTO_END_FALSE(MFX_OK(sts), "Failed to pMfxDEC->DecodeFrameAsync() sts=[%d] \n", sts);

    copy_mfx_surf_to_buf(byteBufOut, pMfxDecodedSurf);

    if (bFirstRun)
    {
        bFirstRun = FALSE;
        copy_mfx_surf_to_buf(pMjpgDecoded, pMfxDecodedSurf);
        DUMP_BUFFER("C:\\Temp\\mft0_mjpg_decode_out", MFVideoFormat_NV12,
            pMjpgDecoded, CSC_BUF_SIZE_MAX, pMfxDecodedSurf->Info.CropW, pMfxDecodedSurf->Info.CropH);
    }
END:
    pMfxDEC->Close();
    DeleteMfxSurfaceMjpeg(pMfxSurface);
    Release();

    return hr;
}
#endif


HRESULT MjpgDecoder::AllocateMfxSurfaceMjpeg(INT iWidth, INT iHeight)
{
    mfxU16 width                  = (mfxU16) MSDK_ALIGN32(iWidth);
    mfxU16 height                 = (mfxU16) MSDK_ALIGN32(iHeight);
    mfxU8  bitsPerPixel           = 12; // NV12 format is a 12 bits per pixel format
    mfxU32 surfaceSize            = width * height * bitsPerPixel / 8;
    mfxU8* pSurfaceBuf            = (mfxU8*) new mfxU8[surfaceSize];

    memset(&m_MfxSurface, 0, sizeof(m_MfxSurface));
    m_MfxSurface.Data.Y     = pSurfaceBuf;
    m_MfxSurface.Data.UV    = m_MfxSurface.Data.Y + width * height;
    m_MfxSurface.Data.Pitch = width;

END:
    return S_OK;
}

HRESULT MjpgDecoder::SetMfxSurfaceMjpeg(mfxFrameInfo FrameInfo)
{
    memcpy(&(m_MfxSurface.Info), &FrameInfo, sizeof(mfxFrameInfo));

    return S_OK;
}


HRESULT MjpgDecoder::GetMfxBitStream(mfxBitstream *pMfxBsOut, BYTE *pMjpgBufIn, INT iMjpgBufLen)
{
    memset(pMfxBsOut, 0, sizeof(*pMfxBsOut));
    pMfxBsOut->DataOffset = 0;
    pMfxBsOut->Data       = pMjpgBufIn;
    pMfxBsOut->MaxLength  = CSC_BUF_SIZE_MAX;
    pMfxBsOut->DataLength = iMjpgBufLen;

    return S_OK;
}

HRESULT MjpgDecoder::GetMfxVideoParamsMjpeg(mfxVideoParam *pMfxVp)
{
    memset(pMfxVp, 0, sizeof(*pMfxVp));
    pMfxVp->mfx.CodecId                = MFX_CODEC_JPEG;
    pMfxVp->IOPattern                  = MFX_IOPATTERN_OUT_SYSTEM_MEMORY;
    pMfxVp->mfx.FrameInfo.FourCC       = MFX_FOURCC_NV12;
    pMfxVp->mfx.FrameInfo.ChromaFormat = MFX_CHROMAFORMAT_YUV420;

    return S_OK;
}

HRESULT MjpgDecoder::Init(FOURCC eFourccOut, INT iWidth, INT iHeight)
{
    mfxStatus         sts = MFX_ERR_NONE;
    HRESULT           hr = S_OK;
    PLOG(">>>=== MjpgDecoder::Init() FOURCC=[%d] iWidth=[%d] iHeight=[%d] \n", eFourccOut, iWidth, iHeight);

    // initiated
    if (m_pMfxDEC)
    {
        goto END;
    }

    m_bMfxVideoDecodeInited = FALSE;

    m_iWidth = iWidth;
    m_iHeight = iHeight;

    memset(m_pMjpgDecoded, 0, DECODED_BUF_SIZE_MAX);

    sts = Initialize(MFX_IMPL_AUTO_ANY, /* mfxVersion */{{0, 1}}, &m_session, NULL);
    GOTO_END_FALSE(MFX_OK(sts), "Failed to Initialize() \n");

    m_pMfxDEC = new MFXVideoDECODE(m_session);
    GOTO_END_NULL(m_pMfxDEC, "Failed to MjpgDecoder::Init() due to null pointer m_pMfxDEC \n");

    hr = GetMfxVideoParamsMjpeg(&m_mfxVideoParams);
    GOTO_END_FAILED(hr, "Failed to GetMfxVideoParamsMjpeg() \n");

    hr = AllocateMfxSurfaceMjpeg(iWidth, iHeight);
    GOTO_END_FAILED(hr, "Failed to AllocateMfxSurfaceMjpeg() \n");

    PLOG("<<<=== MjpgDecoder::Init() \n");
END:
    return hr;
}

HRESULT MjpgDecoder::Destroy()
{
    mfxStatus         sts = MFX_ERR_NONE;
    HRESULT           hr = S_OK;

    m_bMfxVideoDecodeInited = FALSE;

    if (m_pMfxDEC)
    {
        m_pMfxDEC->Close();
        delete m_pMfxDEC;
        m_pMfxDEC = NULL;
    }

    if (m_MfxSurface.Data.Y)
    {
        delete m_MfxSurface.Data.Y;
        m_MfxSurface.Data.Y = NULL;
    }

END:
    return hr;
}

HRESULT MjpgDecoder::Decode(BYTE *pMjpgBitStreamIn, INT iMjpgBitStreamLenIn, BYTE *byteBufOut)
{
    mfxStatus         sts            = MFX_ERR_NONE;
    mfxFrameSurface1* pMfxDecodedSurf = NULL;
    mfxBitstream      mfxBS;
    HRESULT           hr = S_OK;
    static BOOL       bFirstRun = TRUE;

    if (!byteBufOut || !pMjpgBitStreamIn)
    {
        PLOG("Failed to mjpg_decode() due to null pointer byteBufOut or pMjpgBitStreamIn \n");
        goto END;
    }

    hr = GetMfxBitStream(&mfxBS, pMjpgBitStreamIn, iMjpgBitStreamLenIn);
    GOTO_END_FAILED(hr, "Failed to GetMfxBitStream() \n");

    hr = GetMfxVideoParamsMjpeg(&m_mfxVideoParams);
    GOTO_END_FAILED(hr, "Failed to GetMfxVideoParamsMjpeg() \n");

    sts = m_pMfxDEC->DecodeHeader(&mfxBS, &m_mfxVideoParams);
    MSDK_IGNORE_MFX_STS(sts, MFX_WRN_PARTIAL_ACCELERATION);
    GOTO_END_FALSE(MFX_OK(sts), "Failed to pMfxDEC->DecodeHeader() \n");

    hr = SetMfxSurfaceMjpeg(m_mfxVideoParams.mfx.FrameInfo);
    GOTO_END_FAILED(hr, "Failed to pMfxDEC->DecodeHeader() \n");

    if (!m_bMfxVideoDecodeInited)
    {
        m_bMfxVideoDecodeInited = TRUE;
        // Initialize the Media SDK decoder
        sts = m_pMfxDEC->Init(&m_mfxVideoParams);
        MSDK_IGNORE_MFX_STS(sts, MFX_WRN_PARTIAL_ACCELERATION);
        GOTO_END_FALSE(MFX_OK(sts), "Failed to pMfxDEC->Init() \n");
    }

    for (int iTryTimes = 0; iTryTimes < 10; iTryTimes++)
    {
        mfxSyncPoint syncp;
        sts = m_pMfxDEC->DecodeFrameAsync(&mfxBS, &m_MfxSurface, &pMfxDecodedSurf, &syncp);

        // Wait if device is busy, then repeat the same call to DecodeFrameAsync
        if (MFX_WRN_DEVICE_BUSY == sts) { MSDK_SLEEP(1); goto CONTINUE; }

        // Ignore warnings if output is available, if no output and no action required just repeat the DecodeFrameAsync call
        if (MFX_ERR_NONE < sts && syncp) { sts = MFX_ERR_NONE; }

        if (MFX_ERR_NONE == sts)
        {
            sts = m_session.SyncOperation(syncp, 60000); // Synchronize. Wait until decoded frame is ready
            GOTO_END_FALSE(MFX_OK(sts), "Failed to m_session.SyncOperation() \n");
            break;
        }
CONTINUE:
        PLOG("retry pMfxDEC->DecodeFrameAsync() iTryTimes=[%d] sts=[%d] \n", iTryTimes, sts);
    }

    GOTO_END_FALSE(MFX_OK(sts), "Failed to pMfxDEC->DecodeFrameAsync() sts=[%d] \n", sts);

    copy_mfx_surf_to_buf(byteBufOut, pMfxDecodedSurf);

    if (bFirstRun)
    {
        bFirstRun = FALSE;
        copy_mfx_surf_to_buf(m_pMjpgDecoded, pMfxDecodedSurf);
        DUMP_BUFFER("C:\\Temp\\mft0_mjpg_decode_out", MFVideoFormat_NV12,
            m_pMjpgDecoded, CSC_BUF_SIZE_MAX, pMfxDecodedSurf->Info.CropW, pMfxDecodedSurf->Info.CropH);
    }
END:
    return hr;
}
