// FDFBSample.cpp : Defines the entry point for the console application.
// This app should work on SKL, windows 10, with FDFB enabled drivers
// Change this line to use FB only or FDFB mode:
// VPE_FDFB_MODE_ENUM FDFBMode = VPE_FDFB_MODE_FDFB;

#include <atlcomcli.h>
#include <d3d11.h>
//#include <memory>
#include "fdfb.h"

FdfbFilter::FdfbFilter()
{
    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextureIn               = NULL;
    m_pTextOut                 = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;
                               
    m_pD3d11DeviceContext      = NULL;
    m_pD3d11Device             = NULL;
};

FdfbFilter::~FdfbFilter()
{

}

HRESULT FdfbFilter::CreateStagingTexture(CComPtr<ID3D11Texture2D> &pTextureStaging, INT iWidth, INT iHeight, DXGI_FORMAT eFormat)
{
    D3D11_TEXTURE2D_DESC stagingDesc = {0};
    HRESULT hr;

    stagingDesc.Width            = iWidth;
    stagingDesc.Height           = iHeight;
    stagingDesc.MipLevels        = 1;
    stagingDesc.ArraySize        = 1;
    stagingDesc.Format           = eFormat;
    stagingDesc.SampleDesc.Count = 1;
    stagingDesc.Usage            = D3D11_USAGE_STAGING;
    stagingDesc.CPUAccessFlags   = D3D11_CPU_ACCESS_WRITE | D3D11_CPU_ACCESS_READ;

    hr = m_pD3d11Device->CreateTexture2D(&stagingDesc, NULL, &pTextureStaging);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateStagingTexture \n");
    }

    return hr;
}

HRESULT FdfbFilter::CreateInputFrame(CComPtr<ID3D11Texture2D> &pTextureIn, CComPtr<ID3D11VideoProcessorInputView> &pViewIn, INT iWidth, INT iHeight, DXGI_FORMAT eFormat, UINT uiFourCC)
{
    D3D11_TEXTURE2D_DESC inputDesc = {0};
    D3D11_VIDEO_PROCESSOR_INPUT_VIEW_DESC inputViewDesc = {0};
    HRESULT hr;

    inputDesc.Width = iWidth;
    inputDesc.Height = iHeight;
    inputDesc.MipLevels = 1;
    inputDesc.ArraySize = 1;
    inputDesc.Format = eFormat;
    inputDesc.SampleDesc.Count = 1;
    inputDesc.Usage = D3D11_USAGE_DEFAULT;

    hr = m_pD3d11Device->CreateTexture2D(&inputDesc, NULL, &pTextureIn);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateTexture2D input \n");
        goto END;
    }

    inputViewDesc.FourCC = uiFourCC;
    inputViewDesc.ViewDimension = D3D11_VPIV_DIMENSION_TEXTURE2D;
    inputViewDesc.Texture2D.MipSlice = 0;
    inputViewDesc.Texture2D.ArraySlice = 0;
    hr = m_pD3d11VideoDevice->CreateVideoProcessorInputView(pTextureIn, m_pD3d11VideoProcessorEnum, &inputViewDesc, &pViewIn);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateVideoProcessorInputView \n");
        goto END;
    }
END:
    return hr;
}

HRESULT FdfbFilter::CreateOutputFrame(CComPtr<ID3D11Texture2D> &pTextureOut, CComPtr<ID3D11VideoProcessorOutputView> &pViewOut, INT iWidth, INT iHeight, DXGI_FORMAT eFormat)
{
    D3D11_TEXTURE2D_DESC outputDesc = {0};
    HRESULT hr;
    D3D11_VIDEO_PROCESSOR_OUTPUT_VIEW_DESC outputViewDesc;

    outputDesc.Width = iWidth;
    outputDesc.Height = iHeight;
    outputDesc.MipLevels = 1;
    outputDesc.ArraySize = 1;
    outputDesc.Format = eFormat;
    outputDesc.SampleDesc.Count = 1;
    outputDesc.Usage = D3D11_USAGE_DEFAULT;
    outputDesc.BindFlags = D3D11_BIND_RENDER_TARGET;

    hr = m_pD3d11Device->CreateTexture2D(&outputDesc, NULL, &pTextureOut);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateTexture2D output \n");
        goto END;
    }

    ZeroMemory(&outputViewDesc, sizeof(outputViewDesc));
    outputViewDesc.ViewDimension = D3D11_VPOV_DIMENSION_TEXTURE2D;
    hr = m_pD3d11VideoDevice->CreateVideoProcessorOutputView(pTextureOut, m_pD3d11VideoProcessorEnum, &outputViewDesc, &pViewOut);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateVideoProcessorOutputView \n");
        goto END;
    }

END:
    return hr;
}

// please be noticed only staging texture can be accessed by both CPU and GPU
// D3D11_USAGE enumeration: https://msdn.microsoft.com/en-us/library/windows/desktop/ff476259(v=vs.85).aspx
HRESULT FdfbFilter::CopyBufToTextureStaging(ID3D11Texture2D *pTextureStaging, BYTE *pBufInNv12, INT iWidth, INT iHeight)
{
    HRESULT hr = S_OK;
    BYTE *dst;
    UINT pitch;
    D3D11_MAPPED_SUBRESOURCE lr;

    ZeroMemory(&lr, sizeof(lr));
    hr = m_pD3d11DeviceContext->Map(pTextureStaging, 0, D3D11_MAP_WRITE, 0, &lr);
    if (FAILED(hr))
    {
        PLOG("ERROR: map staging for input \n");
        goto END;
    }

    dst   = (BYTE*)lr.pData;
    pitch = lr.RowPitch;

    if (FDFB_INPUT_FORMAT == FOURCC_YUY2)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(dst, pBufInNv12, iWidth * 2);
	        dst += pitch;
	        pBufInNv12 += iWidth * 2;
        }
    }
    else if (FDFB_INPUT_FORMAT == FOURCC_NV12)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(dst, pBufInNv12, iWidth);
	        dst += pitch;
	        pBufInNv12 += iWidth;
        }

        // U,V plane
        for (INT cnt = 0; cnt < iHeight / 2; cnt++)
        {
            memcpy(dst, pBufInNv12, iWidth);
            dst += pitch;
            pBufInNv12 += iWidth;
        }
    }
    else
    {
        PLOG("Failed to CopyBufToTextureStaging() due to unsupported format \n");
        hr = E_FAIL;
        goto END;
    }

    m_pD3d11DeviceContext->Unmap(pTextureStaging, 0);
END:
    return hr;
}

HRESULT FdfbFilter::CopyTextureStagingToBuf(BYTE *pBufOut, ID3D11Texture2D *pTextureStaging, INT iWidth, INT iHeight)
{
    HRESULT hr = S_OK;
    UINT pitch;
    D3D11_MAPPED_SUBRESOURCE lr;

    ZeroMemory(&lr, sizeof(lr));
    hr = m_pD3d11DeviceContext->Map(pTextureStaging, 0, D3D11_MAP_READ, 0, &lr);
    if (FAILED(hr))
    {
        PLOG("ERROR: map staging for output \n");
        goto END;
    }

    // output the result
    BYTE *src = (BYTE*)lr.pData;
    pitch = lr.RowPitch;

    if (FDFB_OUTPUT_FORMAT == FOURCC_YUY2)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(pBufOut, src, iWidth*2);
            pBufOut += iWidth*2;
            src += pitch;
        }
    }
    else if (FDFB_OUTPUT_FORMAT == FOURCC_NV12)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(pBufOut, src, iWidth);
            pBufOut += iWidth;
            src += pitch;
        }

        // U,V plane
        for (INT cnt = 0; cnt < iHeight / 2; cnt++)
        {
            memcpy(pBufOut, src, iWidth);
            pBufOut += iWidth;
            src += pitch;
        }
    }
    else
    {
        PLOG("Failed to CopyBufToTextureStaging() due to unsupported format \n");
        hr = E_FAIL;
        goto END;
    }

    m_pD3d11DeviceContext->Unmap(pTextureStaging, 0);
    if (FAILED(hr))
    {
        PLOG("ERROR: unmap staging for output \n");
        goto END;
    }

END:
    return hr;
}

HRESULT FdfbFilter::VideoProcessorBlt(CComPtr<ID3D11VideoProcessorOutputView> pOutputView,  CComPtr<ID3D11VideoProcessorInputView> pInputView)
{
    HRESULT hr = S_OK;

    // blt
    D3D11_VIDEO_PROCESSOR_STREAM StreamData;
    ZeroMemory(&StreamData, sizeof(StreamData));
    StreamData.Enable = TRUE;
    StreamData.OutputIndex = 0;
    StreamData.InputFrameOrField = 0;
    StreamData.PastFrames = 0;
    StreamData.FutureFrames = 0;
    StreamData.ppPastSurfaces = NULL;
    StreamData.ppFutureSurfaces = NULL;
    StreamData.pInputSurface = pInputView;
    StreamData.ppPastSurfacesRight = NULL;
    StreamData.ppFutureSurfacesRight = NULL;

    hr = m_pD3d11VideoContext->VideoProcessorBlt(m_pD3d11VideoProcessor, pOutputView, 0, 1, &StreamData);
    if (FAILED(hr))
    {
        PLOG("ERROR: VideoProcessorBlt \n");
        goto END;
    }

END:
    return hr;
}

HRESULT FdfbFilter::VideoProcessorWait()
{
    VPE_FUNCTION vprepParams = {0};
    VPE_GET_STATUS_PARAMS preprocStatusQuery = {0};
    VPE_STATUS_PARAM preprocStatusQueryBuffer = {0};
    BOOL blFinished = FALSE;
    HRESULT hr = S_OK;

    preprocStatusQuery.StatusCount = 1;
    preprocStatusQuery.pStatusBuffer = &preprocStatusQueryBuffer;

    vprepParams.Function = VPE_FN_GET_STATUS_PARAM;
    vprepParams.pGetStatusParams = &preprocStatusQuery;

    for (UINT i = 0; i < 10; i++)
    {
        hr = m_pD3d11VideoContext->VideoProcessorGetOutputExtension(m_pD3d11VideoProcessor, m_pVprepOperationGUID, sizeof(vprepParams), &vprepParams);
        if (FAILED(hr))
        {
            PLOG("ERROR: Status query failed \n");
            goto END;
        }
        if (preprocStatusQuery.pStatusBuffer->Status == VPREP_STATUS_OK)
        {
            blFinished = TRUE;
            break;
        }
        else if (preprocStatusQuery.pStatusBuffer->Status == VPREP_STATUS_ERROR)
        {
            PLOG("ERROR: Status query is VPREP_STATUS_ERROR \n");
            hr = E_FAIL;
            goto END;
        }
        Sleep(5); // more or less arbitrary sleep
    }

    if (!blFinished)
    {
        // log, not return E_FAIL
        PLOG("Status query wait too long and FDFB is still not blFinished \n");
    }
END:
    return hr;
}

HRESULT FdfbFilter::CreateDevice()
{
    HRESULT hr = S_OK;

    D3D_FEATURE_LEVEL __FeatureLevel;
    D3D_FEATURE_LEVEL pFeatureLevels[] = { D3D_FEATURE_LEVEL_11_0 };
    INT flCount = sizeof(pFeatureLevels) / sizeof(D3D_FEATURE_LEVEL);
    IDXGIFactory* dxgiFactory   = NULL;
    UINT          adapterIdx    = 0;
    IDXGIAdapter* pAdapter      = NULL;
    IDXGIAdapter* pIntelAdapter = NULL;
    D3D_DRIVER_TYPE eDriverType = D3D_DRIVER_TYPE_HARDWARE;

    PLOG(">>>=== FdfbFilter::CreateDevice() \n");

    hr = CreateDXGIFactory(__uuidof(IDXGIFactory), reinterpret_cast<void**>(&dxgiFactory));
    GOTO_END_FAILED(hr, "Failed to CreateDXGIFactory() \n");
    GOTO_END_NULL(dxgiFactory, "Failed to CreateDXGIFactory() \n");

    for (adapterIdx = 0; dxgiFactory->EnumAdapters(adapterIdx, &pAdapter) != DXGI_ERROR_NOT_FOUND; adapterIdx++)
    {
        DXGI_ADAPTER_DESC adapterDesc;
        hr = pAdapter->GetDesc(&adapterDesc);
        if (FAILED(hr))
        {
            PLOG("Failed to pAdapter->GetDesc() \n");
            pAdapter->Release();
            pAdapter = NULL;
            continue;
        }

        // eDriverType must be D3D_DRIVER_TYPE_UNKNOWN if pIntelAdapter isn't NULL
        // https://msdn.microsoft.com/en-us/library/windows/desktop/ff476082(v=vs.85).aspx
        if (wcsstr(adapterDesc.Description, L"Intel"))
        {
            pIntelAdapter = pAdapter;
            eDriverType   = D3D_DRIVER_TYPE_UNKNOWN;
            break;
        }
        else
        {
            pAdapter->Release();
            pAdapter = NULL;
        }
    }

    if (pIntelAdapter)
    {
        PLOG("====== FdfbFilter::CreateDevice() Found Intel graphic adapter \n");
    } else {
        PLOG("====== Warning: cannot find Intel graphic adapter, D3D11CreateDevice() with NULL pIntelAdapter \n");
    }

    hr = D3D11CreateDevice(pIntelAdapter, eDriverType, NULL, D3D11_CREATE_DEVICE_BGRA_SUPPORT,
        pFeatureLevels, flCount, D3D11_SDK_VERSION, &m_pD3d11Device, &__FeatureLevel, &m_pD3d11DeviceContext);
    GOTO_END_FAILED(hr, "Failed toD3D11CreateDevice() \n");

END:
    if (pIntelAdapter)
    {
        pIntelAdapter ->Release();
        pIntelAdapter = NULL;
    }

    if (dxgiFactory)
    {
        dxgiFactory->Release();
        dxgiFactory = NULL;
    }

    PLOG("<<<=== FdfbFilter::CreateDevice() hr=[%08x] \n", hr);
    return hr;
}

HRESULT FdfbFilter::CreateVideoProcessor(INT iWidth, INT iHeight)
{
    HRESULT hr;
    D3D11_VIDEO_PROCESSOR_CAPS caps;
    UINT flags;
    D3D11_VIDEO_PROCESSOR_CONTENT_DESC desc;
    PLOG(">>>=== FdfbFilter::CreateVideoProcessor iWidth=[%d] iHeight=[%d] \n", iWidth, iHeight);

    hr = m_pD3d11Device->QueryInterface(__uuidof(ID3D11VideoDevice), (void**)&m_pD3d11VideoDevice);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11Device->QueryInterface() ID3D11VideoDevice \n");

    hr = m_pD3d11DeviceContext->QueryInterface(__uuidof(ID3D11VideoContext), (void**)&m_pD3d11VideoContext);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11DeviceContext->QueryInterface() ID3D11VideoContext \n");

    ZeroMemory(&desc, sizeof(D3D11_VIDEO_PROCESSOR_CONTENT_DESC));
    desc.InputFrameFormat = D3D11_VIDEO_FRAME_FORMAT_PROGRESSIVE;
    desc.InputFrameRate.Numerator = desc.OutputFrameRate.Numerator = 30;
    desc.InputFrameRate.Denominator = desc.OutputFrameRate.Denominator = 1;
    desc.InputWidth = desc.OutputWidth = iWidth;
    desc.InputHeight = desc.OutputHeight = iHeight;
    desc.Usage = D3D11_VIDEO_USAGE_PLAYBACK_NORMAL;

    hr = m_pD3d11VideoDevice->CreateVideoProcessorEnumerator(&desc, &m_pD3d11VideoProcessorEnum);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoDevice->CreateVideoProcessorEnumerator() \n");

    hr = m_pD3d11VideoProcessorEnum->GetVideoProcessorCaps(&caps);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoProcessorEnum->GetVideoProcessorCaps() \n");

    m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat(format_fourcc_to_dxgi(FDFB_INPUT_FORMAT), &flags);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat() FDFB_INPUT_FORMAT \n");
    GOTO_END_FALSE((flags & D3D11_VIDEO_PROCESSOR_FORMAT_SUPPORT_INPUT), "Failed to CreateVideoProcessor() due to unsupported format \n");

    m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat(format_fourcc_to_dxgi(FDFB_OUTPUT_FORMAT), &flags);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat() FDFB_OUTPUT_FORMAT \n");
    GOTO_END_FALSE((flags & D3D11_VIDEO_PROCESSOR_FORMAT_SUPPORT_OUTPUT), "Failed to CreateVideoProcessor() due to unsupported format \n");

    hr = m_pD3d11VideoDevice->CreateVideoProcessor(m_pD3d11VideoProcessorEnum, 0, &m_pD3d11VideoProcessor);
    GOTO_END_NULL(m_pD3d11VideoProcessor, "Failed to m_pD3d11VideoDevice->CreateVideoProcessor() due to null pointer \n");
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoDevice->CreateVideoProcessor() \n");

END:
    PLOG("<<<=== FdfbFilter::CreateVideoProcessor hr=[0x%08x] \n", hr);
    return hr;
}

HRESULT FdfbFilter::SetParams()
{
    HRESULT hr;
    BOOL blFBRedLipStrength = TRUE;
    UINT FBRedLipStrength = 100;
    VPE_FUNCTION functionParams;
    VPE_VERSION SetVersion;
    BOOL bIsKeyExisted;
    DWORD dwFaceCriteria;
    PLOG(">>>=== FdfbFilter::SetParams() \n");

    ZeroMemory(&functionParams, sizeof(functionParams));
    ZeroMemory(&SetVersion, sizeof(SetVersion));

    if (!m_pD3d11VideoProcessor)
    {
        PLOG("Failed to FdfbFilter::SetParams() due to null pointer m_pD3d11VideoProcessor \n");
        hr = E_POINTER;
        goto END;
    }

    {
        // set VPE version = 3, get FDFB caps
        functionParams.Function = VPE_FN_SET_VERSION_PARAM;
        functionParams.pVersion = &SetVersion;
        SetVersion.Version = 3;
        hr = m_pD3d11VideoContext->VideoProcessorSetOutputExtension(m_pD3d11VideoProcessor, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorSetOutputExtension() VPE_FN_SET_VERSION_PARAM, graphics driver didn't support FDFB hr=[0x%08x]\n", hr);
            goto END;
        }

        VPE_MODE_PARAM ModeParams;
        ZeroMemory(&ModeParams, sizeof(ModeParams));

        functionParams.Function = VPE_FN_MODE_PARAM;
        functionParams.pModeParam = &ModeParams;
        functionParams.pModeParam->Mode = VPE_MODE_FDFB;
        hr = m_pD3d11VideoContext->VideoProcessorSetOutputExtension(m_pD3d11VideoProcessor, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorSetOutputExtension() VPE_MODE_FDFB, graphics driver didn't support FDFB \n");
            goto END;
        }

        // FDFB caps check
        VPE_FDFB_CAPS_PARAM CapsParam;
        ZeroMemory(&CapsParam, sizeof(CapsParam));
        
        CapsParam.nParamName = VPE_FDFB_CAPS_FB_FEATURE_SUPPORT;
        CapsParam.nDataSize  = sizeof(VPE_FB_FEATURE_SURPORT);

        functionParams.Function = VPE_FN_FDFB_QUERY_CAPS;
        functionParams.pFDFBParam = (VPE_FDFB_PARAM*)(&CapsParam);
        hr = m_pD3d11VideoContext->VideoProcessorGetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorGetStreamExtension() VPE_FN_FDFB_QUERY_CAPS \n");
            goto END;
        }
        else
        {
            VPE_FB_FEATURE_SURPORT flag = *(VPE_FB_FEATURE_SURPORT*)(&CapsParam.Data);
        }

        // start of FDFB parameters
        VPE_FDFB_PARAM FDFBParam;
        ZeroMemory(&functionParams, sizeof(functionParams));
        functionParams.Function = VPE_FN_FDFB_PARAM;
        functionParams.pFDFBParam = &FDFBParam;
        ZeroMemory(&FDFBParam, sizeof(FDFBParam));
        FDFBParam.nDataSize = sizeof(UINT) > sizeof(BYTE)* 4 ? sizeof(UINT) : sizeof(BYTE)* 4;
        FDFBParam.nParamName = VPE_FDP_MODE;
        FDFBParam.nValue = VPE_FDP_MODE_CONFIG;
        hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if FAILED(hr)
        {
            PLOG("Failed to VideoProcessorSetStreamExtension() VPE_FN_FDFB_PARAM \n");
            goto END;
        }

        // mode
        FDFBParam.nParamName = VPE_FDFB_MODE;
        FDFBParam.nValue = m_cFDFBMode;
        hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorSetStreamExtension() VPE_FDFB_MODE! \n");
            goto END;
        }

        // we don't expose this setting to registrykey since it cannot be applied on the fly.
        ZeroMemory(&FDFBParam, sizeof(FDFBParam));
        FDFBParam.nDataSize  = sizeof(UINT) > sizeof(BYTE)* 4 ? sizeof(UINT) : sizeof(BYTE)* 4;
        FDFBParam.nParamName = VPE_FDP_INIT_FACE_SELECTION_CRITERIA;
        bIsKeyExisted        = GetDwordRegKey("VPE_FDP_INIT_FACE_SELECTION_CRITERIA", dwFaceCriteria);
        if ((!bIsKeyExisted) || (dwFaceCriteria >= VPE_FD_FACE_SORT_CRITERIA_END))
        {
            dwFaceCriteria = VPE_FD_FACE_SORT_BY_MAX_SIZE;
        }
        FDFBParam.nValue = dwFaceCriteria;

        PLOG("Enabled feature=[VPE_FDP_INIT_FACE_SELECTION_CRITERIA] nValue=[%d]=[%s] \n",
            FDFBParam.nValue, (FDFBParam.nValue == VPE_FD_FACE_SORT_BY_MAX_SIZE) ? 
            "VPE_FD_FACE_SORT_BY_MAX_SIZE" : "VPE_FD_FACE_SORT_BY_CENTER");

        hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorSetStreamExtension() VPE_FDP_INIT_FACE_SELECTION_CRITERIA! \n");
            goto END;
        }

        SetFeatureParams();

        // end of FDFB parameters
        ZeroMemory(&FDFBParam, sizeof(FDFBParam));
        FDFBParam.nDataSize = sizeof(UINT) > sizeof(BYTE)* 4 ? sizeof(UINT) : sizeof(BYTE)* 4;
        FDFBParam.nParamName = VPE_FDP_MODE;
        FDFBParam.nValue = VPE_FDP_MODE_EXEC;
        hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorSetStreamExtension() VPE_FDP_MODE \n");
            goto END;
        }

        // enable status query
        VPE_SET_STATUS_PARAM preprocStatus;
        ZeroMemory(&preprocStatus, sizeof(preprocStatus));
        preprocStatus.bEnableStatusCollection = 1;

        ZeroMemory(&functionParams, sizeof(functionParams));
        functionParams.Function = VPE_FN_SET_STATUS_PARAM;
        functionParams.pSetStatusParam = &preprocStatus;

        hr = m_pD3d11VideoContext->VideoProcessorSetOutputExtension(m_pD3d11VideoProcessor, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
        if (FAILED(hr))
        {
            PLOG("Failed to VideoProcessorSetOutputExtension() VPE_FN_SET_STATUS_PARAM \n");
            goto END;
        }
    }

END:
    PLOG("<<<=== FdfbFilter::SetParams() \n");
    return hr;
}


HRESULT FdfbFilter::Init(ID3D11Device *pD3d11Device, ID3D11DeviceContext *pD3d11DeviceContext, FOURCC eFourCcOut, FOURCC eFourCcIn, INT iWidth, INT iHeight)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== FdfbFilter::Init() \n");

    CAutoLock cObjectLock(&m_MutexLock);

    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextureIn               = NULL;
    m_pTextOut                 = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;

    GOTO_END_NULL(pD3d11Device, "Failed to FdfbFilter::Init() due to null pointer \n");
    GOTO_END_NULL(pD3d11DeviceContext, "Failed to FdfbFilter::Init() due to null pointer \n");
                               
    if (eFourCcIn != FOURCC_NV12 && eFourCcIn != FOURCC_YUY2) {
        PLOG("Failed to Init(), fdfb doesn't support this input format \n");
        hr = E_INVALIDARG;
        goto END;
    }

    if (eFourCcOut != FOURCC_NV12 && eFourCcOut != FOURCC_YUY2) {
        PLOG("Failed to Init(), fdfb doesn't support this output format \n");
        hr = E_INVALIDARG;
        goto END;
    }

    m_eFourCcIn  = eFourCcIn;
    m_eFourCcOut = eFourCcOut;

#if 1
    m_pD3d11Device        = pD3d11Device;
    m_pD3d11DeviceContext = pD3d11DeviceContext;
#else
    hr = CreateDevice();
    GOTO_END_FAILED(hr, "Failed to CreateDevice() \n");
#endif

    hr = CreateVideoProcessor(iWidth, iHeight);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessor() \n");

    hr = CreateStagingTexture(m_pTextureStagingIn, iWidth, iHeight, format_fourcc_to_dxgi(eFourCcIn));
    GOTO_END_FAILED(hr, "Failed to CreateStagingTexture() for m_pTextureStagingIn \n");

    hr = CreateStagingTexture(m_pTextureStagingOut, iWidth, iHeight, format_fourcc_to_dxgi(eFourCcOut));
    GOTO_END_FAILED(hr, "Failed to CreateStagingTexture() for m_pTextureStagingOut \n");

    hr = CreateInputFrame(m_pTextureIn, m_pViewIn, iWidth, iHeight, format_fourcc_to_dxgi(eFourCcIn), 0);
    GOTO_END_FAILED(hr, "Failed to CreateInputFrame() \n");

    hr = CreateOutputFrame(m_pTextOut, m_pViewOut, iWidth, iHeight, format_fourcc_to_dxgi(eFourCcOut));
    GOTO_END_FAILED(hr, "Failed to CreateOutputFrame() \n");

    hr = FeatureInit(&m_FeatureList);
    GOTO_END_FAILED(hr, "Failed to FeatureInit() \n");

#if 0 // UWP (Metro) apps doesn't have write permission, so we cannot automatically create registry keys if not exist.
    InitFeatureReg(&m_FeatureList);
    GOTO_END_FAILED(hr, "Failed to InitFeatureReg() \n");
#endif

    hr = SetParams();
    GOTO_END_FAILED(hr, "Failed to SetParams(), possibly because device doesn't support fdfb feature \n");

END:
    PLOG("<<<=== FdfbFilter::Init() \n");
    return hr;
}

HRESULT FdfbFilter::Destroy()
{
    HRESULT hr = S_OK;
    PLOG(">>>=== FdfbFilter::Destroy() \n");

    CAutoLock cObjectLock(&m_MutexLock);

    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextureIn               = NULL;
    m_pTextOut                 = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;

    PLOG("<<<=== FdfbFilter::Destroy() \n");
    return hr;
}

HRESULT FdfbFilter::Blt(BYTE *pFrameBufOut, BYTE *pFrameBufIn, INT iWidth, INT iHeight, BOOL blRegKeyHasBeenChanged)
{
    HRESULT hr = S_OK;
    CAutoLock cObjectLock(&m_MutexLock);

    if (blRegKeyHasBeenChanged)
    {
        hr = SetParams();
        GOTO_END_FAILED(hr, "Failed to SetParams() \n");
    }

    // smart pointer for auto release these resource
    hr = CopyBufToTextureStaging(m_pTextureStagingIn, pFrameBufIn, iWidth, iHeight);
    GOTO_END_FAILED(hr, "Failed to CopyBufToTextureStaging() \n");

    // copy staging to input
    m_pD3d11DeviceContext->CopyResource(m_pTextureIn, m_pTextureStagingIn); // we must need a temp texture for copy CPU memory to GPU memory.

    hr = VideoProcessorBlt(m_pViewOut, m_pViewIn);
    GOTO_END_FAILED(hr, "Failed to VideoProcessorBlt() \n");

    hr = VideoProcessorWait();
    GOTO_END_FAILED(hr, "Failed to VideoProcessorWait() \n");

    // copy output to staging
    m_pD3d11DeviceContext->CopyResource(m_pTextureStagingOut, m_pTextOut);  // we must need a temp texture for GPU memory to CPU memory.
    hr = CopyTextureStagingToBuf(pFrameBufOut, m_pTextureStagingOut, iWidth, iHeight);
    GOTO_END_FAILED(hr, "Failed to CopyTextureStagingToBuf() \n");

END:
    return hr;
}


HRESULT FdfbFilter::BltText(ID3D11Texture2D *pTextIn, BOOL blRegKeyHasBeenChanged)
{
    HRESULT hr = S_OK;
    D3D11_VIDEO_PROCESSOR_INPUT_VIEW_DESC inputViewDesc   = {m_eFourCcIn, D3D11_VPIV_DIMENSION_TEXTURE2D, 0};
    D3D11_VIDEO_PROCESSOR_OUTPUT_VIEW_DESC outputViewDesc = {D3D11_VPOV_DIMENSION_TEXTURE2D, 0};
    ID3D11VideoProcessorInputView *pVpInputView = NULL;
    ID3D11VideoProcessorOutputView *pOutputView = NULL;
    D3D11_VIDEO_PROCESSOR_STREAM StreamData = {0};

    CAutoLock cObjectLock(&m_MutexLock);

    if (blRegKeyHasBeenChanged)
    {
        hr = SetParams();
        GOTO_END_FAILED(hr, "Failed to SetParams() \n");
    }

    m_pD3d11DeviceContext->CopyResource(m_pTextureIn, pTextIn);

    hr = m_pD3d11VideoDevice->CreateVideoProcessorInputView(m_pTextureIn, m_pD3d11VideoProcessorEnum, &inputViewDesc, &pVpInputView);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessorInputView() \n");

    hr = m_pD3d11VideoDevice->CreateVideoProcessorOutputView(m_pTextOut, m_pD3d11VideoProcessorEnum, &outputViewDesc, &pOutputView);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessorOutputView() \n");

    StreamData.Enable        = TRUE;
    StreamData.pInputSurface = pVpInputView;

    hr = VideoProcessorBlt(pOutputView, pVpInputView);
    GOTO_END_FAILED(hr, "Failed to VideoProcessorBlt() \n");

    hr = VideoProcessorWait();
    GOTO_END_FAILED(hr, "Failed to VideoProcessorWait() \n");

END:
    if (pVpInputView)
    {
        pVpInputView->Release();
    }
    if (pOutputView)
    {
        pOutputView->Release();
    }
    return hr;
}


ID3D11Texture2D* FdfbFilter::GetBltTextOut()
{
    return m_pTextOut;
};


HRESULT FeatureInit(FDFB_FILTER_FEATURE_LIST *pFeatureList)
{
    HRESULT hr = S_OK;

    if (!pFeatureList || pFeatureList->uiNum < FDFB_FEATURE_NUM_MAX)
    {
        return E_POINTER;
    }

    pFeatureList->paEntry[0].blActive = TRUE;
    pFeatureList->paEntry[0].eParam   = VPE_FBP_RED_LIP_STRENGTH;
    pFeatureList->paEntry[0].uiValue  = FDFB_FEATURE_VALUE_MAX;
    sprintf_s(pFeatureList->paEntry[0].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_RED_LIP_STRENGTH");

    pFeatureList->paEntry[1].blActive = TRUE;
    pFeatureList->paEntry[1].eParam   = VPE_FBP_BIG_EYE_STRENGTH;
    pFeatureList->paEntry[1].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[1].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_BIG_EYE_STRENGTH");

    pFeatureList->paEntry[2].blActive = TRUE;
    pFeatureList->paEntry[2].eParam   = VPE_FBP_SLIM_FACE_STRENGTH;
    pFeatureList->paEntry[2].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[2].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_SLIM_FACE_STRENGTH");

    pFeatureList->paEntry[3].blActive = TRUE;
    pFeatureList->paEntry[3].eParam   = VPE_FBP_HAPPY_FACE_STRENGTH;
    pFeatureList->paEntry[3].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[3].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_HAPPY_FACE_STRENGTH");

    pFeatureList->paEntry[4].blActive = TRUE;
    pFeatureList->paEntry[4].eParam   = VPE_FBP_CUTE_NOSE_STRENGTH;
    pFeatureList->paEntry[4].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[4].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_CUTE_NOSE_STRENGTH");

    pFeatureList->paEntry[5].blActive = TRUE;
    pFeatureList->paEntry[5].eParam   = VPE_FBP_SKIN_RESERVED1_STRENGTH;
    pFeatureList->paEntry[5].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[5].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_SKIN_RESERVED1_STRENGTH");

    pFeatureList->paEntry[6].blActive = TRUE;
    pFeatureList->paEntry[6].eParam   = VPE_FBP_SKIN_RESERVED2_STRENGTH;
    pFeatureList->paEntry[6].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[6].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_SKIN_RESERVED2_STRENGTH");

    pFeatureList->paEntry[7].blActive = TRUE;
    pFeatureList->paEntry[7].eParam   = VPE_FBP_SKIN_FOUNDATION_STRENGTH;
    pFeatureList->paEntry[7].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[7].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_SKIN_FOUNDATION_STRENGTH");

    pFeatureList->paEntry[8].blActive = TRUE;
    pFeatureList->paEntry[8].eParam   = VPE_FBP_SKIN_FOUNDATION_COLOR;
    pFeatureList->paEntry[8].uiValue  = FDFB_COLOR_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[8].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_SKIN_FOUNDATION_COLOR");

    pFeatureList->paEntry[9].blActive = TRUE;
    pFeatureList->paEntry[9].eParam   = VPE_FBP_SKIN_SMOOTHING_STRENGTH;
    pFeatureList->paEntry[9].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[9].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_SKIN_SMOOTHING_STRENGTH");

    pFeatureList->paEntry[10].blActive = TRUE;
    pFeatureList->paEntry[10].eParam   = VPE_FBP_CIRCLES_REMOVAL_STRENGTH;
    pFeatureList->paEntry[10].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[10].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_CIRCLES_REMOVAL_STRENGTH");

    pFeatureList->paEntry[11].blActive = TRUE;
    pFeatureList->paEntry[11].eParam   = VPE_FBP_EYE_BAGS_REMOVAL_STRENGTH;
    pFeatureList->paEntry[11].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[11].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_EYE_BAGS_REMOVAL_STRENGTH");

    pFeatureList->paEntry[12].blActive = TRUE;
    pFeatureList->paEntry[12].eParam   = VPE_FBP_EYE_WRINKLES_REMOVAL_STRENGTH;
    pFeatureList->paEntry[12].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[12].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_EYE_WRINKLES_REMOVAL_STRENGTH");

    pFeatureList->paEntry[13].blActive = TRUE;
    pFeatureList->paEntry[13].eParam   = VPE_FBP_BLUSH_STRENGTH;
    pFeatureList->paEntry[13].uiValue  = FDFB_FEATURE_VALUE_MIN;
    sprintf_s(pFeatureList->paEntry[13].szKeyName, FDFB_FEATURE_KEYNAME_SIZE, "VPE_FBP_BLUSH_STRENGTH");

    return hr;
}

HRESULT InitFeatureReg(FDFB_FILTER_FEATURE_LIST *pFeatureList)
{
    HRESULT hr          = S_OK;
    DWORD   dwFaceCount = 0;

    if (!pFeatureList || pFeatureList->uiNum < FDFB_FEATURE_NUM_MAX)
    {
        return E_POINTER;
    }

    if (!GetDwordRegKey("VPE_FBP_MAX_FACE_COUNT", dwFaceCount))
    {
        SetDwordRegKey("VPE_FBP_MAX_FACE_COUNT", FDFB_FACE_COUNT_MAX);
    }

    for (UINT uiIdx = 0; uiIdx < pFeatureList->uiNum; uiIdx++)
    {
        if (pFeatureList->paEntry[uiIdx].blActive)
        {
            DWORD               dwVal      = 0;
            CHAR                *szKeyName = pFeatureList->paEntry[uiIdx].szKeyName;
            VPE_FDFB_PARAM_ENUM eParam     = pFeatureList->paEntry[uiIdx].eParam;

            if (eParam == VPE_FBP_SKIN_FOUNDATION_COLOR)
            {
                if (!GetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_Y", dwVal))
                {
                    SetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_Y", 0);
                }
                if (!GetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_U", dwVal))
                {
                    SetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_U", 0);
                }
                if (!GetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_V", dwVal))
                {
                    SetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_V", 0);
                }

                continue;
            }

            if (!GetDwordRegKey(szKeyName, dwVal))
            {
                SetDwordRegKey(szKeyName, pFeatureList->paEntry[uiIdx].uiValue);
            }
        }
    }

    return hr;
}

HRESULT ReadFeatureReg(FDFB_FILTER_FEATURE_LIST *pFeatureList)
{
    HRESULT hr = S_OK;
    DWORD dwVal = 0;

    if (!pFeatureList || pFeatureList->uiNum < FDFB_FEATURE_NUM_MAX)
    {
        return E_POINTER;
    }

    for (UINT uiIdx = 0; uiIdx < pFeatureList->uiNum; uiIdx++)
    {
        DWORD               dwVal      = 0;
        CHAR                *szKeyName = pFeatureList->paEntry[uiIdx].szKeyName;
        VPE_FDFB_PARAM_ENUM eParam     = pFeatureList->paEntry[uiIdx].eParam;

        if (eParam == VPE_FBP_SKIN_FOUNDATION_COLOR)
        {
            BOOL  bIsKeyY, bIsKeyU, bIsKeyV;
            DWORD dwFoundationColorY, dwFoundationColorU, dwFoundationColorV;
            
            bIsKeyY = GetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_Y", dwFoundationColorY);
            bIsKeyU = GetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_U", dwFoundationColorU);
            bIsKeyV = GetDwordRegKey("VPE_FBP_SKIN_FOUNDATION_COLOR_V", dwFoundationColorV);

            if (bIsKeyY && bIsKeyU && bIsKeyV &&
                dwFoundationColorY <= FDFB_COLOR_VALUE_MAX &&
                dwFoundationColorU <= FDFB_COLOR_VALUE_MAX &&
                dwFoundationColorV <= FDFB_COLOR_VALUE_MAX)
            {
                pFeatureList->paEntry[uiIdx].blActive = TRUE;
                pFeatureList->paEntry[uiIdx].uiValue = (dwFoundationColorY << 16) | (dwFoundationColorU << 8) | dwFoundationColorV;
            }
            else
            {
                pFeatureList->paEntry[uiIdx].blActive = FALSE;
            }

            continue;
        }


        if (GetDwordRegKey(szKeyName, dwVal))
        {
            if (FDFB_FEATURE_VALUE_MAX < dwVal) // disable the feature if its value is out of valid range 
            {
                pFeatureList->paEntry[uiIdx].blActive = FALSE;
            }
            else
            {
                pFeatureList->paEntry[uiIdx].blActive = TRUE;
                pFeatureList->paEntry[uiIdx].uiValue  = dwVal;
            }
        }
        else
        {
            pFeatureList->paEntry[uiIdx].blActive = FALSE;
        }
    }

    return hr;
}

// set feature params that can be applied on the fly.
HRESULT FdfbFilter::SetFeatureParams()
{
    HRESULT                hr;
    VPE_FUNCTION           functionParams;
    VPE_FDFB_PARAM         FDFBParam;
    DWORD                  dwFaceCount;
    BOOL                   bIsKeyExisted;
    VPE_FBP_FEATURE_ACTIVE FeatureActive;

    PLOG(">>>=== FdfbFilter::SetFeatureParams() \n");

    hr = ReadFeatureReg(&m_FeatureList);
    if (FAILED(hr))
    {
        PLOG("Failed to ReadFeatureReg() \n");
        goto END;
    }

    ZeroMemory(&functionParams, sizeof(functionParams));
    functionParams.Function   = VPE_FN_FDFB_PARAM;
    functionParams.pFDFBParam = &FDFBParam;

    ZeroMemory(&FDFBParam, sizeof(FDFBParam));
    FDFBParam.nDataSize  = sizeof(UINT) > sizeof(BYTE)* 4 ? sizeof(UINT) : sizeof(BYTE)* 4;
    FDFBParam.nParamName = VPE_FBP_MAX_FACE_COUNT;
    bIsKeyExisted        = GetDwordRegKey("VPE_FBP_MAX_FACE_COUNT", dwFaceCount);
    if ((!bIsKeyExisted) || (dwFaceCount >= FDFB_FACE_COUNT_MAX))
    {
        dwFaceCount = FDFB_FACE_COUNT_DEFAULT;
    }
    FDFBParam.nValue = dwFaceCount;
    PLOG("Enabled feature=[VPE_FBP_MAX_FACE_COUNT] nValue=[%d] \n", FDFBParam.nValue);

    hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
    if (FAILED(hr))
    {
        PLOG("Failed to VideoProcessorSetStreamExtension() VPE_FBP_MAX_FACE_COUNT! \n");
        goto END;
    }

    ZeroMemory(&FeatureActive, sizeof(FeatureActive));
    for (UINT uiIdx = 0; uiIdx < m_FeatureList.uiNum; uiIdx++)
    {
        VPE_FDFB_PARAM_ENUM eParam = m_FeatureList.paEntry[uiIdx].eParam;
        BOOL blActive =  m_FeatureList.paEntry[uiIdx].blActive;

        if (eParam == VPE_FBP_RED_LIP_STRENGTH && blActive)
        {
            FeatureActive.bFB_RED_LIP_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_BIG_EYE_STRENGTH && blActive)
        {
            FeatureActive.bFB_BIG_EYE_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_SLIM_FACE_STRENGTH && blActive)
        {
            FeatureActive.bFB_SLIM_FACE_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_HAPPY_FACE_STRENGTH && blActive)
        {
            FeatureActive.bFB_HAPPY_FACE_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_CUTE_NOSE_STRENGTH && blActive)
        {
            FeatureActive.bFB_CUTE_NOSE_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_SKIN_RESERVED1_STRENGTH && blActive)
        {
            FeatureActive.bFB_SKIN_RESERVED1_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_SKIN_RESERVED2_STRENGTH && blActive)
        {
            FeatureActive.bFB_SKIN_RESERVED2_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_SKIN_FOUNDATION_STRENGTH && blActive)
        {
            FeatureActive.bFB_SKIN_FOUNDATION_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_SKIN_FOUNDATION_COLOR && blActive)
        {
            FeatureActive.bFB_SKIN_FOUNDATION_Color_Active = TRUE;
        }
        if (eParam == VPE_FBP_SKIN_SMOOTHING_STRENGTH && blActive)
        {
            FeatureActive.bFB_SKIN_SMOOTHING_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_CIRCLES_REMOVAL_STRENGTH && blActive)
        {
            FeatureActive.bFB_CIRCLES_REMOVAL_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_EYE_BAGS_REMOVAL_STRENGTH && blActive)
        {
            FeatureActive.bFB_EYE_BAGS_REMOVAL_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_EYE_WRINKLES_REMOVAL_STRENGTH && blActive)
        {
            FeatureActive.bFB_EYE_WRINKLES_REMOVAL_STRENGTH_Active = TRUE;
        }
        if (eParam == VPE_FBP_BLUSH_STRENGTH && blActive)
        {
            FeatureActive.bFB_BLUSH_STRENGTH_Active = TRUE;
        }
    }

    ZeroMemory(&functionParams, sizeof(functionParams));
    ZeroMemory(&FDFBParam, sizeof(FDFBParam));

    functionParams.Function                    = VPE_FN_FDFB_PARAM;
    functionParams.pFDFBParam                  = &FDFBParam;
    functionParams.pFDFBParam->nDataSize       = sizeof(VPE_FBP_FEATURE_ACTIVE);
    functionParams.pFDFBParam->nParamName      = VPE_FBP_FEATURE_STATE;
    functionParams.pFDFBParam->FbFeatureActive = FeatureActive;

    hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
    if (FAILED(hr))
    {
        PLOG("Failed to VideoProcessorSetStreamExtension() VPE_FN_FDFB_PARAM \n");
        goto END;
    }

    BOOL blIsAnyFeatureActive;
    blIsAnyFeatureActive = (*(QWORD*)(&FeatureActive) != 0);
    if (!blIsAnyFeatureActive)
    {
        PLOG("Didn't enable any FDFB feature \n");
        goto END;
    }
        
    for (UINT uiIdx = 0; uiIdx < m_FeatureList.uiNum; uiIdx++)
    {
        if (m_FeatureList.paEntry[uiIdx].blActive)
        {
            PLOG("Enabled feature=[%s] uiValue=[%d] eParam=[%d] \n", m_FeatureList.paEntry[uiIdx].szKeyName, m_FeatureList.paEntry[uiIdx].uiValue, m_FeatureList.paEntry[uiIdx].eParam);
            VPE_FDFB_PARAM FDFBParam;

            ZeroMemory(&FDFBParam, sizeof(FDFBParam));
            ZeroMemory(&functionParams, sizeof(functionParams));

            functionParams.Function               = VPE_FN_FDFB_PARAM;
            functionParams.pFDFBParam             = &FDFBParam;
            functionParams.pFDFBParam->nDataSize  = MAX(sizeof(UINT), sizeof(BYTE)* 4);
            functionParams.pFDFBParam->nParamName = m_FeatureList.paEntry[uiIdx].eParam;
            functionParams.pFDFBParam->nValue     = m_FeatureList.paEntry[uiIdx].uiValue;

            hr = m_pD3d11VideoContext->VideoProcessorSetStreamExtension(m_pD3d11VideoProcessor, 0, m_pVprepOperationGUID, sizeof(functionParams), &functionParams);
            if (FAILED(hr))
            {
                PLOG("Failed to VideoProcessorSetStreamExtension() nParamName=[%d] \n", m_FeatureList.paEntry[uiIdx].eParam);
                goto END;
            }
        }
    }
END:
    PLOG("<<<=== FdfbFilter::SetFeatureParams() \n");
    return hr;
}
