// FDFBSample.cpp : Defines the entry point for the console application.
// This app should work on SKL, windows 10, with FDFB enabled drivers
// Change this line to use FB only or FDFB mode:
// VPE_FDFB_MODE_ENUM FDFBMode = VPE_FDFB_MODE_FB_ONLY;
// VPE_FDFB_MODE_ENUM FDFBMode = VPE_FDFB_MODE_FDFB;

#include <atlcomcli.h>
#include <d3d11.h>
#include "rotation.h"
#include "../utils_mft.h"


HRESULT RotationFilter::CreateStagingTexture(CComPtr<ID3D11Texture2D> &TextureStaging, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    D3D11_TEXTURE2D_DESC stagingDesc = {0};
    HRESULT hr;

    stagingDesc.Width            = iWidth;
    stagingDesc.Height           = iHeight;
    stagingDesc.MipLevels        = 1;
    stagingDesc.ArraySize        = 1;
    stagingDesc.Format           = format_fourcc_to_dxgi(eFourcc);
    stagingDesc.SampleDesc.Count = 1;
    stagingDesc.Usage            = D3D11_USAGE_STAGING;
    stagingDesc.CPUAccessFlags   = D3D11_CPU_ACCESS_WRITE | D3D11_CPU_ACCESS_READ;

    hr = m_pD3d11Device->CreateTexture2D(&stagingDesc, NULL, &TextureStaging);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateStagingTexture() eFourcc=[%d] \n", eFourcc);
    }

    return hr;
}

HRESULT RotationFilter::CreateInputFrame(CComPtr<ID3D11Texture2D> &pTextureIn, CComPtr<ID3D11VideoProcessorInputView> &pViewIn, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    D3D11_TEXTURE2D_DESC inputDesc = {0};
    D3D11_VIDEO_PROCESSOR_INPUT_VIEW_DESC inputViewDesc = {0};
    HRESULT hr;

    inputDesc.Width                    = iWidth;
    inputDesc.Height                   = iHeight;
    inputDesc.MipLevels                = 1;
    inputDesc.ArraySize                = 1;
    inputDesc.Format                   = format_fourcc_to_dxgi(eFourcc);
    inputDesc.SampleDesc.Count         = 1;
    inputDesc.Usage                    = D3D11_USAGE_DEFAULT;

    inputViewDesc.FourCC               = eFourcc;
    inputViewDesc.ViewDimension        = D3D11_VPIV_DIMENSION_TEXTURE2D;
    inputViewDesc.Texture2D.MipSlice   = 0;
    inputViewDesc.Texture2D.ArraySlice = 0;

    hr = m_pD3d11Device->CreateTexture2D(&inputDesc, NULL, &pTextureIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateTexture2D() input \n");
        goto END;
    }

    hr = m_pD3d11VideoDevice->CreateVideoProcessorInputView(pTextureIn, m_pD3d11VideoProcessorEnum, &inputViewDesc, &pViewIn);
    if (FAILED(hr))
    {
        PLOG("Failed to CreateVideoProcessorInputView() \n");
        goto END;
    }
END:
    return hr;
}

HRESULT RotationFilter::CreateOutputFrame(CComPtr<ID3D11Texture2D> &pTextureOut, CComPtr<ID3D11VideoProcessorOutputView> &pViewOut, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    D3D11_TEXTURE2D_DESC outputDesc = {0};
    HRESULT hr;
    D3D11_VIDEO_PROCESSOR_OUTPUT_VIEW_DESC outputViewDesc;

    outputDesc.Width            = iWidth;
    outputDesc.Height           = iHeight;
    outputDesc.MipLevels        = 1;
    outputDesc.ArraySize        = 1;
    outputDesc.Format           = format_fourcc_to_dxgi(eFourcc);
    outputDesc.SampleDesc.Count = 1;
    outputDesc.Usage            = D3D11_USAGE_DEFAULT;
    outputDesc.BindFlags        = D3D11_BIND_RENDER_TARGET;

    hr = m_pD3d11Device->CreateTexture2D(&outputDesc, NULL, &pTextureOut);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateTexture2D output \n");
        goto END;
    }

    ZeroMemory(&outputViewDesc, sizeof(outputViewDesc));
    outputViewDesc.ViewDimension = D3D11_VPOV_DIMENSION_TEXTURE2D;
    hr = m_pD3d11VideoDevice->CreateVideoProcessorOutputView(pTextureOut, m_pD3d11VideoProcessorEnum, &outputViewDesc, &pViewOut);
    if (FAILED(hr))
    {
        PLOG("ERROR: CreateVideoProcessorOutputView \n");
        goto END;
    }

END:
    return hr;
}

HRESULT RotationFilter::CopyBufToTextureStaging(ID3D11Texture2D *pStagingTexture, BYTE *pBufIn, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    HRESULT hr = S_OK;
    BYTE *dst;
    UINT pitch;
    D3D11_MAPPED_SUBRESOURCE lr;

    ZeroMemory(&lr, sizeof(lr));
    hr = m_pD3d11DeviceContext->Map(pStagingTexture, 0, D3D11_MAP_WRITE, 0, &lr);
    if (FAILED(hr))
    {
        PLOG("Failed to m_pD3d11DeviceContext->Map() \n");
        goto END;
    }

    dst   = (BYTE*)lr.pData;
    pitch = lr.RowPitch;

    if (eFourcc == FOURCC_YUY2)
    {
        // YUY2 plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth*2);
	        dst += pitch;
	        pBufIn += iWidth*2;
        }
    }
    else if (eFourcc == FOURCC_I420)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth);
	        dst += pitch;
	        pBufIn += iWidth;
        }
        // U plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth/2);
	        dst += pitch;
	        pBufIn += iWidth;
        }
        // V plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
	        memcpy(dst, pBufIn, iWidth/2);
	        dst += pitch;
	        pBufIn += iWidth;
        }
    }
    else if (eFourcc == FOURCC_NV12)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(dst, pBufIn, iWidth);
            dst += pitch;
            pBufIn += iWidth;
        }

        // U,V plane
        for (INT cnt = 0; cnt < iHeight / 2; cnt++)
        {
            memcpy(dst, pBufIn, iWidth);
            dst += pitch;
            pBufIn += iWidth;
        }
    }
    else
    {
        PLOG("Failed to CopyBufToTextureStaging due to invalid eFourcc=[%d] \n", eFourcc);
        hr = E_INVALIDARG;
        goto END;
    }

    m_pD3d11DeviceContext->Unmap(pStagingTexture, 0);
END:
    return hr;
}

HRESULT RotationFilter::CopyTextureStagingToBuf(BYTE *pBufNv12Out, ID3D11Texture2D *pStagingTexture, INT iWidth, INT iHeight, FOURCC eFourcc)
{
    HRESULT hr = S_OK;
    UINT pitch;
    D3D11_MAPPED_SUBRESOURCE lr;

    ZeroMemory(&lr, sizeof(lr));
    hr = m_pD3d11DeviceContext->Map(pStagingTexture, 0, D3D11_MAP_READ, 0, &lr);
    if (FAILED(hr))
    {
        PLOG("Failed to m_pD3d11DeviceContext->Map() \n");
        goto END;
    }

    // output the result
    BYTE *src = (BYTE*)lr.pData;
    pitch = lr.RowPitch;

    if (eFourcc == FOURCC_YUY2)
    {
        // YUY2 plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(pBufNv12Out, src, iWidth*2);
	        src += pitch;
	        pBufNv12Out += iWidth*2;
        }
    }
    else if (eFourcc == FOURCC_I420)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
	        memcpy(pBufNv12Out, src, iWidth);
	        src += pitch;
	        pBufNv12Out += iWidth;
        }
        // U plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
	        memcpy(pBufNv12Out, src, iWidth/2);
	        src += pitch;
	        pBufNv12Out += iWidth;
        }
        // V plane
        for (INT cnt = 0; cnt < iHeight/2; cnt++)
        {
	        memcpy(pBufNv12Out, src, iWidth/2);
	        src += pitch;
	        pBufNv12Out += iWidth;
        }
    }
    else if (eFourcc == FOURCC_NV12)
    {
        // Y plane
        for (INT cnt = 0; cnt < iHeight; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth);
            src += pitch;
            pBufNv12Out += iWidth;
        }

        // U,V plane
        for (INT cnt = 0; cnt < iHeight / 2; cnt++)
        {
            memcpy(pBufNv12Out, src, iWidth);
            src += pitch;
            pBufNv12Out += iWidth;
        }
    }
    else
    {
        PLOG("Failed to CopyTextureStagingToBuf() due to invalid eFourcc=[%d] \n", eFourcc);
        hr = E_INVALIDARG;
        goto END;
    }

    m_pD3d11DeviceContext->Unmap(pStagingTexture, 0);
    if (FAILED(hr))
    {
        PLOG("ERROR: unmap staging for output \n");
        goto END;
    }

END:
    return hr;
}

HRESULT RotationFilter::VideoProcessorBlt(CComPtr<ID3D11VideoProcessorOutputView> pViewOut,  CComPtr<ID3D11VideoProcessorInputView> pViewIn)
{
    HRESULT hr = S_OK;

    // blt
    D3D11_VIDEO_PROCESSOR_STREAM StreamData;

    ZeroMemory(&StreamData, sizeof(StreamData));
    StreamData.Enable                = TRUE;
    StreamData.OutputIndex           = 0;
    StreamData.InputFrameOrField     = 0;
    StreamData.PastFrames            = 0;
    StreamData.FutureFrames          = 0;
    StreamData.ppPastSurfaces        = NULL;
    StreamData.ppFutureSurfaces      = NULL;
    StreamData.pInputSurface         = pViewIn;
    StreamData.ppPastSurfacesRight   = NULL;
    StreamData.ppFutureSurfacesRight = NULL;

    m_pD3d11VideoContext->VideoProcessorSetStreamRotation(m_pD3d11VideoProcessor, 0, TRUE, D3D11_VIDEO_PROCESSOR_ROTATION_180);

    hr = m_pD3d11VideoContext->VideoProcessorBlt(m_pD3d11VideoProcessor, pViewOut, 0, 1, &StreamData);
    if (FAILED(hr))
    {
        PLOG("Failed to VideoProcessorBlt() \n");
        goto END;
    }

END:
    return hr;
}


HRESULT RotationFilter::CreateDevice()
{
    HRESULT hr = S_OK;

    D3D_FEATURE_LEVEL __FeatureLevel;
    D3D_FEATURE_LEVEL pFeatureLevels[] = { D3D_FEATURE_LEVEL_11_0 };
    INT flCount = sizeof(pFeatureLevels) / sizeof(D3D_FEATURE_LEVEL);
    IDXGIFactory* dxgiFactory   = NULL;
    UINT          adapterIdx    = 0;
    IDXGIAdapter* pAdapter      = NULL;
    IDXGIAdapter* pIntelAdapter = NULL;
    D3D_DRIVER_TYPE eDriverType = D3D_DRIVER_TYPE_HARDWARE;

    PLOG(">>>=== RotationFilter::CreateDevice() \n");

    CreateDXGIFactory(__uuidof(IDXGIFactory), reinterpret_cast<void**>(&dxgiFactory));
    if (FAILED(hr))
    {
        PLOG("Failed to CreateDXGIFactory() \n");
        goto END;
    }
    GOTO_END_NULL(dxgiFactory, "Failed to CreateDXGIFactory() \n");

    for (adapterIdx = 0; dxgiFactory->EnumAdapters(adapterIdx, &pAdapter) != DXGI_ERROR_NOT_FOUND; adapterIdx++)
    {
        DXGI_ADAPTER_DESC adapterDesc;
        hr = pAdapter->GetDesc(&adapterDesc);
        if (FAILED(hr))
        {
            PLOG("Failed to pAdapter->GetDesc() \n");
            pAdapter->Release();
            pAdapter = NULL;
            continue;
        }

        // eDriverType must be D3D_DRIVER_TYPE_UNKNOWN if pIntelAdapter isn't NULL
        // https://msdn.microsoft.com/en-us/library/windows/desktop/ff476082(v=vs.85).aspx
        if (wcsstr(adapterDesc.Description, L"Intel"))
        {
            pIntelAdapter = pAdapter;
            eDriverType   = D3D_DRIVER_TYPE_UNKNOWN;
            break;
        }
        else
        {
            pAdapter->Release();
            pAdapter = NULL;
        }
    }

    if (pIntelAdapter)
    {
        PLOG("====== RotationFilter::CreateDevice() Found Intel graphic adapter \n");
    } else {
        PLOG("====== Warning: cannot find Intel graphic adapter, D3D11CreateDevice() with NULL pIntelAdapter \n");
    }

    hr = D3D11CreateDevice(pIntelAdapter, eDriverType, NULL, D3D11_CREATE_DEVICE_BGRA_SUPPORT,
        pFeatureLevels, flCount, D3D11_SDK_VERSION, &m_pD3d11Device, &__FeatureLevel, &m_pD3d11DeviceContext);
    if (FAILED(hr))
    {
        PLOG("Failed to D3D11CreateDevice() hr=[%08x] \n", hr);
        goto END;
    }

END:
    if (pIntelAdapter)
    {
        pIntelAdapter ->Release();
        pIntelAdapter = NULL;
    }

    if (dxgiFactory)
    {
        dxgiFactory->Release();
        dxgiFactory = NULL;
    }

    PLOG("<<<=== RotationFilter::CreateDevice() hr=[%08x] \n", hr);
    return hr;
}


HRESULT RotationFilter::CreateVideoProcessor(INT iStreamW, INT iStreamH, INT iTargetW, INT iTargetH)
{
    HRESULT hr;
    D3D11_VIDEO_PROCESSOR_CAPS caps;
    UINT flags;
    D3D11_VIDEO_PROCESSOR_CONTENT_DESC desc;

    GOTO_END_NULL(m_pD3d11Device, "Failed to RotationFilter::CreateVideoProcessor() due to null pointer \n");
    GOTO_END_NULL(m_pD3d11DeviceContext, "Failed to RotationFilter::CreateVideoProcessor() due to null pointer \n");

    hr = m_pD3d11Device->QueryInterface(__uuidof(ID3D11VideoDevice), (void**)&m_pD3d11VideoDevice);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11Device->QueryInterface() ID3D11VideoDevice \n");

    hr = m_pD3d11DeviceContext->QueryInterface(__uuidof(ID3D11VideoContext), (void**)&m_pD3d11VideoContext);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11Device->QueryInterface() ID3D11VideoContext \n");

    ZeroMemory(&desc, sizeof(D3D11_VIDEO_PROCESSOR_CONTENT_DESC));
    desc.InputFrameFormat           = D3D11_VIDEO_FRAME_FORMAT_PROGRESSIVE;
    desc.InputFrameRate.Numerator   = desc.OutputFrameRate.Numerator   = 30;
    desc.InputFrameRate.Denominator = desc.OutputFrameRate.Denominator = 1;
    desc.InputWidth                 = iStreamW;
    desc.InputHeight                = iStreamH;
    desc.OutputWidth                = iTargetW;
    desc.OutputHeight               = iTargetH;
    desc.Usage                      = D3D11_VIDEO_USAGE_PLAYBACK_NORMAL;

    hr = m_pD3d11VideoDevice->CreateVideoProcessorEnumerator(&desc, &m_pD3d11VideoProcessorEnum);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessorEnumerator() \n");

    hr = m_pD3d11VideoProcessorEnum->GetVideoProcessorCaps(&caps);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoProcessorEnum->GetVideoProcessorCaps() \n");

    m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat(DXGI_FORMAT_NV12, &flags);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoProcessorEnum->CheckVideoProcessorFormat() \n");

    GOTO_END_FALSE((flags & D3D11_VIDEO_PROCESSOR_FORMAT_SUPPORT_INPUT), "Failed to RotationFilter::CreateVideoProcessor() due to unspoorted input format \n");
    GOTO_END_FALSE((flags & D3D11_VIDEO_PROCESSOR_FORMAT_SUPPORT_OUTPUT), "Failed to RotationFilter::CreateVideoProcessor() due to unspoorted output format \n");

    hr = m_pD3d11VideoDevice->CreateVideoProcessor(m_pD3d11VideoProcessorEnum, 0, &m_pD3d11VideoProcessor);
    GOTO_END_FAILED(hr, "Failed to m_pD3d11VideoDevice->CreateVideoProcessor() \n");

    m_pD3d11VideoContext->VideoProcessorSetStreamAutoProcessingMode(m_pD3d11VideoProcessor, 0, FALSE);

END:
    return hr;
}


HRESULT RotationFilter::Init(ID3D11Device *pD3d11Device, ID3D11DeviceContext *pD3d11DeviceContext, FOURCC eFourcc, INT iStreamW, INT iStreamH, INT iTargetW, INT iTargetH)
{
    HRESULT hr = S_OK;
    PLOG(">>>=== RotationFilter::Init() \n");
    CAutoLock cObjectLock(&m_MutexLock);

    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextIn                  = NULL;
    m_pTextOut                 = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;

    m_iStreamW                 = iStreamW;
    m_iStreamH                 = iStreamH;

    m_iTargetW                 = iTargetW;
    m_iTargetH                 = iTargetH;

    m_eFourccIn  = eFourcc;
    m_eFourccOut = eFourcc;
    
    GOTO_END_NULL(pD3d11Device, "Failed to Init() due to null pointer \n");
    GOTO_END_NULL(pD3d11DeviceContext, "Failed to Init() due to null pointer \n");
    GOTO_END_FALSE((eFourcc == FOURCC_YUY2 || eFourcc == FOURCC_NV12), "Failed to Init() due to unspoorted format \n");

#if 1
    m_pD3d11Device             = pD3d11Device;
    m_pD3d11DeviceContext      = pD3d11DeviceContext;
#else
    hr = CreateDevice();
    GOTO_END_FAILED(hr, "Failed to CreateDevice() \n");
#endif

    hr = CreateVideoProcessor(m_iStreamW, m_iStreamH, m_iTargetW, m_iTargetH);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessor() \n");

    hr = CreateStagingTexture(m_pTextureStagingIn, m_iStreamW, m_iStreamH, eFourcc);
    GOTO_END_FAILED(hr, "Failed to CreateStagingTexture() m_pTextureStagingIn \n");

    hr = CreateStagingTexture(m_pTextureStagingOut, m_iTargetW, m_iTargetH, eFourcc);
    GOTO_END_FAILED(hr, "Failed to CreateStagingTexture() m_pTextureStagingOut \n");

    hr = CreateInputFrame(m_pTextIn, m_pViewIn, m_iStreamW, m_iStreamH, eFourcc);
    GOTO_END_FAILED(hr, "Failed to CreateInputFrame() \n");

    hr = CreateOutputFrame(m_pTextOut, m_pViewOut, m_iTargetW, m_iTargetH, eFourcc);
    GOTO_END_FAILED(hr, "Failed to CreateOutputFrame() \n");

END:
    PLOG("<<<=== RotationFilter::Init() \n");
    return hr;
}

HRESULT RotationFilter::Destroy()
{
    PLOG(">>>=== RotationFilter::Destroy() \n");
    CAutoLock cObjectLock(&m_MutexLock);

    m_pTextureStagingIn        = NULL;
    m_pTextureStagingOut       = NULL;
    m_pTextIn                  = NULL;
    m_pTextOut                 = NULL;

    m_pViewIn                  = NULL;
    m_pViewOut                 = NULL;
                               
    m_pD3d11VideoProcessor     = NULL;
    m_pD3d11VideoProcessorEnum = NULL;
    m_pD3d11VideoContext       = NULL;
    m_pD3d11VideoDevice        = NULL;
                               
    PLOG("<<<=== RotationFilter::Destroy() \n");
    return S_OK;
}


HRESULT RotationFilter::SetStreamTargetRect(D3D11_VIDEO_PROCESSOR_ROTATION eRotateAngle)
{
    RECT rectStreamSrc = {0};
    RECT rectStreamDest = {0};
    RECT rectTargetDest = {0 , 0, m_iTargetW, m_iTargetH};

    if (eRotateAngle == D3D11_VIDEO_PROCESSOR_ROTATION_IDENTITY || eRotateAngle == D3D11_VIDEO_PROCESSOR_ROTATION_180)
    {
        if (m_iStreamW >= m_iTargetW)
        {
            rectStreamSrc.left   = (m_iStreamW-m_iTargetW)/2;
            rectStreamSrc.right  = (m_iStreamW+m_iTargetW)/2;
            rectStreamDest.left  = 0;
            rectStreamDest.right = m_iTargetW;
        }
        else
        {
            rectStreamSrc.left   = 0;
            rectStreamSrc.right  = m_iStreamW;
            rectStreamDest.left  = (m_iTargetW-m_iStreamW)/2;
            rectStreamDest.right = (m_iTargetW+m_iStreamW)/2;
        }

        if (m_iStreamH >= m_iTargetH)
        {
            rectStreamSrc.top    = (m_iStreamH-m_iTargetH)/2;
            rectStreamSrc.bottom = (m_iStreamH+m_iTargetH)/2;
            rectStreamDest.top   = 0;
            rectStreamDest.bottom = m_iTargetH;
        }
        else
        {
            rectStreamSrc.top     = 0;
            rectStreamSrc.bottom  = m_iStreamH;
            rectStreamDest.top    = (m_iTargetH-m_iStreamH)/2;
            rectStreamDest.bottom = (m_iTargetH+m_iStreamH)/2;
        }
    }
    else // D3D11_VIDEO_PROCESSOR_ROTATION_270 or D3D11_VIDEO_PROCESSOR_ROTATION_90
    {
        if (m_iStreamW >= m_iTargetH)
        {
            rectStreamSrc.left    = (m_iStreamW-m_iTargetH)/2;
            rectStreamSrc.right   = (m_iStreamW+m_iTargetH)/2;
            rectStreamDest.top    = 0;
            rectStreamDest.bottom = m_iTargetH;
        }
        else
        {
            rectStreamSrc.left    = 0;
            rectStreamSrc.right   = m_iStreamW;
            rectStreamDest.top    = (m_iTargetH-m_iStreamW)/2;
            rectStreamDest.bottom = (m_iTargetH+m_iStreamW)/2;
        }

        if (m_iStreamH >= m_iTargetW)
        {
            rectStreamSrc.top    = (m_iStreamH-m_iTargetW)/2;
            rectStreamSrc.bottom = (m_iStreamH+m_iTargetW)/2;
            rectStreamDest.left  = 0;
            rectStreamDest.right = m_iTargetW;
        }
        else
        {
            rectStreamSrc.top    = 0;
            rectStreamSrc.bottom = m_iStreamH;
            rectStreamDest.left  = (m_iTargetW-m_iStreamH)/2;
            rectStreamDest.right = (m_iTargetW+m_iStreamH)/2;
        }
    }

    m_pD3d11VideoContext->VideoProcessorSetStreamSourceRect(m_pD3d11VideoProcessor, 0, TRUE, &rectStreamSrc);
    m_pD3d11VideoContext->VideoProcessorSetStreamDestRect(m_pD3d11VideoProcessor, 0, TRUE, &rectStreamDest);
    m_pD3d11VideoContext->VideoProcessorSetOutputTargetRect(m_pD3d11VideoProcessor, TRUE, &rectTargetDest);

    return S_OK;
}


HRESULT RotationFilter::BltText(ID3D11Texture2D *pTextIn , D3D11_VIDEO_PROCESSOR_ROTATION eRotateAngle)
{
    HRESULT hr = S_OK;
    D3D11_VIDEO_PROCESSOR_INPUT_VIEW_DESC inputViewDesc   = {FOURCC_NV12, D3D11_VPIV_DIMENSION_TEXTURE2D, 0};
    D3D11_VIDEO_PROCESSOR_OUTPUT_VIEW_DESC outputViewDesc = {D3D11_VPOV_DIMENSION_TEXTURE2D, 0};
    ID3D11VideoProcessorInputView *pVpInputView = NULL;
    ID3D11VideoProcessorOutputView *pOutputView = NULL;
    D3D11_VIDEO_PROCESSOR_STREAM StreamData = {0};
    D3D11_VIDEO_COLOR VideoColor = {0};

    CAutoLock cObjectLock(&m_MutexLock);

    hr = m_pD3d11VideoDevice->CreateVideoProcessorInputView(pTextIn, m_pD3d11VideoProcessorEnum, &inputViewDesc, &pVpInputView);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessorInputView() \n");

    hr = m_pD3d11VideoDevice->CreateVideoProcessorOutputView(m_pTextOut, m_pD3d11VideoProcessorEnum, &outputViewDesc, &pOutputView);
    GOTO_END_FAILED(hr, "Failed to CreateVideoProcessorOutputView() \n");

    StreamData.Enable        = TRUE;
    StreamData.pInputSurface = pVpInputView;

    hr = SetStreamTargetRect(eRotateAngle);
    GOTO_END_FAILED(hr, "Failed to SetStreamTargetRect() \n");

    VideoColor.RGBA = {0};
    m_pD3d11VideoContext->VideoProcessorSetOutputBackgroundColor(m_pD3d11VideoProcessor, FALSE, &VideoColor);

    m_pD3d11VideoContext->VideoProcessorSetStreamRotation(m_pD3d11VideoProcessor, 0, TRUE, eRotateAngle);
    hr = m_pD3d11VideoContext->VideoProcessorBlt(m_pD3d11VideoProcessor, pOutputView, 0, 1, &StreamData);
    GOTO_END_FAILED(hr, "Failed to VideoProcessorBlt() \n");

END:
    if (pVpInputView)
    {
        pVpInputView->Release();
    }
    if (pOutputView)
    {
        pOutputView->Release();
    }

    return hr;
}


HRESULT RotationFilter::BltBuf(BYTE *pBufIn, UINT uiWidth, UINT uiHeight, D3D11_VIDEO_PROCESSOR_ROTATION eRotateAngle)
{
    HRESULT hr = S_OK;

    GOTO_END_NULL(pBufIn, "Failed to RotationFilter:Blt() due to null pointer \n");
    GOTO_END_FALSE(uiWidth == m_iStreamW, "Failed to RotationFilter:Blt() due to unvalid uiWidth=[%d] \n", uiWidth);
    GOTO_END_FALSE(uiHeight == m_iStreamH, "Failed to RotationFilter:Blt() due to unvalid uiHeight=[%d] \n", uiHeight);

    hr = CopyBufToTextureStaging(m_pTextureStagingIn, pBufIn, m_iStreamW, m_iStreamH, m_eFourccIn);
    GOTO_END_FAILED(hr, "Failed to RotationFilter::CopyBufToTextureStaging() \n");

    // copy staging to vpblt input, needing a temp texture for copying CPU to GPU memory.
    m_pD3d11DeviceContext->CopyResource(m_pTextIn, m_pTextureStagingIn);

    hr = BltText(m_pTextIn, eRotateAngle);
    GOTO_END_FAILED(hr, "Failed to RotationFilter::Blt() \n");

END:
    return hr;
}


ID3D11Texture2D* RotationFilter::GetBltTextOut()
{
    return m_pTextOut;
};


HRESULT RotationFilter::GetBltBufOut(BYTE *pBufOut, UINT uiBufSize)
{
    HRESULT hr = S_OK;
    CAutoLock cObjectLock(&m_MutexLock);

    GOTO_END_NULL(pBufOut, "Failed to CopyTextureStagingToBuf() due to null pointer \n");
    GOTO_END_FALSE(uiBufSize >= (m_iTargetW*m_iTargetH*3/2), "Failed to GetBltBufOut() invalid uiBufSize=[%d] \n", uiBufSize);

    // copy vpblt output to staging, needing a temp texture for copying GPU to CPU memory.
    m_pD3d11DeviceContext->CopyResource(m_pTextureStagingOut, GetBltTextOut());

    hr = CopyTextureStagingToBuf(pBufOut, m_pTextureStagingOut, m_iTargetW, m_iTargetH, m_eFourccOut);
    GOTO_END_FAILED(hr, "Failed to CopyTextureStagingToBuf() \n");

END:
    return hr;
}
