#if 1
#include <stdio.h>
#include <fstream>
#include <Windows.h>
#include <dxgiformat.h>
#include <mfapi.h>
#include <tlhelp32.h>
#include "utils_mft.h"
#include "config.h"

VOID dump_buffer_path(CHAR *szFilePath, BYTE *pBuf, INT iBufLen)
{
    std::fstream inputFileStream;

    if (!szFilePath || !pBuf || iBufLen <= 0)
    {
        PLOG("Failed to dump_buffer_path() due to invalid szFilePath/pBuf/iBufLen \n");
        goto END;
    }

    PLOG(">>>=== dump_buffer_path() szFilePath=[%s] iBufLen=[%d] \n", szFilePath, iBufLen);

    inputFileStream.open(szFilePath, std::ios::out | std::ios::binary);
    inputFileStream.write((char*)pBuf, iBufLen);
    inputFileStream.close();
END:
    PLOG("<<<=== dump_buffer_path() \n");
}


VOID dump_buffer(CHAR *szPathPrefix, CONST GUID stSubType, BYTE *pBuffer, INT iBufferLen, UINT uiWidth, UINT uiHeight)
{
    CHAR szFilePath[256] = {0};
    HRESULT hr = S_OK;
    INT iDumpSize  = 0;

    PLOG(">>>=== dump_buffer() szPathPrefix=[%s] iBufferLen=[%d] uiWidth=[%d] uiHeight=[%d] \n", szPathPrefix, iBufferLen, uiWidth, uiHeight);

    if (!szPathPrefix)
    {
        PLOG("Failed to dump_buffer() due to null pointer szPathPrefix \n");
        goto END;
    }

    if (!pBuffer)
    {
        PLOG("Failed to dump_buffer() due to null pointer pBuffer \n");
        goto END;
    }

    if (stSubType== MFVideoFormat_RGB24)
    {
        iDumpSize = MIN(uiWidth*uiHeight*3, iBufferLen);
        sprintf_s(szFilePath, sizeof(szFilePath), "%s_%dx%d.bgr24", szPathPrefix, uiWidth, uiHeight);
    }
    else if (stSubType == MFVideoFormat_YUY2)
    {
        iDumpSize = MIN(uiWidth*uiHeight*2, iBufferLen);
        sprintf_s(szFilePath, sizeof(szFilePath), "%s_%dx%d.yuy2", szPathPrefix, uiWidth, uiHeight);
    }
    else if (stSubType == MFVideoFormat_I420)
    {
        iDumpSize = MIN(uiWidth*uiHeight*3/2, iBufferLen);
        sprintf_s(szFilePath, sizeof(szFilePath), "%s_%dx%d_i420.yuv", szPathPrefix, uiWidth, uiHeight);
    }
    else if (stSubType == MFVideoFormat_NV12)
    {
        iDumpSize = MIN(uiWidth*uiHeight*3/2, iBufferLen);
        sprintf_s(szFilePath, sizeof(szFilePath), "%s_%dx%d.nv12", szPathPrefix, uiWidth, uiHeight);
    }
    else if (stSubType == MFVideoFormat_MJPG)
    {
        iDumpSize =  iBufferLen;
        sprintf_s(szFilePath, sizeof(szFilePath), "%s_%dx%d.mjpg", szPathPrefix, uiWidth, uiHeight);
    }
    else
    {
        iDumpSize = iBufferLen;
        sprintf_s(szFilePath, sizeof(szFilePath), "%s_%dx%d.unknown_color_format", szPathPrefix, uiWidth, uiHeight);
    }
    dump_buffer_path(szFilePath, pBuffer, iDumpSize);

END:
    PLOG("<<<=== dump_buffer() szFilePath=[%s] \n", szFilePath);
}


VOID printlog(CHAR* szFileName, INT iLineNum, CHAR* szFormat, ...)
{
    static BOOL bsFirstPrint = TRUE;
    CHAR szBuf[512] = {0};
    CHAR *pHead;
    size_t sizetHeadSize;
    va_list args;

    sprintf_s(szBuf, 512, "[%s:%d] ", szFileName, iLineNum);
    pHead = szBuf + strlen(szBuf);
    sizetHeadSize = 512 - strlen(szBuf);

    va_start(args, szFormat);
    vsprintf_s(pHead, sizetHeadSize, szFormat, args);
    va_end(args);

    // delete C:\\Temp\\mft0_fb.log for the first run.
    if (bsFirstPrint)
    {
        bsFirstPrint = FALSE;
        remove("C:\\Temp\\mft0_fb.log");
    }

    std::fstream inputFileStream;
    inputFileStream.open("C:\\Temp\\mft0_fb.log", std::ios::out | std::ios::binary | std::ios_base::app);
    inputFileStream.write((char*)szBuf, strlen(szBuf));
    inputFileStream.close();
}

VOID printlog_mediatype(CHAR *szFileName, INT iLineNum, IMFMediaType *pMediaType, CHAR* szPrefixLog, ...)
{
    CHAR szMayjorGuid[128] = {0};
    CHAR szSubGuid[128]    = {0};
    GUID GuidMajType = GUID_NULL;
    GUID GuidSubType = GUID_NULL;
    UINT uiNumerator, uiDenominator;
    HRESULT hr;
    UINT uiWidth = 0;
    UINT uiHeight = 0;
    UINT uiFps = 0;

    CHAR szBuf[512] = {0};
    va_list args;

    va_start(args, szPrefixLog);
    vsprintf_s(szBuf, 512, szPrefixLog, args);
    va_end(args);

    if (!pMediaType)
    {
        printlog(szFileName, iLineNum, "%s pMediaType=[NULL] \n", szPrefixLog);
        return;
    }

    hr = pMediaType->GetGUID(MF_MT_MAJOR_TYPE, &GuidMajType);
    GOTO_END_FAILED(hr, "Failed to GetGUID() MF_MT_MAJOR_TYPE \n");

    hr = pMediaType->GetGUID(MF_MT_SUBTYPE, &GuidSubType);
    GOTO_END_FAILED(hr, "Failed to GetGUID() MF_MT_SUBTYPE \n");

    hr = MFGetAttributeSize(pMediaType, MF_MT_FRAME_SIZE, &uiWidth, &uiHeight);
    GOTO_END_FAILED(hr, "Failed to MFGetAttributeSize() MF_MT_FRAME_SIZE \n");

    hr = MFGetAttributeRatio(pMediaType, MF_MT_FRAME_RATE, &uiNumerator, &uiDenominator);
    GOTO_END_FAILED(hr, "Failed to MFGetAttributeRatio() MF_MT_FRAME_SIZE \n");

    uiFps = (uiDenominator == 0) ? 0 : (UINT)(uiNumerator/uiDenominator);
    sGuidPrint(szMayjorGuid, sizeof(szMayjorGuid), GuidMajType);
    sGuidPrint(szSubGuid, sizeof(szSubGuid), GuidSubType);

END:
    if (GuidMajType == MFMediaType_Video)
    {
        printlog(szFileName, iLineNum, "%s subtype=[%s] W=[%4d] H=[%4d] FPS=[%3d] \n", szBuf, szSubGuid, uiWidth, uiHeight, uiFps);
    }
    else
    {
        printlog(szFileName, iLineNum, "%s mayjortype=[%s] subtype=[%s] \n", szBuf, szMayjorGuid, szSubGuid);
    }
}

VOID sGuidPrint(CHAR *szGuid, INT iInSize, CONST GUID &guid)
{
    if (!szGuid || iInSize < 40) { return; }

    SecureZeroMemory(szGuid, iInSize);

    if (guid == MFVideoFormat_RGB24)
    {
        sprintf_s(szGuid, 40, "MFVideoFormat_RGB24");
    }
    else if (guid == MFVideoFormat_YUY2)
    {
        sprintf_s(szGuid, 40, "MFVideoFormat_YUY2");
    }
    else if (guid == MFVideoFormat_NV12)
    {
        sprintf_s(szGuid, 40, "MFVideoFormat_NV12");
    }
    else if (guid == MFVideoFormat_I420)
    {
        sprintf_s(szGuid, 40, "MFVideoFormat_I420");
    }
    else if (guid == MFVideoFormat_MJPG)
    {
        sprintf_s(szGuid, 40, "MFVideoFormat_MJPG");
    }
    else
    {
        sprintf_s(szGuid, 40, "%08X-%04X-%04X-%02X%02X-%02X%02X%02X%02X%02X%02X",
        guid.Data1, guid.Data2, guid.Data3,
        guid.Data4[0], guid.Data4[1], guid.Data4[2], guid.Data4[3], guid.Data4[4], guid.Data4[5], guid.Data4[6], guid.Data4[7]);
    }
}

HRESULT I420_to_Nv12(BYTE *pNv12Out, int iNv12OutLen, BYTE *pI420In, int iWidth, int iHeight, BOOL bHorizontalFlipCpu)
{
    HRESULT hr = S_OK;

    if (!pNv12Out || !pI420In)
    {
        PLOG("Failed to I420_to_Nv12() due to null pointer \n");
        hr = E_FAIL;
        goto END;
    }

    if (iNv12OutLen < iHeight * iWidth * 3/2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to I420_to_Nv12() due to not enough iNv12OutLen=[%d] iHeight=[%d] iWidth=[%d] \n", iNv12OutLen, iHeight, iWidth);
        goto END;
    }

    if (bHorizontalFlipCpu)
    {
        for (INT y = 0; y < iHeight; y++)
        {
            for (INT x = 0; x < iWidth; x++)
            {
                pNv12Out[x + y*iWidth] = pI420In[(iWidth - x - 1) + y*iWidth];
            }
        }

        pI420In  += iHeight * iWidth;
        pNv12Out += iHeight * iWidth;

        BYTE *pI420_U = pI420In;
        BYTE *pI420_V = pI420In + (iHeight/2) * (iWidth/2);

        for (INT y = 0; y < iHeight/2; y++)
        {
            for (INT x = 0; x < iWidth/2; x++)
            {
                pNv12Out[2*x + y*iWidth]     = pI420_U[(iWidth/2 - x - 1) + y*iWidth/2];
                pNv12Out[2*x + 1 + y*iWidth] = pI420_V[(iWidth/2 - x - 1) + y*iWidth/2];
            }
        }
    }
    else
    {
        // directly copy to NV12 since they share the same Y-plane
        memcpy_s(pNv12Out, iHeight * iWidth, pI420In, iHeight * iWidth);

        pI420In  += iHeight * iWidth;
        pNv12Out += iHeight * iWidth;

        BYTE *pI420_U = pI420In;
        BYTE *pI420_V = pI420In + (iHeight/2) * (iWidth/2);

        for (int i = 0; i < (iHeight/2) * (iWidth/2); i++)
        {
            pNv12Out[0] = *pI420_U; // I420's U to NV12's UV-plane
            pNv12Out[1] = *pI420_V; // I420's V to NV12's UV-plane
            pI420_U++;
            pI420_V++;
            pNv12Out += 2;
        }
    }
END:
    return hr;
}

HRESULT Nv12_to_I420(BYTE *pI420Out, int iI420OutLen, BYTE *pNv12In, int iWidth, int iHeight)
{
    HRESULT hr = S_OK;

    if (!pI420Out || !pNv12In)
    {
        hr = E_FAIL;
        PLOG("Failed to Nv12_to_I420() due to null pointer \n");
        goto END;
    }

    if (iI420OutLen < iHeight * iWidth * 3/2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to Nv12_to_I420() due to not enough iI420OutLen=[%d] iHeight=[%d] iWidth=[%d] \n", iI420OutLen, iHeight, iWidth);
        goto END;
    }

    // directly copy to NV12 since they share the same Y-plane
    memcpy_s(pI420Out, iHeight * iWidth, pNv12In, iHeight * iWidth);
    pNv12In  += iHeight * iWidth;
    pI420Out += iHeight * iWidth;

    BYTE *pI420_U = pI420Out;
    BYTE *pI420_V = pI420Out + (iHeight/2) * (iWidth/2);

    for (int i = 0; i < (iHeight/2) * (iWidth/2); i++)
    {
        *pI420_U = pNv12In[0];
        *pI420_V = pNv12In[1];
        pI420_U++;
        pI420_V++;
        pNv12In += 2;
    }
END:
    return hr;
}

HRESULT Yuy2_to_Nv12(BYTE *pNv12Out, int iNv12OutLen, BYTE *pYuy2In, int iWidth, int iHeight)
{
    HRESULT hr = S_OK;

    if (!pNv12Out || !pYuy2In)
    {
        hr = E_FAIL;
        PLOG("Failed to Yuy2_to_Nv12() due to null pointer \n");
        goto END;
    }

    if (iNv12OutLen < iHeight * iWidth * 2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to Yuy2_to_Nv12() due to not enough iNv12OutLen=[%d] iHeight=[%d] iWidth=[%d] \n", iNv12OutLen, iHeight, iWidth);
        goto END;
    }

    BYTE *pY_plane = pNv12Out;
    BYTE *pUVplane = pNv12Out + iHeight * iWidth;

    BYTE *pY_SubPixel = pYuy2In;
    BYTE *pU_SubPixel = pYuy2In + 1;
    BYTE *pV_SubPixel = pYuy2In + 3;

    // YUY2 pattern: (Y0 U0 Y1 V0)(Y2 U1 Y3 V1)(Y4 U2 Y5 V2)(Y6 U3 Y7 V3) ...

    for (INT iH = 0; iH < iHeight; iH++)
    {
        for (INT iW = 0; iW < iWidth/2; iW++)
        {
            pY_plane[0] = pY_SubPixel[0]; // Y0
            pY_plane[1] = pY_SubPixel[2]; // Y1
            pY_plane += 2;

            if (iH % 2 == 0) // vertical 2:1 sample
            {
                pUVplane[0] = pU_SubPixel[0]; // U0
                pUVplane[1] = pV_SubPixel[0]; // V0
                pUVplane += 2;
            }

            pY_SubPixel += 4;
            pU_SubPixel += 4;
            pV_SubPixel += 4;
        }
    }
END:
    return hr;
}

HRESULT Nv12_to_Yuy2(BYTE *pYuy2Out, int iYuy2OutLen, BYTE *pNv12In, int iWidth, int iHeight)
{
    HRESULT hr = S_OK;

    if (!pYuy2Out || !pNv12In)
    {
        hr = E_FAIL;
        PLOG("Failed to Nv12_to_Yuy2() due to null pointer \n");
        goto END;
    }

    if (iYuy2OutLen < iHeight * iWidth * 3/2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to Nv12_to_Yuy2() due to not enough iNv12OutLen=[%d] iHeight=[%d] iWidth=[%d] \n", iYuy2OutLen, iHeight, iWidth);
        goto END;
    }

    BYTE *pY_plane = pNv12In;
    BYTE *pUVplane = pNv12In + iHeight * iWidth;

    BYTE *pY_SubPixel = pYuy2Out;
    BYTE *pU_SubPixel = pYuy2Out + 1;
    BYTE *pV_SubPixel = pYuy2Out + 3;

    // YUY2 pattern: (Y0 U0 Y1 V0)(Y2 U1 Y3 V1)(Y4 U2 Y5 V2)(Y6 U3 Y7 V3) ...

    for (INT iH = 0; iH < iHeight; iH++)
    {
        BYTE *pUVplaneTmp = pUVplane;
        for (INT iW = 0; iW < iWidth/2; iW++)
        {
            pY_SubPixel[0] = pY_plane[0]; // Y0
            pY_SubPixel[2] = pY_plane[1]; // Y1
            pY_plane += 2;

            pU_SubPixel[0] = pUVplaneTmp[0]; // U0
            pV_SubPixel[0] = pUVplaneTmp[1]; // V0
            pUVplaneTmp += 2;

            pY_SubPixel += 4;
            pU_SubPixel += 4;
            pV_SubPixel += 4;
        }

        if (iH % 2 != 0)
        {
            pUVplane += iWidth; // point to next row of UV-plane
        }
    }
END:
    return hr;
}

DXGI_FORMAT format_fourcc_to_dxgi(FOURCC eFourcc)
{
    DXGI_FORMAT eFormat = DXGI_FORMAT_UNKNOWN;

    switch (eFourcc)
    {
        case FOURCC_I420:
            eFormat = DXGI_FORMAT_NV12; // FIXME: D3D11 doesn't support I420 format, for I420->NV12 we do via software rather than GPU
            break;
        case FOURCC_YUY2:
            eFormat = DXGI_FORMAT_YUY2;
            break;
        case FOURCC_NV12:
            eFormat = DXGI_FORMAT_NV12;
            break;
        default:
            eFormat = DXGI_FORMAT_UNKNOWN;
            break;
    }
    return  eFormat;
}


FOURCC format_dxgi_to_fourcc(DXGI_FORMAT eFormat)
{
    FOURCC eFourcc = 0;

    switch (eFormat)
    {
        case DXGI_FORMAT_YUY2:
            eFourcc = FOURCC_YUY2;
            break;
        case DXGI_FORMAT_NV12:
            eFourcc = FOURCC_NV12;
            break;
        default:
            eFourcc = 0;
            break;
    }
    return  eFourcc;
}

BOOL GetDwordRegKey(CONST CHAR *szKeyName, DWORD &dwVal)
{
    HKEY    hKey;
    DWORD   dwSizeofDword = sizeof(DWORD);
    BOOL    blKeyExisted  = FALSE;
    LSTATUS lRes          = RegOpenKeyExA(HKEY_LOCAL_MACHINE, VCAM_CONTROL_REGKEY_PATH, 0, KEY_READ, &hKey);

    if (lRes != ERROR_SUCCESS)
    {
        PLOG("Cannot get key=[%s] via RegOpenKeyExA()  \n", szKeyName);
        goto END;
    }

    lRes = RegQueryValueExA(hKey, szKeyName, 0, NULL, (LPBYTE)(&dwVal), &dwSizeofDword);
    if (lRes != ERROR_SUCCESS)
    {
        PLOG("Cannot get key=[%s] via RegQueryValueExA() \n", szKeyName);
        RegCloseKey(hKey);
        goto END;
    }

    RegCloseKey(hKey);

    blKeyExisted = TRUE;
END:
    return blKeyExisted;
}

BOOL SetDwordRegKey(CONST CHAR *szKeyName, DWORD dwVal)
{
    DWORD dwSizeofDword = sizeof(DWORD);
    DWORD dwDisposition;
    HKEY  hKey;
    LSTATUS lRes;
    BOOL blKeyCreated = FALSE;
    PLOG(">>>=== SetDwordRegKey() szKeyName=[%s] dwVal=[%d] \n", szKeyName, dwVal);

    lRes = RegOpenKeyExA(HKEY_LOCAL_MACHINE, VCAM_CONTROL_REGKEY_PATH, 0, KEY_WRITE, &hKey);
    if (lRes != ERROR_SUCCESS)
    {
        lRes = RegCreateKeyExA(HKEY_LOCAL_MACHINE, VCAM_CONTROL_REGKEY_PATH, 0, NULL, 0, KEY_ALL_ACCESS, NULL, &hKey, &dwDisposition);
        if (lRes != ERROR_SUCCESS)
        {
            PLOG("Failed to RegCreateKeyExA() \n");
            goto END;
        }
    }

    lRes = RegSetValueExA(hKey, szKeyName, 0, REG_DWORD, (PBYTE)&dwVal, dwSizeofDword);
    if (lRes != ERROR_SUCCESS)
    {
        RegCloseKey(hKey);
        PLOG("Failed to RegSetValueExA() \n");
        goto END;
    }

    RegCloseKey(hKey);
    blKeyCreated = TRUE;

END:
    PLOG("<<<=== SetDwordRegKey() \n");
    return blKeyCreated;
}

VOID CALLBACK RegKeyListener::ChangeCb(PTP_CALLBACK_INSTANCE Instance, PVOID pvParameter, PTP_WAIT Wait, TP_WAIT_RESULT WaitResult) // Not used.  Required for callback to match
{
    PLOG(">>>=== RegKeyListener::ChangeCb() \n");
    RegKeyListener *pListener = (RegKeyListener*) pvParameter;
    pListener->m_blRegKeyHasBeenChanged = TRUE;
    PLOG("<<<=== RegKeyListener::ChangeCb() \n");
}

HRESULT RegKeyListener::Listen()
{
    HRESULT hr = E_FAIL;
    LONG    lErrorCode;

    if (!m_hRegKey || !m_pRegWaitEvent || !m_hRegEvent)
    {
        m_blRegKeyHasBeenChanged = FALSE;
        lErrorCode      = RegOpenKeyExA(HKEY_LOCAL_MACHINE, VCAM_CONTROL_REGKEY_PATH, 0, KEY_READ, &m_hRegKey);
        m_hRegEvent     = CreateEventEx(NULL, NULL, 0, STANDARD_RIGHTS_ALL | EVENT_MODIFY_STATE);
        m_pRegWaitEvent = CreateThreadpoolWait(ChangeCb, this, NULL);

        if (lErrorCode != ERROR_SUCCESS || !m_hRegEvent || !m_pRegWaitEvent)
        {
            PLOG("Failed to RegKeyListener::Listen() \n");
            goto END;
        }
    }

    // Creates the wait event that listens for changes
    lErrorCode = RegNotifyChangeKeyValue(m_hRegKey, FALSE, REG_NOTIFY_CHANGE_LAST_SET, m_hRegEvent, TRUE);
    if (lErrorCode != ERROR_SUCCESS)
    {
        goto END;
    }
   
    SetThreadpoolWait(m_pRegWaitEvent, m_hRegEvent, NULL);

    WaitForThreadpoolWaitCallbacks(m_pRegWaitEvent, FALSE); 

    hr = S_OK;

END:
    return hr;
}

HRESULT RegKeyListener::Release()
{
    if (m_hRegKey)
    {
        RegCloseKey(m_hRegKey);
        m_hRegKey = NULL;
    }

    if (m_pRegWaitEvent)
    {
        // Unregister the wait by setting the event to NULL.
        SetThreadpoolWait(m_pRegWaitEvent, NULL, NULL);
        // Close the wait.
        CloseThreadpoolWait(m_pRegWaitEvent);
        m_pRegWaitEvent = NULL;
    }

    if (m_hRegEvent)
    {
        // Close the event.
        CloseHandle(m_hRegEvent);
        m_hRegEvent = NULL;
    }
    return S_OK;
}


struct EnumData
{
    DWORD dwProcessId;
    HWND hWnd;
};


BOOL CALLBACK EnumProc(HWND hWnd, LPARAM lParam)
{
    EnumData& ed = *(EnumData*)lParam;
    DWORD dwProcessId = 0x0;

    GetWindowThreadProcessId(hWnd, &dwProcessId);
PLOG("EnumProc() dwProcessId=[%d] \n", dwProcessId);
    if (ed.dwProcessId == dwProcessId)
    {
        // Found a window matching the process ID
        ed.hWnd = hWnd;
        // Report success
        SetLastError( ERROR_SUCCESS );
        // Stop enumeration
        return FALSE;
    }
    return TRUE;
}


HWND FindWindowFromProcessId(DWORD dwProcessId)
{
    EnumData ed = { dwProcessId };

    if (!EnumWindows(EnumProc, (LPARAM)&ed) && (GetLastError() == ERROR_SUCCESS))
    {
        return ed.hWnd;
    }
    return NULL;
}


DWORD GetProcessPidByName(CHAR *szProcessName)
{
    PROCESSENTRY32 PorcessEntry = {0};
    PorcessEntry.dwSize         = sizeof(PROCESSENTRY32);
    DWORD dwPid = 0;

    PLOG(">>>=== GetProcessPidByName() szProcessName=[%s] \n", szProcessName);

    HANDLE hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, NULL);

    if (Process32First(hSnapshot, &PorcessEntry))
    {
        while (Process32Next(hSnapshot, &PorcessEntry))
        {
            if (strcmp(PorcessEntry.szExeFile, szProcessName) == 0)
            {
                dwPid = PorcessEntry.th32ProcessID;
                break;
            }
        }
    }

    CloseHandle(hSnapshot);

    PLOG("<<<=== GetProcessPidByName() dwPid=[%d] \n", dwPid);
    return dwPid;
}


// SendMessageToProcess() doesn't work with MFT0 + Metro App, because we only can receive hWindow of WUDFHost.exe.
// But still leave this function just in case.
HRESULT SendMessageToProcess(CHAR* szProcessName)
{
    DWORD dwProcessPid;
    HWND hWindow;
    HRESULT hr = S_OK;
    BOOL bIsSucceed;

    PLOG(">>>=== SendMessageToProcess() szProcessName=[%d] \n", szProcessName);

    if (!szProcessName)
    {
        PLOG("Failed to SendMessageToProcess() \n");
        hr = E_FAIL;
        goto END;
    }

    dwProcessPid = GetProcessPidByName(szProcessName);
    if (dwProcessPid == 0)
    {
        PLOG("Failed to SendMessageToProcess() return 0 from GetProcessPidByName() \n");
        hr = E_FAIL;
        goto END;
    }

    hWindow = FindWindowFromProcessId(dwProcessPid);
    if (!hWindow)
    {
        hr = E_FAIL;
        PLOG("Failed to SendMessageToProcess() cannot get hWindow from dwProcessPid=[%d] \n", dwProcessPid);
        goto END;
    }

    // https://msdn.microsoft.com/en-us/library/windows/desktop/ms644944(v=vs.85).aspx
    bIsSucceed = PostMessage(hWindow, WM_SYSCOMMAND, SC_MINIMIZE, 0);
    if (!bIsSucceed)
    {
        PLOG("Failed to SendMessageToProcess() \n");
    }

END:
    PLOG("<<<=== SendMessageToProcess() hr=[%d] \n", hr);
    return hr;
}


BOOL IsProcessExisted(CHAR *szProcessName)
{
    BOOL bIsExist               = FALSE;
    PROCESSENTRY32 PorcessEntry = {0};
    PorcessEntry.dwSize         = sizeof(PROCESSENTRY32);

    PLOG(">>>=== IsProcessExisted() szProcessName=[%s] \n", szProcessName);

    HANDLE hSnapshot = CreateToolhelp32Snapshot(TH32CS_SNAPPROCESS, NULL);

    if (Process32First(hSnapshot, &PorcessEntry))
    {
        while (Process32Next(hSnapshot, &PorcessEntry))
        {
            if (strcmp(PorcessEntry.szExeFile, szProcessName) == 0)
            {
                bIsExist = TRUE;
                break;
            }
        }
    }

    CloseHandle(hSnapshot);

    PLOG("<<<=== IsProcessExisted() bIsExist=[%d] \n", bIsExist);
    return bIsExist;
}


HRESULT ShellRunExe(CHAR *szExePath, CHAR *szExeParam)
{
    HRESULT hr = S_OK;

    PLOG(">>>=== ShellRunExe() szExePath=[%s] szExeParam=[%s] \n", szExePath, szExeParam);

    if (!szExePath)
    {
        hr = E_FAIL;
        PLOG("Failed to szExePath() due to null pointer szExePath \n");
        goto END;
    }

    // https://msdn.microsoft.com/en-us/library/windows/desktop/bb762153(v=vs.85).aspx
    HINSTANCE hInstance = ShellExecute(NULL, "open", szExePath, szExeParam, NULL, SW_SHOWDEFAULT);
    if ((INT)hInstance <= 32)
    {
        PLOG("Failed to ShellExecute() \n");
        hr = E_FAIL;
    }

END:
    PLOG("<<<=== ShellRunExe() \n");

    return hr;
}

// This function only for NV12 format, 90 degree clockwise
HRESULT RotateNv12SurfaceDegree90(BYTE* pByteOut, BYTE* pByteIn, UINT uiWidthIn, UINT uiHeightIn)
{
    BYTE *pInPlaneY = pByteIn;
    BYTE *pOutPlaneY = pByteOut;

    // Y plane rotation
    for (INT y = 0; y < uiHeightIn; y++)
    {
        for (INT x = 0; x < uiWidthIn; x++)
        {
            pOutPlaneY[(uiHeightIn- 1 - y) + x*uiHeightIn] = pInPlaneY[x + y*uiWidthIn];
        }
    }

    UINT uiWidthInUV = uiWidthIn/2;
    UINT uiHeightInUV = uiHeightIn/2;
    WORD *pInPlaneUV = (WORD*)(pByteIn + uiWidthIn*uiHeightIn);
    WORD *pOutPlaneUV = (WORD*)(pByteOut + uiWidthIn*uiHeightIn);

    // UV plane rotation
    for (INT y = 0; y < uiHeightInUV; y++)
    {
        for (INT x = 0; x < uiWidthInUV; x++)
        {
            pOutPlaneUV[(uiHeightInUV- 1 - y) + x*uiHeightInUV] = pInPlaneUV[x + y*uiWidthInUV];
        }
    }
    return S_OK;
}


// This function only for NV12 format, 180 degree clockwise
HRESULT RotateNv12SurfaceDegree180(BYTE* pByteOut, BYTE* pByteIn, UINT uiWidthIn, UINT uiHeightIn)
{
    BYTE *pInPlaneY = pByteIn;
    BYTE *pOutPlaneY = pByteOut;

    // Y plane rotation
    for (INT y = 0; y < uiHeightIn; y++)
    {
        for (INT x = 0; x < uiWidthIn; x++)
        {
            pOutPlaneY[(uiWidthIn - 1 - x) + (uiHeightIn - 1 - y)*uiWidthIn] = pInPlaneY[x + y*uiWidthIn];
        }
    }

    UINT uiWidthInUV = uiWidthIn/2;
    UINT uiHeightInUV = uiHeightIn/2;
    WORD *pInPlaneUV = (WORD*)(pByteIn + uiWidthIn*uiHeightIn);
    WORD *pOutPlaneUV = (WORD*)(pByteOut + uiWidthIn*uiHeightIn);

    // UV plane rotation
    for (INT y = 0; y < uiHeightInUV; y++)
    {
        for (INT x = 0; x < uiWidthInUV; x++)
        {
            pOutPlaneUV[(uiWidthInUV - 1 - x) + (uiHeightInUV - 1 - y)*uiWidthInUV] = pInPlaneUV[x + y*uiWidthInUV];
        }
    }
    return S_OK;
}


// This function only for NV12 format, 270 degree clockwise
HRESULT RotateNv12SurfaceDegree270(BYTE* pByteOut, BYTE* pByteIn, UINT uiWidthIn, UINT uiHeightIn)
{
    BYTE *pInPlaneY = pByteIn;
    BYTE *pOutPlaneY = pByteOut;

    // Y plane rotation
    for (INT y = 0; y < uiHeightIn; y++)
    {
        for (INT x = 0; x < uiWidthIn; x++)
        {
            pOutPlaneY[y + (uiWidthIn - 1 - x)*uiHeightIn] = pInPlaneY[x + y*uiWidthIn];
        }
    }

    UINT uiWidthInUV = uiWidthIn/2;
    UINT uiHeightInUV = uiHeightIn/2;
    WORD *pInPlaneUV = (WORD*)(pByteIn + uiWidthIn*uiHeightIn);
    WORD *pOutPlaneUV = (WORD*)(pByteOut + uiWidthIn*uiHeightIn);

    // UV plane rotation
    for (INT y = 0; y < uiHeightInUV; y++)
    {
        for (INT x = 0; x < uiWidthInUV; x++)
        {
            pOutPlaneUV[y + (uiWidthInUV - 1 - x)*uiHeightInUV] = pInPlaneUV[x + y*uiWidthInUV];
        }
    }

    return S_OK;
}


HRESULT CropSurfaceTopBottomNv12(BYTE *pByteOut, BYTE *pByteIn, UINT uiWidthIn, UINT uiHeightIn, UINT uiCropHeightIn)
{
    UINT top = (uiHeightIn - uiCropHeightIn)/2;
    UINT bottom = (uiHeightIn + uiCropHeightIn)/2;

    UINT uiWidthUV_in = uiWidthIn/2;
    BYTE *pPlaneY_in = pByteIn;
    WORD *pPlaneUV_in = (WORD*)(pByteIn + uiWidthIn*uiHeightIn);

    UINT uiWidthY_out = uiWidthIn;
    UINT uiHeightY_out = bottom - top;
    UINT uiWidthUV_out = uiWidthY_out/2;
    UINT uiHeightUV_out = uiHeightY_out/2;
    BYTE *pPlaneY_out = pByteOut;
    WORD *pPlaneUV_out = (WORD*)(pByteOut + uiWidthY_out*uiHeightY_out);

    UINT uiWidthIn_times_y_plus_top; // uiWidthIn*(y+top);
    UINT uiWidthY_out_times_y; // uiWidthY_out*y;

    UINT uiWidthUV_out_times_y; // uiWidthUV_out*y;
    UINT uiWidthUV_in_times_y_plus_top_div2; // uiWidthUV_in*(y+top/2);

    // Y plane
    for (INT y = 0; y < uiHeightY_out; y++)
    {
        uiWidthIn_times_y_plus_top = uiWidthIn*(y+top);
        uiWidthY_out_times_y = uiWidthY_out*y;
        for (INT x = 0; x < uiWidthY_out; x++)
        {
            pPlaneY_out[x + uiWidthY_out_times_y] = pPlaneY_in[x + uiWidthIn_times_y_plus_top];
        }
    }

    // UV plane
    for (INT y = 0; y < uiHeightUV_out; y++)
    {
        uiWidthUV_in_times_y_plus_top_div2 = uiWidthUV_in*(y+top/2);
        uiWidthUV_out_times_y = uiWidthUV_out*y;
        for (INT x = 0; x < uiWidthUV_out; x++)
        {
            pPlaneUV_out[x + uiWidthUV_out_times_y] = pPlaneUV_in[x + uiWidthUV_in_times_y_plus_top_div2];
        }
    }

    return S_OK;
}


HRESULT CopyAndFillTopBottomSurfNv12(BYTE *pByteOut, UINT uiHeightOut, BYTE *pByteIn, UINT uiWidthIn, UINT uiHeightIn)
{
    UINT top = (uiHeightOut-uiHeightIn)/2;
    UINT bottom = (uiHeightOut+uiHeightIn)/2;

    UINT uiWidthUV_in = uiWidthIn/2;
    BYTE *pPlaneY_in = pByteIn;
    WORD *pPlaneUV_in = (WORD*)(pByteIn + uiWidthIn*uiHeightIn);

    UINT uiWidthY_out = uiWidthIn;
    UINT uiHeightY_out = uiHeightOut;
    UINT uiWidthUV_out = uiWidthY_out/2;
    UINT uiHeightUV_out = uiHeightY_out/2;
    BYTE *pPlaneY_out = pByteOut;
    WORD *pPlaneUV_out = (WORD*)(pByteOut + uiWidthY_out*uiHeightY_out);

    UINT uiWidthY_out_times_y; // uiWidthY_out*y
    UINT uiWidthUV_out_times_y; // uiWidthUV_out*y
    // Y plane
    for (INT y = 0; y < uiHeightY_out; y++)
    {
        uiWidthY_out_times_y = uiWidthY_out*y;
        for (INT x = 0; x < uiWidthY_out; x++)
        {
            if (top <= y && y < bottom)
            {
                pPlaneY_out[x + uiWidthY_out_times_y] = *pPlaneY_in;
                pPlaneY_in++;
            }
            else
            {
                pPlaneY_out[x + uiWidthY_out_times_y] = 16;
            }
        }
    }

    // UV plane
    for (INT y = 0; y < uiHeightUV_out; y++)
    {
        uiWidthUV_out_times_y = uiWidthUV_out*y;
        for (INT x = 0; x < uiWidthUV_out; x++)
        {
            if (top/2 <= y && y < bottom/2)
            {
                pPlaneUV_out[x + uiWidthUV_out_times_y] = *pPlaneUV_in;
                pPlaneUV_in++;
            }
            else
            {
                pPlaneUV_out[x + uiWidthUV_out_times_y] = 0x8080; // (128<<8) + 128;
            }
        }
    }

    return S_OK;
}


#else
#include <streams.h>
#include <stdio.h>
#include <fstream>
#include <dxgiformat.h>
#include <D3D9Types.h>
#include <tlhelp32.h>

#include "utils_mft.h"
#include "config.h"

#include <uuids.h>
#include <WinBase.h>
#include <strmif.h>
#include <stdarg.h>

#pragma warning(disable:4244) // disable "warning C4244: 'initializing' : conversion from 'REFERENCE_TIME' to 'INT', possible loss of data"

FOURCC format_subtype_to_fourcc(CONST GUID guidSubtype)
{
    FOURCC eFourcc = D3DFMT_UNKNOWN;

    if (guidSubtype == MEDIASUBTYPE_YUY2)
    {
        eFourcc = FOURCC_YUY2;
    }
    else if (guidSubtype == MEDIASUBTYPE_NV12)
    {
        eFourcc = FOURCC_NV12;
    }
    else if (guidSubtype == MEDIASUBTYPE_I420)
    {
        eFourcc = FOURCC_I420;
    }
    else
    {
        eFourcc = D3DFMT_UNKNOWN;
    }

    return  eFourcc;
}

HRESULT Yuy2_set_black(BYTE *pBufInOut, int iBufLen, int iWidth, int iHeight)
{
    HRESULT hr = S_OK;

    if (!pBufInOut)
    {
        hr = E_FAIL;
        PLOG("Failed to Yuy2_set_black() due to null pointer pBufInOut \n");
        goto END;
    }

    if (iBufLen < iHeight * iWidth * 2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to Yuy2_set_black() due to not enough iBufLen=[%d] iHeight=[%d] iWidth=[%d] \n", iBufLen, iHeight, iWidth);
        goto END;
    }

    BYTE *pSubPixel = pBufInOut;

    // YUY2 pattern: (Y0 U0 Y1 V0)(Y2 U1 Y3 V1)(Y4 U2 Y5 V2)(Y6 U3 Y7 V3) ...
    for (INT iH = 0; iH < iHeight; iH++)
    {
        for (INT iW = 0; iW < iWidth / 2; iW++)
        {
            // https://msdn.microsoft.com/zh-tw/library/windows/desktop/bb530104%28v=vs.85%29.aspx black = (Y, Cr, Cb) = (16, 128, 128)
            pSubPixel[0] = 16;  // Y0
            pSubPixel[1] = 128; // U0
            pSubPixel[2] = 16;  // Y1
            pSubPixel[3] = 128; // V0
            pSubPixel   += 4;
        }
    }
END:
    return hr;
}

HRESULT I420_set_black(BYTE *pBufInOut, int iBufLen, int iWidth, int iHeight)
{
    HRESULT hr = S_OK;

    if (!pBufInOut)
    {
        hr = E_FAIL;
        PLOG("Failed to I420_set_black() due to null pointer pBufInOut \n");
        goto END;
    }

    if (iBufLen < iHeight * iWidth * 3 / 2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to I420_set_black() due to not enough iBufLen=[%d] iHeight=[%d] iWidth=[%d] \n", iBufLen, iHeight, iWidth);
        goto END;
    }

    BYTE *Yplane = pBufInOut;
    BYTE *Uplane = pBufInOut + iHeight * iWidth;
    BYTE *Vplane = Uplane + (iHeight/2) * (iWidth/2);

    // https://msdn.microsoft.com/zh-tw/library/windows/desktop/bb530104%28v=vs.85%29.aspx black = (Y, Cr, Cb) = (16, 128, 128)

    memset(Yplane, 16, iHeight * iWidth);
    memset(Uplane, 128, (iHeight/2) * (iWidth/2));
    memset(Vplane, 128, (iHeight/2) * (iWidth/2));

END:
    return hr;
}

HRESULT NV12_set_black(BYTE *pBufInOut, int iBufLen, int iWidth, int iHeight)
{
    HRESULT hr = S_OK;

    if (!pBufInOut)
    {
        hr = E_FAIL;
        PLOG("Failed to NV12_set_black() due to null pointer pBufInOut \n");
        goto END;
    }

    if (iBufLen < iHeight * iWidth * 3 / 2)
    {
        hr = RPC_S_INVALID_ARG;
        PLOG("Failed to NV12_set_black() due to not enough iBufLen=[%d] iHeight=[%d] iWidth=[%d] \n", iBufLen, iHeight, iWidth);
        goto END;
    }

    BYTE *Yplane = pBufInOut;
    BYTE *UVplane = pBufInOut + iHeight * iWidth;

    // https://msdn.microsoft.com/zh-tw/library/windows/desktop/bb530104%28v=vs.85%29.aspx black = (Y, Cr, Cb) = (16, 128, 128)

    memset(Yplane, 16, iHeight * iWidth);
    memset(UVplane, 128, (iHeight/2) * (iWidth));

END:
    return hr;
}
#endif