#pragma once

//#include <streams.h>
#include <winnt.h>
#include <stdio.h>
#include <dxgiformat.h>

#include <mmiscapi.h>
#include <mfobjects.h>
#include "config.h"

typedef unsigned __int64 QWORD;

#define FOURCC_YUY2 MAKEFOURCC('Y', 'U', 'Y', '2')
#define FOURCC_I420 MAKEFOURCC('I', '4', '2', '0')
#define FOURCC_NV12 MAKEFOURCC('N', 'V', '1', '2')
#define CSC_BUF_SIZE_MAX (1920*1080*2) // max buffer size for 1920*1080 YUY2 and NV12

#define ABS(a) (a >= 0? a: -a)
#define MIN(a, b) ((a <= b)? a : b)
#define MAX(a, b) ((a >= b)? a : b)
#define IS_XY_WITHIN_RECT(x, y, rect) (y >= rect.top && y <= rect.bottom && x >= rect.left && x <= rect.right)

#if MFT0_DEBUG_LOG
#define PLOG(pszFormat, ...) printlog(__FILE__, __LINE__, pszFormat, ##__VA_ARGS__)
#define PLOG_MEDIATYPE(pMediaType, szPrefixLog, ...) printlog_mediatype(__FILE__, __LINE__, pMediaType, szPrefixLog, ##__VA_ARGS__)
#else
#define PLOG(pszFormat, ...)
#define PLOG_MEDIATYPE(pMediaType, szPrefixLog, ...)
#endif

#if MFT0_DEBUG_DUMP
#define DUMP_BUFFER(szPathPrefix, stSubType, pBuffer, iBufferLen, uiWidth, uiHeight) \
    dump_buffer(szPathPrefix, stSubType, pBuffer, iBufferLen, uiWidth, uiHeight)
#else
#define DUMP_BUFFER(szPathPrefix, stSubType, pBuffer, iBufferLen, uiWidth, uiHeight)
#endif


#define GOTO_END_FALSE(bIsTrue, szMsg, ...) do {\
                            if (!(bIsTrue)) {hr = E_FAIL; PLOG(szMsg, ##__VA_ARGS__); goto END;}\
                        } while (FALSE)

#define GOTO_END_NULL(pPointer, szMsg) do {\
                            if (pPointer == NULL) {hr = E_FAIL; PLOG(szMsg); goto END;}\
                        } while (FALSE)

#define GOTO_END_FAILED(hr, szMsg) do {\
                            if (FAILED(hr)) { PLOG(szMsg); goto END;}\
                        } while (FALSE)

const GUID MEDIASUBTYPE_I420 = { 0x30323449, 0x0000, 0x0010, 0x80, 0x00, 0x00, 0xaa, 0x00, 0x38, 0x9b, 0x71};

VOID sGuidPrint(char *szGuid, INT iInSize, const GUID &guid); // 128 bit GUID to human-readable string
VOID printlog(CHAR* szFile, INT iLine, CHAR *szFormat, ...);
VOID printlog_mediatype(CHAR *szFileName, INT iLineNum, IMFMediaType *pMediaType, CHAR* szPrefixLog, ...);
VOID dump_buffer_path(CHAR *szFilePath, BYTE *pBuf, INT iBufLen);
VOID dump_buffer(CHAR *szPathPrefix, CONST GUID stSubType, BYTE *pBuffer, INT iBufferLen, UINT uiWidth, UINT uiHeight);

HRESULT I420_to_Nv12(BYTE *pNv12Out, int iNv12OutLen, BYTE *pI420In, int iWidth, int iHeight, BOOL bHorizontalFlipCpu);
HRESULT Nv12_to_I420(BYTE *pI420Out, int iI420OutLen, BYTE *pNv12In, int iWidth, int iHeight);
HRESULT Yuy2_to_Nv12(BYTE *pNv12Out, int iNv12OutLen, BYTE *pYuy2In, int iWidth, int iHeight);
HRESULT Nv12_to_Yuy2(BYTE *pYuy2Out, int iYuy2OutLen, BYTE *pNv12In, int iWidth, int iHeight);

DXGI_FORMAT format_fourcc_to_dxgi(FOURCC eFourcc);
FOURCC format_subtype_to_fourcc(CONST GUID guidSubtype);

BOOL GetDwordRegKey(CONST CHAR *szKeyName, DWORD &dwVal);
BOOL SetDwordRegKey(CONST CHAR *szKeyName, DWORD dwVal);


HRESULT Yuy2_set_black(BYTE *pBufInOut, int iBufLen, int iWidth, int iHeight);
HRESULT I420_set_black(BYTE *pBufInOut, int iBufLen, int iWidth, int iHeight);
HRESULT NV12_set_black(BYTE *pBufInOut, int iBufLen, int iWidth, int iHeight);

BOOL IsProcessExisted(CHAR *szProcessName);
HRESULT ShellRunExe(CHAR *szExePath, CHAR *szExeParam);

HRESULT RotateNv12SurfaceDegree90(BYTE* pByteOut, BYTE* pByteIn, UINT uiWidthIn, UINT uiHeightIn);
HRESULT RotateNv12SurfaceDegree270(BYTE* pByteOut, BYTE* pByteIn, UINT uiWidthIn, UINT uiHeightIn);
HRESULT RotateNv12SurfaceDegree180(BYTE* pByteOut, BYTE* pByteIn, UINT uiWidthIn, UINT uiHeightIn);

HRESULT CropSurfaceTopBottomNv12(BYTE *pByteOut, BYTE *pByteIn, UINT uiWidthIn, UINT uiHeightIn, UINT uiCropHeightIn);
HRESULT CopyAndFillTopBottomSurfNv12(BYTE *pByteOut, UINT uiHeightOut, BYTE *pByteIn, UINT uiWidthIn, UINT uiHeightIn);


class RegKeyListener
{

public:
    RegKeyListener() {
        m_hRegKey                = NULL;
        m_hRegEvent              = NULL;
        m_pRegWaitEvent          = NULL;
        m_blRegKeyHasBeenChanged = FALSE;
    }

    ~RegKeyListener() {
        Release();
    }

    HRESULT Listen();

    HRESULT Init() {
        m_hRegKey                = NULL;
        m_hRegEvent              = NULL;
        m_pRegWaitEvent          = NULL;
        m_blRegKeyHasBeenChanged = FALSE;
        return S_OK;
    }

    HRESULT Release();
    BOOL CheckIfChanged() {return m_blRegKeyHasBeenChanged;}
    VOID Reset() {m_blRegKeyHasBeenChanged = 0;}

private:

    HKEY     m_hRegKey;
    HANDLE   m_hRegEvent;
    PTP_WAIT m_pRegWaitEvent;
    BOOL     m_blRegKeyHasBeenChanged;

    static VOID CALLBACK ChangeCb(PTP_CALLBACK_INSTANCE Instance, PVOID pvParameter, PTP_WAIT Wait, TP_WAIT_RESULT WaitResult);
};


// NOTE: Copy the class from wxutil.h
// wrapper for whatever critical section we have
class CCritSec {

    // make copy constructor and assignment operator inaccessible

    CCritSec(const CCritSec &refCritSec);
    CCritSec &operator=(const CCritSec &refCritSec);

    CRITICAL_SECTION m_CritSec;

#ifdef DEBUG
public:
    DWORD   m_currentOwner;
    DWORD   m_lockCount;
    BOOL    m_fTrace;        // Trace this one
public:
    CCritSec();
    ~CCritSec();
    void Lock();
    void Unlock();
#else

public:
    CCritSec() {
        InitializeCriticalSection(&m_CritSec);
    };

    ~CCritSec() {
        DeleteCriticalSection(&m_CritSec);
    };

    void Lock() {
        EnterCriticalSection(&m_CritSec);
    };

    void Unlock() {
        LeaveCriticalSection(&m_CritSec);
    };
#endif
};

// NOTE: Copy the class from wxutil.h
// locks a critical section, and unlocks it automatically
// when the lock goes out of scope
class CAutoLock {

    // make copy constructor and assignment operator inaccessible

    CAutoLock(const CAutoLock &refAutoLock);
    CAutoLock &operator=(const CAutoLock &refAutoLock);

protected:
    CCritSec * m_pLock;

public:
    CAutoLock(CCritSec * plock)
    {
        m_pLock = plock;
        m_pLock->Lock();
    };

    ~CAutoLock() {
        m_pLock->Unlock();
    };
};